@@ -1,6 +1,6 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -37,27 +37,29 @@
 
 
 #{compiler
-  build-information compiler-arguments process-command-line
-  default-analysis-database-size default-standard-bindings default-extended-bindings side-effecting-standard-bindings
+  build-information compiler-arguments process-command-line dump-nodes
+  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home compiler-cleanup-hook
-  foreign-type-table-size file-io-only nonwinding-call/cc
-  unit-name insert-timer-checks used-units inlining
-  debug-info-vector-name
+  foreign-type-table-size file-io-only
+  unit-name insert-timer-checks used-units
+  debugging perform-lambda-lifting!
+  posix-is-available register-ffi-macro ffi-include-path
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
-  target-heap-size target-stack-size try-harder default-installation-home 
+  target-heap-size target-stack-size try-harder default-installation-home target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
+  target-initial-heap-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  dependency-list broken-constant-nodes inline-substitutions-enabled
+  broken-constant-nodes inline-substitutions-enabled enable-sharp-greater-read-syntax
   emit-profile profile-lambda-list profile-lambda-index profile-info-vector-name
-  direct-call-ids foreign-type-table first-analysis debug-lambda-list debug-variable-list
+  direct-call-ids foreign-type-table first-analysis
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
+  reorganize-recursive-bindings substitution-table simplify-named-call
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas!
+  transform-direct-lambdas! source-filename compressed-literals literal-compression-threshold
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
@@ -65,14 +67,14 @@
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database write-dependency-list
+  topological-sort print-version print-usage initialize-analysis-database
   product copyright compiler-features default-declarations units-used-by-default words-per-flonum default-debugging-declarations
-  default-profiling-declarations default-optimization-passes
-  foreign-string-result-reserve parameter-limit default-output-filename eq-inline-operator optimizable-rest-argument-operators
+  default-profiling-declarations default-optimization-passes compressed-literals-initializer
+  foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  chop-separator chop-extension current-source-file source-file-changed
+  chop-separator chop-extension display-real-name-table display-line-number-database
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion}
+  foreign-argument-conversion foreign-result-conversion parse-script-header}
 
 
 (include "tweaks")
@@ -98,25 +100,37 @@
 				   (symbol->string oname)
 				   oname) ) ) ]
 		       [(memq 'to-stdout options) #f]
-		       [else default-output-filename] ) ]
+		       [else (make-pathname #f (if filename (pathname-file filename) "out") "c")] ) ]
 	[home (chop-separator (or (getenv "CHICKEN_HOME") default-installation-home ""))]
 	[ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))]
-	[ilimit (memq 'inline-limit options)]
-	[ipasses (memq 'inline-passes options)]
 	[strict (memq 'strict options)]
 	[strict-srfi0 (memq 'strict-srfi-0 options)]
-	[ipass 1]
+	[strict-letrec (memq 'strict-letrec options)]
+	[strictr (memq 'strict-reader options)]
 	[opasses default-optimization-passes]
 	[time0 #f]
 	[time-breakdown #f]
 	[forms '()]
 	[cleanup-forms '(((##sys#implicit-exit-handler)))]
 	[profile (memq 'profile options)]
-	[dsize (memq 'database-size options)]
 	[hsize (memq 'heap-size options)]
+	[hisize (memq 'heap-initial-size options)]
+	[hgrowth (memq 'heap-growth options)]
+	[hshrink (memq 'heap-shrinkage options)]
+	[kwstyle (memq 'keyword-style options)]
+	[lcthreshold (memq 'compress-literals options)]
 	[srfi7 (memq 'srfi-7 options)]
+	[hygienic-at-rt (memq 'hygienic-at-run-time options)]
+	[uunit (memq 'unit options)]
+	[a-only (memq 'analyze-only options)]
+	[dynamic (memq 'dynamic options)]
+	[script-extra-units #f]
+	[dumpnodes #f]
+	[ffi-mode (memq 'ffi options)]
 	[ssize (or (memq 'nursery options) (memq 'stack-size options))] )
 
+    (define (cputime) (##sys#fudge 6))
+
     (define (print-header mode dbgmode)
       (when verbose (printf "pass: ~a~%~!" mode))
       (and (memq dbgmode debugging-chicken)
@@ -126,7 +140,9 @@
 
     (define (print-node mode dbgmode n)
       (when (print-header mode dbgmode)
-	(pretty-print (build-expression-tree n)) ) )
+	(if dumpnodes
+	    (dump-nodes n)
+	    (pretty-print (build-expression-tree n)) ) ) )
 
     (define (print-db mode dbgmode db pass)
       (when (print-header mode dbgmode)
@@ -165,34 +181,63 @@
 	      [else '()] ) ) )
 
     (define (begin-time)
-      (when time-breakdown (set! time0 (current-milliseconds))) )
+      (when time-breakdown (set! time0 (cputime))) )
 
     (define (end-time pass)
       (when time-breakdown
-	(printf "milliseconds needed for ~a: ~s~%" pass (- (current-milliseconds) time0)) ) )
+	(printf "milliseconds needed for ~a: ~s~%" pass (- (cputime) time0)) ) )
 
     (define (install-macros)
-      (##syncase#install-macro-package (or strict strict-srfi0) strict-srfi0) )
+      (fluid-let ([##sys#read-line-counter ##sys#read-line-counter])
+	(##syncase#install-macro-package (or strict strict-srfi0) (not strict) (or strict strict-srfi0 strict-letrec)) ) )
 
-    (define (output-dependency-list out)
-      (fprintf out "~A: " outfile)
-      (write-dependency-list dependency-list 70 out) )
+    (define (strictness s s0)
+      (cond [s
+	     (set! strict #t)
+	     (when verbose (display "Strict mode\n"))
+	     (set! ##sys#strict-mode #t)
+	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-0 #:srfi-16 #:srfi-9 #:srfi-26)
+	     (register-feature! #:strict)
+	     (##sys#undefine-non-standard-macros '()) ]
+	    [s0 
+	     (set! strict-srfi0 #t)
+	     (when verbose (display "Strict mode with SRFI-0\n"))
+	     (set! ##sys#strict-mode #t)
+	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-16 #:srfi-9 #:srfi-26)
+	     (register-feature! #:strict)
+	     (##sys#undefine-non-standard-macros '(cond-expand)) ] )
+      (if (or s s0 strictr) 
+	  (begin
+	    (strict-reader #t)
+	    (case-sensitive #f) )
+	  (enable-sharp-greater-read-syntax) ) )
 
+    (define (read-form in)
+      (if ffi-mode
+	  (let ([s (read-string #f in)])
+	    (if (string=? s "")
+		(end-of-file)
+		`(declare 
+		   (foreign-declare ,s)
+		   (foreign-parse ,s) ) ) )
+	  (##sys#read in infohook) ) )
+
+    (when (and uunit (pair? (cdr uunit)))
+      (set! unit-name (string->c-identifier (stringify (cadr uunit)))) )
     (set! debugging-chicken 
       (append-map
        (lambda (do)
 	 (map (lambda (c) (string->symbol (string c)))
 	      (string->list do) ) )
        (collect-options 'debug) ) )
-    (set! try-harder (memq 'optimize options))
-    (when dsize (set! analysis-database-size (arg-val (cadr dsize))))
+    (set! try-harder #t)
+    (set! dumpnodes (memq '|D| debugging-chicken))
     (when (memq 't debugging-chicken) (##sys#start-timer))
     (when (memq 'b debugging-chicken) (set! time-breakdown #t))
     (when (memq 'no-warnings options) 
       (when verbose (printf "Warnings are disabled~%~!"))
       (set! warnings-enabled #f)
       (set! ##sys#warnings-enabled #f) )
-    (when (memq 'inline options) (set! inlining #t))
     (when (memq 'optimize-leaf-routines options) (set! optimize-leaf-routines #t))
     (when (memq 'unsafe options) 
       (set! unsafe #t)
@@ -200,112 +245,103 @@
     (when (memq 'disable-interrupts options) (set! insert-timer-checks #f))
     (when (memq 'fixnum-arithmetic options) (set! number-type 'fixnum))
     (when (memq 'block options) (set! block-compilation #t))
-    (when (memq 'no-fancy-ports options) (set! file-io-only #t))
-    (when (memq 'no-winding-callcc options) (set! nonwinding-call/cc #t))
-    (when (memq 'case-sensitive options) 
-      (when verbose (printf "Identifiers and symbols are case sensitive~%~!"))
-      (register-feature! 'case-sensitive)
-      (##sys#case-sensitive #t) )
-    (set! expand-only (memq 'expand-only options))
-    (set! write-dependencies (memq 'write-dependencies options))
-    (cond [(memq 'dependency-output options) => (lambda (o) (set! dependency-output (cadr o)))])
+    (when (memq 'case-insensitive options) 
+      (when verbose (printf "Identifiers and symbols are case insensitive~%~!"))
+      (register-feature! 'case-insensitive)
+      (case-sensitive #f) )
+    (when kwstyle
+      (cond [(string=? "prefix" (cadr kwstyle))
+	     (keyword-style #:prefix) ]
+	    [(string=? "suffix" (cadr kwstyle))
+	     (keyword-style #:suffix) ] ) )
+    (when lcthreshold
+      (let ([t (cadr lcthreshold)])
+	(set! literal-compression-threshold 
+	  (or (string->number t)
+	      (quit "invalid argument to `-compress-literals' option: ~A" t) ) ) ) )
     (set! verbose-mode verbose)
     (set! ##sys#read-error-with-line-number #t)
-    (set! ##sys#load-verbose #f)
     (set! ##sys#include-pathnames
       (append (map chop-separator (collect-options 'include-path))
 	      ##sys#include-pathnames
 	      ipath
 	      (if home (list home) '()) ) )
-    (when verbose
-      (printf "Nursery size is ~A~%Heap size is ~A~%~!" ssize hsize) )
+    (when (and outfile filename (string=? outfile filename))
+      (quit "source- and output-filename are the same") )
 
     ;; Handle feature options:
     (for-each register-feature! (collect-options 'feature))
     (for-each unregister-feature! (collect-options 'no-feature))
 
+    ;; Handle FFI defines and include-paths:
+    (for-each register-ffi-macro (collect-options 'ffi-define))
+    (set! ffi-include-path (collect-options 'ffi-include-path))
+
     ;; Load extensions:
+    (set! ##sys#features (cons #:compiler-extension ##sys#features))
     (let ([extends (collect-options 'extend)])
       (when verbose
-	(printf "loading extensions...~%~!")
-	(set! ##sys#load-verbose #t) )
-      (for-each load extends) )
+	(printf "Loading extensions...~%~!")
+	(load-verbose #t) )
+      (for-each (lambda (f) (load (##sys#resolve-include-filename f))) extends) )
+    (set! ##sys#features (delete #:compiler-extension ##sys#features eq?))
 
     (set! ##sys#features (cons '#:compiling ##sys#features))
-    ;; Install hilevel macros:
-    (cond [(memq 'hygienic options)
-	   (when verbose 
-	     (printf "Using hygienic macros~A~!"
-		     (cond [strict " - strict mode"]
-			   [strict-srfi0 " - strict mode with SRFI-0"]
-			   [else ""] ) ) )
-	   (install-macros) ]
-	  [strict
-	   (when verbose (display "Strict mode\n"))
-	   (set! ##sys#strict-mode #t)
-	   (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-0 #:srfi-16)
-	   (register-feature! #:strict)
-	   (##sys#undefine-non-standard-macros '()) ]
-	  [strict-srfi0
-	   (when verbose (display "Strict mode with SRFI-0\n"))
-	   (set! ##sys#strict-mode #t)
-	   (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-16)
-	   (register-feature! #:strict)
-	   (##sys#undefine-non-standard-macros '(cond-expand)) ] )
+    ;; Install highlevel macros:
+    (if (memq 'hygienic options)
+	(begin
+	  (when verbose 
+	    (printf "Using hygienic macros~!") )
+	  (install-macros) )
+	(begin
+	  (##sys#provide 'match) 
+	  (when (or strict strict-srfi0 strict-letrec) (##sys#enable-strict-letrec) ) ) )
 
-    (when ipasses 
-      (set! inline-passes (arg-val (cadr ipasses))) )
-    (when ilimit
-      (set! inline-limit (arg-val (cadr ilimit))) )
+    ;; Loads required extensions:
+    (let ([ids (map string->symbol (collect-options 'require-for-syntax))])
+      (for-each
+       (lambda (id)
+	 (when verbose (printf "Requiring ~A ...~%" id))
+	 (##sys#require id) )
+       ids)
+      (let ([rs (##sys#lookup-runtime-requirements ids)])
+	(when (pair? rs)
+	  (set! initforms
+	    (append initforms (map (lambda (r) `(##sys#require ',r)) rs)) ) ) ) )
+
+    (when (memq 'compile-time-macros-only options)
+      (set! ##sys#enable-runtime-macros #f) )
+    (strictness strict strict-srfi0)
     (set! target-heap-size
       (if hsize
 	  (arg-val (cadr hsize))
 	  (and-let* ([hsize default-default-target-heap-size]
 		     [(not (zero? hsize))] )
 	    hsize) ) )
+    (set! target-initial-heap-size (and hisize (arg-val (cadr hisize))))
+    (set! target-heap-growth (and hgrowth (arg-val (cadr hgrowth))))
+    (set! target-heap-shrinkage (and hshrink (arg-val (cadr hshrink))))
     (set! target-stack-size
       (if ssize
 	  (arg-val (cadr ssize))
 	  (and-let* ([ssize default-default-target-stack-size]
 		     [(not (zero? ssize))] )
 	    ssize) ) )
-    (set! emit-debug-info (memq 'emit-debug-info options))
-    (when (memq 'debug-calls options)
-      (set! debug-calls #t)
-      (set! emit-debug-info #t) )
-    (when (memq 'debug-loops options)
-      (set! debug-calls #t)
-      (set! debug-loops #t)
-      (set! emit-debug-info #t) )
-    (set! emit-trace-info (not (or emit-debug-info (memq 'no-trace options))))
-    (set! ##sys#print-qualifiers #t)
+    (set! emit-trace-info (not (memq 'no-trace options)))
     (when (memq 'm debugging-chicken) (set-gc-report! #t))
     (when (memq 'usual-integrations options)
       (set! standard-bindings default-standard-bindings)
       (set! extended-bindings default-extended-bindings) )
-    (when emit-debug-info
-      (set! initforms (append initforms default-debugging-declarations)) )
     (when verbose
-      (printf "Debugging info: ~A~%~!"
-	      (if emit-debug-info
-		  (if debug-calls
-		      (if debug-loops
-			  "procedures, assignments, calls and loops"
-			  "procedures, assignments and calls")
-		      "procedures and assignments")
-		  (if emit-trace-info
-		      "stacktrace"
-		      "none") ) ) )
+      (printf "debugging info: ~A~%~!"
+	      (if emit-trace-info
+		  "stacktrace"
+		  "none") ) )
     (when profile
-      (when emit-debug-info
-	(warning "debugging information is not available in profiled applications - ignored") 
-	(set! emit-debug-info #f) )
-      (if (pair? (cdr profile))
-	  (set! emit-profile (cadr profile))
-	  (quit "missing argument to `-profile' option") )
+      (set! emit-profile #t)
       (set! initforms (append initforms default-profiling-declarations))
       (when verbose
-	(printf "Generating profile ~S~%~!" emit-profile) ) )
+	(printf "Generating profile~%~!" emit-profile) ) )
 
     (cond ((memq 'version options)
 	   (print-version)
@@ -318,7 +354,8 @@
 
 	   ;; Display header:
 	   (unless (memq 'quiet options)
-	     (printf "compiling file `~a' ...~%" filename) )
+	     (printf "compiling `~a' ...~%" filename) )
+	   (set! source-filename filename)
 	   (debugging 'r "options" options)
 	   (debugging 'r "debugging options" debugging-chicken)
 	   (debugging 'r "home directory" home)
@@ -327,7 +364,6 @@
 	   (debugging 'r "compiler features" compiler-features)
 
 	   ;; Read toplevel expressions:
-	   (set! ##sys#read-line-counter 1)
 	   (set! ##sys#line-number-database (make-vector line-number-database-size '()))
 	   (let ([prelude (collect-options 'prelude)]
 		 [postlude (collect-options 'postlude)] 
@@ -336,9 +372,6 @@
 			 (list filename)
 			 (collect-options 'epilogue) ) ]  )
 
-	     (when write-dependencies
-	       (set! dependency-list (append-reverse files dependency-list)) )
-
 	     (let ([proc (user-read-pass)])
 	       (cond [proc
 		      (when verbose (printf "User read pass...~%~!"))
@@ -353,38 +386,61 @@
 			(let* ([f (car files)]
 			       [in (check-and-open-input-file f)]
 			       [c0 (peek-char in)] )
-			  (set! current-source-file f)
-			  (set! source-file-changed #t)
+			  (set! ##sys#read-line-counter 1)
 
 			  ;; Check for script header:
-			  (when (char=? #\# c0)
-			    (read-char in)
-			    (unless (char=? #\! (peek-char in))
-			      (quit "sorry - file may not begin with `#' character if not a script") )
-			    (read-char in)
-			    (let ([header (open-output-string)]
-				  [cleanup #t] )
-			      (do ([c (read-char in) (read-char in)])
-				  ((char=? c #\newline)
-				   (set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) )
-				(if (eof-object? c)
-				    (quit "unexpected end of file - incorrect header")
-				    (write-char c header) ) )
-			      (case (parse-script-header (get-output-string header))
-				[(script) 
-				 (set! forms (cons '(define command-line-arguments (make-parameter (cdr (argv)))) forms))
-				 (set! cleanup #f) ]
-				[(srfi7) 
-				 (set! srfi7 #t)
-				 (install-macros) ]
-				[(r5rs srfi0) (install-macros)] 
-				[(#f) (warning "unrecognized script header - might be corrupt")] )
-			      (when cleanup (set! cleanup-forms (cons '(##sys#script-main '0 '1) cleanup-forms))) ) )
+			  (cond [ffi-mode]
+				[(char=? #\# c0)
+				 (read-char in)
+				 (unless (char=? #\! (peek-char in))
+				   (quit "source file may not begin with `#' character if not a script") )
+				 (read-char in)
+				 (let ([header (open-output-string)]
+				       [cleanup #t] )
+				   (do ([c (read-char in) (read-char in)])
+				       ((char=? c #\newline)
+					(set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) )
+				     (if (eof-object? c)
+					 (quit "unexpected end of file - incorrect header")
+					 (write-char c header) ) )
+				   (when verbose (printf "Compiling UNIX script~%~!"))
+				   (case (parse-script-header (get-output-string header) in)
+				     [(script)
+				      (set! cleanup #f)
+				      (set! script-extra-units #t) ]
+				     [(chicken) (set! script-extra-units #t)]
+				     [(chicken-hygienic) 
+				      (set! script-extra-units #t)
+				      (install-macros) ]
+				     [(srfi7) 
+				      (set! srfi7 #t)
+				      (strictness #t #f)
+				      (install-macros) ]
+				     [(r5rs)
+				      (install-macros)
+				      (strictness #t #f) ]
+				     [(srfi0)
+				      (install-macros)
+				      (strictness #t #t) ]
+				     [(#f) (warning "unrecognized script header - might be corrupt")]
+				     [else (strictness #t #f)] ) ; r4rs & ieee
+				   (when cleanup
+				     (set! cleanup-forms
+				       (cons '(##sys#script-main '0 '0 main)
+					     cleanup-forms) ) ) ) ]
+				[(char=? #\@ c0)
+				 (read-line in)
+				 (set! script-extra-units #t)
+				 (when verbose (printf "Compiling Windows batch file~%~!")) ] )
 
-			  (do ([x (##sys#read in infohook) (##sys#read in infohook)])
-			      ((eof-object? x) (close-checked-input-file in f))
+			  (do ([x (read-form in) (read-form in)])
+			      ((eof-object? x) 
+			       (close-checked-input-file in f) )
 			    (set! forms (cons x forms)) ) ) ) ] ) ) )
 
+	   ;; Visit files:
+	   (for-each visit (collect-options 'visit))
+
 	   ;; Start compilation passes:
 	   (let ([proc (user-preprocessor-pass)])
 	     (when proc
@@ -399,31 +455,39 @@
 		 (car forms)
 		 ##sys#syntax-error-hook) ) ) )
 
+	   (set! script-extra-units
+	     (if script-extra-units
+		 (map (compose string->c-identifier stringify)
+		      (append
+		       '(extras srfi-1 srfi-4 lolevel srfi-18 srfi-37 format tinyclos regex script-utils tcp)
+		       (if posix-is-available
+			   '(posix)
+			   '() ) ) )
+		 '() ) )
+
 	   (print "source" '|1| forms)
 	   (begin-time)
+	   (let ([us (collect-options 'uses)])
+	     (unless (null? us)
+	       (set! forms (cons `(declare (uses ,@us)) forms)) ) )
 	   (let* ([exps0 (map canonicalize-expression (append initforms forms))]
 		  [pvec (gensym)]
 		  [plen (length profile-lambda-list)]
 		  [exps (append
 			 (map (lambda (ic) `(set! ,(cdr ic) ',(car ic))) immutable-constants)
-			 (map (lambda (n) `(##core#callunit ,n)) used-units)
-			 (if emit-debug-info
-			     `((set! ,debug-info-vector-name 
-				 (make-vector ',(+ (length debug-lambda-list) (length debug-variable-list))) ) )
-			     '() )
-			 (map (lambda (dl) `(##core#inline "C_i_setslot" ,debug-info-vector-name ',(car dl) ',(cdr dl)))
-			      (append debug-lambda-list debug-variable-list) )
-			 (if emit-debug-info
-			     `((##sys#register-debug-info ,debug-info-vector-name ))
-			     '() )
-			 (if (and emit-debug-info (not unit-name))
-			     '((if ##sys#break-on-startup (##sys#break) '#f))
+			 (map (lambda (n) `(##core#callunit ,n))
+			      (begin
+				(set! used-units (append used-units script-extra-units))
+				used-units) ) 
+			 (if hygienic-at-rt
+			     '((##core#callunit "syntax_case")
+			       (##syncase#install-macro-package '#t '#t) )
 			     '() )
 			 (if emit-profile
 			     `((set! ,profile-info-vector-name 
 				 (##sys#register-profile-info
 				  ',plen
-				  ',(if unit-name #f emit-profile) ) ) )
+				  ',(if unit-name #f "PROFILE") ) ) )
 			     '() )
 			 (map (lambda (pl)
 				`(##core#inline 
@@ -432,49 +496,77 @@
 				  ',(* profile-info-entry-size (car pl)) 
 				  ',(cdr pl) ) )
 			      profile-lambda-list)
+			 (let ([is (fold (lambda (clf r)
+					   `(let ([,(gensym) (set! ,(car clf) (##sys#read-from-string ',(cdr clf)))])
+					      ,r) )
+					 '(##core#undefined) 
+					 compressed-literals) ] )
+			   (if compressed-literals-initializer
+			       `((##core#set! ,compressed-literals-initializer 
+					      (lambda () ,is) ) )
+			       (list is) ) )
 			 exps0
-			 (if (not unit-name) 
+			 (if (and (not unit-name) (not dynamic))
 			     cleanup-forms
 			     '() ) ) ] )
-	     (set! always-bound (cons debug-info-vector-name always-bound))
 
-	     (when (and nonwinding-call/cc emit-debug-info)
-	       (warning 
-		"`non-winding-callc' option/declaration is not available in combination with debugging information - ignored")
-	       (set! nonwinding-call/cc #f) )
+	     (when (debugging '|N| "real name table:")
+	       (display-real-name-table) )
+	     (when (debugging 'n "line number database:")
+	       (display-line-number-database) )
 
-	     (when (and nonwinding-call/cc emit-profile)
-	       (warning 
-		"`non-winding-callc' option/declaration is not available in combination with profiling information - ignored")
-	       (set! nonwinding-call/cc #f) )
+	     (when hygienic-at-rt
+	       (set! used-units (cons "syntax_case" used-units))
+	       (when unit-name
+		 (warning "enabling hygienic macros at runtime will apply to all programs using this unit") ) )
+
+	     (when (and block-compilation unit-name)
+	       (warning "compilation of library unit in block-mode - globals may not be accessible outside this unit"
+			unit-name) )
+
+	     (set! ##sys#line-number-database line-number-database-2)
+	     (set! line-number-database-2 #f)
+
+	     (end-time "canonicalization")
+	     (print "canonicalized" '|2| exps)
+
+	     (when (memq 'check-syntax options) (exit))
 
 	     (let ([proc (user-pass)])
 	       (when proc
 		 (when verbose (printf "User pass...~%~!"))
-		 (set! exps (map proc exps)) ) )
+		 (begin-time)
+		 (set! exps (map proc exps))
+		 (end-time "user pass") ) )
 
-	     (set! ##sys#line-number-database line-number-database-2)
-	     (set! line-number-database-2 #f)
-	     (end-time "canonicalization")
-	     (print "canonicalized" '|2| exps)
+	     (let* ([node0 (make-node
+			    'lambda '(())
+			    (list (build-node-graph
+				   (canonicalize-begin-body exps) ) ) ) ] 
+		    [proc (user-pass-2)] )
+	       (when proc
+		 (when verbose (printf "Secondary user pass...~%"))
+		 (begin-time)
+		 (set! first-analysis #f)
+		 (let ([db (analyze-expression node0)])
+		   (print-db "analysis (u)" '|0| db 0)
+		   (end-time "pre-analysis (u)")
+		   (begin-time)
+		   (proc node0 db)
+		   (end-time "secondary user pass")
+		   (print-node "secondary user pass" '|U| node0) ) )
 
-	     (when write-dependencies
-	       (cond [dependency-output
-		      (call-with-output-file dependency-output output-dependency-list) ]
-		     [else
-		      (output-dependency-list (current-output-port)) ] ) )
-	     (cond [(memq 'check-syntax options) (exit)]
-		   [expand-only
-		    (pprint-expressions-to-file 
-		     (if (not unit-name)
-			 (butlast exps)
-			 exps) outfile)
-		    (exit) ] )
-
-	     (let ([node0 (make-node
-			   'lambda '(())
-			   (list (build-node-graph
-				  (canonicalize-begin-body exps) ) ) ) ] )
+	       (when (memq 'lambda-lift options)
+		 (begin-time)
+		 (set! first-analysis #f)
+		 (let ([db (analyze-expression node0)])
+		   (print-db "analysis" '|0| db 0)
+		   (end-time "pre-analysis")
+		   (begin-time)
+		   (perform-lambda-lifting! node0 db)
+		   (end-time "lambda lifting")
+		   (print-node "lambda lifted" '|L| node0) )
+		 (set! first-analysis #t) )
 	       (set! ##sys#line-number-database #f)
 	       (set! constant-table #f)
 	       (set! inline-table #f)
@@ -495,7 +587,7 @@
 		     (end-time "analysis")
 		     (print-db "analysis" '|4| db i)
 
-		     (when (pair? debugging-chicken) (print-program-statistics db))
+		     (when (memq 's debugging-chicken) (print-program-statistics db))
 
 		     (cond [(and progress (or try-harder (<= i default-optimization-passes)))
 			    (debugging 'p "optimization pass" i)
@@ -505,46 +597,21 @@
 			      (end-time "optimization")
 			      (print-node "optimized-iteration" '|5| node2)
 
-			      (when (and inlining (not progress-flag) (<= ipass inline-passes))
-				(debugging 'p "inlining pass" ipass)
-				(begin-time)
-				(set! ipass (add1 ipass))
-				(let ([ics (find-inlining-candidates node2 db)])
-				  (when (pair? ics)
-				    (let ([mvars (perform-inlining! ics node2 db)])
-				      (set! always-bound (append (unzip1 mvars) always-bound))
-				      ;; Mutate body of toplevel lambda node:
-				      (set-car!
-				       (node-subexpressions node2)
-				       (fold
-					(lambda (mv r)
-					  (make-node 
-					   'let
-					   (list (car mv))
-					   (list (qnode (cdr mv)) r) ) )
-					(first (node-subexpressions node2))
-					mvars) )
-				      (set! progress-flag #t) 
-				      (print-node "inlined" '|6| node2) ) ) )
-				(end-time "inlining") )
-
 			      (cond [progress-flag (loop (add1 i) node2 #t)]
 				    [(not inline-substitutions-enabled)
 				     (debugging 'p "rewritings enabled...")
 				     (set! inline-substitutions-enabled #t)
 				     (loop (add1 i) node2 #t) ]
 				    [optimize-leaf-routines
-
-				     (begin-time) ; one more analysis pass
+				     (begin-time)
 				     (let ([db (analyze-expression node2)])
 				       (end-time "analysis")
-
 				       (begin-time)
 				       (let ([progress (transform-direct-lambdas! node2 db)])
 					 (end-time "leaf routine optimization")
 					 (loop (add1 i) node2 progress) ) ) ]
-				    [else (loop (add1 i) node2 #f)] ) ) ] 
-
+				    [else (loop (add1 i) node2 #f)] ) ) ]
+			   
 			   [else
 			    (print-node "optimized" '|7| node2)
 
@@ -552,6 +619,12 @@
 			    (let ([node3 (perform-closure-conversion node2 db)])
 			      (end-time "closure conversion")
 			      (print-db "final-analysis" '|8| db i)
+			      (let ([upap (user-post-analysis-pass)])
+				(when upap 
+				  (upap db
+					(lambda (k p) (cut get db <> <>))
+					(lambda (k p x) (cut put! db <> <>)) ) ) )
+			      (when a-only (exit 0))
 			      (print-node "closure-converted" '|9| node3)
 
 			      (begin-time)
@@ -560,7 +633,7 @@
 
 				(let ([out (if outfile (open-output-file outfile) (current-output-port))])
 				  (begin-time)
-				  (generate-code literals lambdas out filename)
+				  (generate-code literals lambdas out filename db)
 				  (end-time "code generation")
 
 				  (when outfile (close-output-port out))
@@ -572,7 +645,7 @@
 
 ;;; Parse script header:
 
-(define (parse-script-header line)
+(define (parse-script-header line in)
   (let ([len (string-length line)])
     (define (cmp-suffix str)
       (let ([index (- len (string-length str))])
@@ -580,6 +653,11 @@
 	     (string=? str (substring line index len))
 	     (memq (string-ref line (sub1 index)) '(#\space #\/)) ) ) )
     (cond [(cmp-suffix "-script") 'script]
+	  [(cmp-suffix "-script-meta")
+	   (read-line in)
+	   'script]
+	  [(cmp-suffix "scheme-chicken") 'chicken]
+	  [(cmp-suffix "scheme-chicken-hygienic") 'chicken-hygienic]
 	  [(cmp-suffix "scheme-r4rs") 'r4rs]
 	  [(cmp-suffix "scheme-r5rs") 'r5rs]
 	  [(cmp-suffix "scheme-srfi-7") 'srfi7]
