@@ -1,6 +1,6 @@
 ;;; c-backend.scm - C-generating backend for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -38,18 +38,19 @@
 
 #{compiler
   build-information compiler-arguments process-command-line
-  default-analysis-database-size default-standard-bindings default-extended-bindings side-effecting-standard-bindings
+  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  installation-home optimization-iterations
-  foreign-type-table-size file-io-only nonwinding-call/cc
+  installation-home optimization-iterations debugging cleanup
+  foreign-type-table-size file-io-only
   unit-name insert-timer-checks used-units inlining external-variables
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
-  target-heap-size target-stack-size try-harder default-installation-home 
+  target-heap-size target-stack-size try-harder default-installation-home target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
+  target-initial-heap-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  dependency-list broken-constant-nodes inline-substitutions-enabled
+  broken-constant-nodes inline-substitutions-enabled
   direct-call-ids foreign-type-table first-analysis block-variable-literal?
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
@@ -64,9 +65,9 @@
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
-  generate-external-variables
+  generate-external-variables real-name real-name2
   product copyright compiler-features default-declarations units-used-by-default words-per-flonum
-  foreign-string-result-reserve parameter-limit default-output-filename eq-inline-operator optimizable-rest-argument-operators
+  foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   default-optimization-iterations generate-foreign-callback-header generate-foreign-callback-stub-prototypes
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
@@ -95,11 +96,13 @@
 
 ;;; Generate target code:
 
-(define (generate-code literals lambdas out source-file)
+(define (generate-code literals lambdas out source-file db)
 
   (define (find-lambda id)
     (or (find (lambda (ll) (eq? id (lambda-literal-id ll))) lambdas)
 	(bomb "can't find lambda" id) ) )
+
+  (define (slashify s) (string-translate (->string s) "\\" "/"))
   
   (define (expression node temps ll)
 
@@ -201,10 +204,13 @@
 	  ((##core#global)
 	   (let ([index (first params)]
 		 [safe (second params)] 
-		 [block (third params)] ) ; block implies safe
-	     (cond [block (gen "lf[" index "]")]
-		   [safe (gen "C_retrieve(lf[" index "])")]
-		   [else (gen "*((C_word*)lf[" index "]+1)")] ) ) )
+		 [block (third params)] )
+	     (cond [block
+		    (if safe
+			(gen "lf[" index "]")
+			(gen "C_retrieve2(lf[" index "]," (c-ify-string (symbol->string (fourth params))) #\)) ) ]
+		   [safe (gen "*((C_word*)lf[" index "]+1)")]
+		   [else (gen "C_retrieve(lf[" index "])")] ) ) )
 
 	  ((##core#setglobal)
 	   (let ([index (first params)]
@@ -240,7 +246,7 @@
 		  (customizable (and call-id (fourth params)))
 		  (empty-closure (and customizable (zero? (lambda-literal-closure-size (find-lambda call-id)))))
 		  (fn (car subs)) )
-	     (when (and emit-trace-info name) (gen #t "C_trace(\"" name "\");"))
+	     (when (and emit-trace-info name) (gen #t "C_trace(\"" (slashify name) "\");"))
 	     (cond ((eq? '##core#proc (node-class fn))
 		    (gen #t (first (node-parameters fn)) #\( nf ",0,") 
 		    (expr-args args i)
@@ -288,7 +294,7 @@
 		  [nf (add1 n)]
 		  [tailcall (first params)]
 		  [call-id (second params)] 
-		  [emtpy-closure (zero? (lambda-literal-closure-size ll))] )
+		  [empty-closure (zero? (lambda-literal-closure-size ll))] )
 	     (cond (tailcall
 		    (let* ([temps (lambda-literal-temporaries ll)]
 			   [ts (iota n (+ temps nf) 1)] )
@@ -318,21 +324,19 @@
 		  (allocating (not (zero? demand)))
 		  (empty-closure (zero? (lambda-literal-closure-size (find-lambda call-id))))
 		  (fn (car subs)) )
-	     (when allocating (gen "(a+=" demand #\,))
 	     (gen call-id #\()
 	     (when allocating 
-	       (gen "a-" demand)
+	       (gen "C_a_i(&a," demand #\))
 	       (when (or (not empty-closure) (pair? args)) (gen #\,)) )
 	     (unless empty-closure
 	       (expr fn i)
 	       (when (pair? args) (gen #\,)) )
 	     (when (pair? args) (expr-args args i))
-	     (gen #\))
-	     (when allocating (gen #\))) ) )
+	     (gen #\)) ) )
 
 	  ((##core#callunit)
 	   ;; The code generated here does not use the extra temporary needed for standard calls, so we have
-	   ;;  one unused varable (but, who cares!):
+	   ;;  one unused varable:
 	   (let* ((n (length subs))
 		  (nf (+ n 1)) )
 	     (gen #t "C_" (first params) "_toplevel(" nf ",C_SCHEME_UNDEFINED,")
@@ -361,9 +365,24 @@
 	   (gen (foreign-result-conversion (second params) "a") (first params) #\)) )
 
 	  ((##core#inline_update)
-	   (gen #\( (first params) #\= (foreign-argument-conversion (second params)))
-	   (expr (first subs) i)
-	   (gen "),C_SCHEME_UNDEFINED)") )
+	   (let ([t (second params)])
+	     (gen #\( (first params) "=(" (foreign-type-declaration t "") #\) (foreign-argument-conversion t)) 
+	     (expr (first subs) i)
+	     (gen "),C_SCHEME_UNDEFINED)") ) )
+
+	  ((##core#inline_loc_ref)
+	   (let ([t (first params)])
+	     (gen (foreign-result-conversion t "a") "*((" (foreign-type-declaration t "") "*)C_data_pointer(")
+	     (expr (first subs) i)
+	     (gen ")))") ) )
+
+	  ((##core#inline_loc_update)
+	   (let ([t (first params)])
+	     (gen "((*(" (foreign-type-declaration t "") "*)C_data_pointer(")
+	     (expr (first subs) i)
+	     (gen "))=" (foreign-argument-conversion t))
+	     (expr (second subs) i) 
+	     (gen "),C_SCHEME_UNDEFINED)") ) )
 
 	  ((##core#switch)
 	   (gen #t "switch(")
@@ -401,16 +420,17 @@
     (expr node temps) )
     
   (define (header)
-    (gen "/* Generated from " source-file " by " product ", " build-information #t
-	 "   Command line: ")
+    (gen "/* Generated from " source-file " by " product #t
+	 "   " build-information #t
+	 "   command line: ")
     (gen-list compiler-arguments)
     (gen #t)
-    (cond [unit-name (gen "   Unit: " unit-name)]
+    (cond [unit-name (gen "   unit: " unit-name)]
 	  [else 
-	   (gen "   Used units: ")
+	   (gen "   used units: ")
 	   (gen-list used-units) ] )
     (unless unit-name
-      (gen #t "   default installation home: " default-installation-home #t)
+      (gen #t "   default installation home: " (or default-installation-home "not specified") #t)
       (gen "   default heap size: " default-default-target-heap-size #t)
       (gen "   default nursery (stack) size: " default-default-target-stack-size) )
     (gen #t "*/" #t #t "#include \"" target-include-file "\"")
@@ -419,13 +439,13 @@
       (for-each (lambda (decl) (gen #t decl)) foreign-declarations) ) )
   
   (define (trailer)
-    (gen #t "/* END OF FILE */" #t) )
+    (gen #t "/* end of file */" #t) )
   
   (define (declarations)
     (let ((n (length literals)))
       (gen #t)
       (for-each 
-       (lambda (uu) (gen #t "C_extern void C_" uu "_toplevel(int c,C_word d,C_word k) C_noret;"))
+       (lambda (uu) (gen #t "C_externimport void C_" uu "_toplevel(C_word c,C_word d,C_word k) C_noret;"))
        used-units)
       (unless (zero? n) (gen #t #t "static C_word lf[" n "];")) ) )
   
@@ -456,10 +476,10 @@
 		  (when customizable (gen "C_fcall "))
 		  (gen id) ]
 		 [else
-		  (gen "C_extern void ")
+		  (gen "C_externexport void ")
 		  (gen "C_" (if unit-name (string-append unit-name "_toplevel") "toplevel")) ] )
 	   (gen #\()
-	   (unless customizable (gen "int c,"))
+	   (unless customizable (gen "C_word c,"))
 	   (when (and direct (not (zero? allocated)))
 	     (gen "C_word *a")
 	     (when (pair? varlist) (gen #\,)) )
@@ -479,7 +499,7 @@
        lambdas) 
       (for-each
        (lambda (s)
-	 (gen #t "typedef void (*C_proc" s ")(int")
+	 (gen #t "typedef void (*C_proc" s ")(C_word")
 	 (for-each gen (make-list s ",C_word"))
 	 (gen ") C_noret;") )
        large-signatures) ) )
@@ -499,18 +519,20 @@
       (define (emitter vflag)
 	(lambda (n)
 	  (gen #t #t "static void C_fcall tr" n #\r)
-	  (if vflag (gen #\v))
+	  (when vflag (gen #\v))
 	  (gen "(C_proc" n " k) C_regparm C_noret;")
 	  (gen #t "static void C_fcall tr" n #\r)
-	  (if vflag (gen #\v))
+	  (when vflag (gen #\v))
 	  (gen "(C_proc" n " k){"
 	       #t "int n;"
 	       #t "C_word *a,t" n #\;)
 	  (restore n)
 	  (gen #t "n=C_rest_count(0);")
-	  (gen #t "a=C_alloc(n*3);")
+	  (if vflag
+	      (gen #t "a=C_alloc(n+1);")
+	      (gen #t "a=C_alloc(n*3);") )
 	  (gen #t #\t n "=C_restore_rest")
-	  (if vflag (gen "_vector"))
+	  (when vflag (gen "_vector"))
 	  (gen "(a,n);")
 	  (gen #t "(k)(")
 	  (apply gen (intersperse (make-argument-list (+ n 1) "t") #\,))
@@ -557,7 +579,7 @@
     (do ([i 0 (+ i 1)]
 	 [lits literals (cdr lits)] )
 	((null? lits))
-      (gen-lit (car lits) (sprintf "lf[~s]" i)) ) )
+      (gen-lit (car lits) (sprintf "lf[~s]" i) #t) ) )
 
   (define (bad-literal lit)
     (bomb "type of literal not supported" lit) )
@@ -568,7 +590,7 @@
 	  [(number? lit) words-per-flonum]
 	  [(symbol? lit) (+ 9 (literal-size (##sys#slot lit 1)))]
 	  [(pair? lit) (+ 3 (literal-size (car lit)) (literal-size (cdr lit)))]
-	  [(vector? lit) (+ 2 (vector-length lit) (reduce + 0 (map literal-size (vector->list lit))))]
+	  [(vector? lit) (+ 1 (vector-length lit) (reduce + 0 (map literal-size (vector->list lit))))]
 	  [(block-variable-literal? lit) 0]
 	  [(##sys#immediate? lit) (bad-literal lit)]
 	  [(##sys#bytevector? lit)
@@ -583,18 +605,20 @@
 		   (loop (add1 i) (+ s (literal-size (##sys#slot lit i)))) ) ) ) ]
 	  [else (bad-literal lit)] ) )
 
-  (define (gen-lit lit to)
+  (define (gen-lit lit to lf)
     (cond ((fixnum? lit)
 	   (if (eq? 'flonum number-type)
 	       (gen #t to "=C_flonum(C_heaptop," lit ");")
 	       (gen #t to "=C_fix(" lit ");") ) )
-	  ((or (eq? lit (void)) (block-variable-literal? lit))
+	  ((block-variable-literal? lit))
+	  ((or (eq? lit (void)))
 	   (gen #t to "=C_SCHEME_UNDEFINED;") )
 	  ((number? lit)
-	   (cond ((eq? 'fixnum number-type)
-		  (warning "coerced inexact literal number '~s' to fixnum" lit)
-		  (gen #t to "=C_fix(" lit ");") )
-		 (else (gen #t to "=C_flonum(C_heaptop," lit ");")) ) )
+	   (cond [(eq? 'fixnum number-type)
+		  (let ([flit (##sys#flo2fix lit)])
+		    (warning "coerced inexact literal number ~S to fixnum ~S" lit flit)
+		    (gen #t to "=C_fix(" flit ");") ) ]
+		 [else (gen #t to "=C_flonum(C_heaptop," lit ");")] ) )
           ((boolean? lit) 
            (gen #t to #\= (if lit "C_SCHEME_TRUE" "C_SCHEME_FALSE") #\;) )
           ((char? lit)
@@ -607,29 +631,32 @@
 		  (do ((len 0 (add1 len))
 		       (lst lit (cdr lst)) )
 		      ((null? lst)
-		       (gen #t to "=C_a_i_list(C_heaptop," len)
+		       (gen #t to "=C_h_list(" len)
 		       (do ((k (sub1 len) (sub1 k)))
 			   ((< k 0) (gen ");" #t "C_drop(" len ");"))
 			 (gen ",C_pick(" k #\)) ) )
-		    (gen-lit (car lst) "tmp")
+		    (gen-lit (car lst) "tmp" #f)
 		    (gen #t "C_save(tmp);") ) )
 		 (else
-		  (gen-lit (car lit) "tmp")
+		  (gen-lit (car lit) "tmp" #f)
 		  (gen #t "C_save(tmp);")
-		  (gen-lit (cdr lit) "tmp")
-		  (gen #t to "=C_pair(C_heaptop,C_restore,tmp);") ) ) )
-	  ((vector? lit) (gen-vector-like-lit to lit "C_vector"))
+		  (gen-lit (cdr lit) "tmp" #f)
+		  (gen #t to "=C_h_pair(C_restore,tmp);") ) ) )
+	  ((vector? lit) (gen-vector-like-lit to lit "C_h_vector"))
           ((symbol? lit)
 	   (let* ([str (##sys#slot lit 1)]
 		  [cstr (c-ify-string str)]
 		  [len (##sys#size str)] )
-	     (gen #t to "=C_intern(C_heaptop," len #\, cstr ");") ) )
+	     (gen #t to "=")
+	     (if lf
+		 (gen "C_h_intern(&" to #\, len #\, cstr ");")
+		 (gen "C_intern(C_heaptop," len #\, cstr ");") ) ) )
 	  ((##sys#immediate? lit) (bad-literal lit))
 	  ((##sys#bytevector? lit)
 	   (if (##sys#permanent? lit)
 	       (gen-string-like-lit to lit "C_pbytevector" #f)
 	       (gen-string-like-lit to lit "C_bytevector" #t) ) )
-	  ((##sys#generic-structure? lit) (gen-vector-like-lit to lit "C_structure"))
+	  ((##sys#generic-structure? lit) (gen-vector-like-lit to lit "C_h_structure"))
           (else (bad-literal lit)) ) )
 
   (define (gen-string-like-lit to lit conser top)
@@ -658,11 +685,11 @@
       (do ([j 0 (+ j 1)]
 	   [n len (- n 1)] )
 	  ((zero? n)
-	   (gen #t to #\= conser "(C_heaptop," len)
+	   (gen #t to #\= conser #\( len)
 	   (do ([j (- len 1) (- j 1)])
 	       ((< j 0) (gen ");" #t "C_drop(" len ");"))
 	     (gen ",C_pick(" j #\)) ) )
-	(gen-lit (##sys#slot lit j) "tmp")
+	(gen-lit (##sys#slot lit j) "tmp" #f)
 	(gen #t "C_save(tmp);") ) ) )
 
   (define (procedures)
@@ -670,13 +697,16 @@
      (lambda (ll)
        (let* ([n (lambda-literal-argument-count ll)]
 	      [id (lambda-literal-id ll)]
+	      [rname (real-name id db)]
 	      [demand (lambda-literal-allocated ll)]
 	      [rest (lambda-literal-rest-argument ll)]
 	      [customizable (lambda-literal-customizable ll)]
 	      [empty-closure (and customizable (zero? (lambda-literal-closure-size ll)))]
 	      [nec (- n (if empty-closure 1 0))]
 	      [vlist0 (make-variable-list n "t")]
+	      [alist0 (make-argument-list n "t")]
 	      [varlist (intersperse (if empty-closure (cdr vlist0) vlist0) #\,)]
+	      [arglist (intersperse (if empty-closure (cdr alist0) alist0) #\,)]
 	      [external (lambda-literal-external ll)]
 	      [looping (lambda-literal-looping ll)]
 	      [direct (lambda-literal-direct ll)]
@@ -687,18 +717,21 @@
 			   "toplevel") ] )
 	 (when empty-closure (debugging 'o "dropping unused closure argument" id))
 	 (gen #t #t)
+	 (gen "/* " (cleanup rname) " */" #t)
 	 (cond [(not (eq? 'toplevel id)) 
 		(gen "static " (if direct "C_word " "void "))
 		(when customizable (gen "C_fcall "))
 		(gen id) ]
 	       [else
-		(gen "static int toplevel_initialized=0;"
-		     #t "static void C_fcall toplevel_trampoline(void *dummy) C_regparm C_noret;"
+		(gen "static int toplevel_initialized=0;")
+		(unless unit-name
+		  (gen #t "C_main_entry_point") )
+		(gen #t "static void C_fcall toplevel_trampoline(void *dummy) C_regparm C_noret;"
 		     #t "static void C_fcall toplevel_trampoline(void *dummy){"
 		     #t "C_" topname "(2,C_SCHEME_UNDEFINED,C_restore);}"
 		     #t #t "void C_" topname) ] )
 	 (gen #\()
-	 (unless customizable (gen "int c,"))
+	 (unless customizable (gen "C_word c,"))
 	 (when (and direct (not (zero? demand))) 
 	   (gen "C_word *a")
 	   (when (pair? varlist) (gen #\,)) )
@@ -719,17 +752,30 @@
 		  (gen #t "C_word *a;"
 		       #t "if(toplevel_initialized) C_kontinue(t1,C_SCHEME_UNDEFINED);"
 		       #t "else C_toplevel_entry(C_text(\"" topname "\"));")
-		  (when (and target-heap-size (not unit-name))
-		    (gen #t "C_set_or_change_heap_size(" target-heap-size ",1);") )
-		  (when (and target-stack-size (not unit-name))
-		    (gen #t "C_resize_stack(" target-stack-size ");") )
-		  (gen #t "if(!C_demand(" demand ")){"
+		  (unless unit-name
+		    (cond [target-initial-heap-size
+			   (gen #t "C_set_or_change_heap_size(" target-initial-heap-size ",1);") ]
+			  [target-heap-size
+			   (gen #t "C_set_or_change_heap_size(" target-heap-size ",1);"
+				#t "C_heap_size_is_fixed=1;") ] )
+		    (when target-heap-growth
+		      (gen #t "C_heap_growth=" target-heap-growth #\;) )
+		    (when target-heap-shrinkage
+		      (gen #t "C_heap_shrinkage=" target-heap-shrinkage #\;) )
+		    (when target-stack-size
+		      (gen #t "C_resize_stack(" target-stack-size ");") ) )
+		  (gen #t "C_check_nursery_minimum(" demand ");"
+                       #t "if(!C_demand(" demand ")){"
 		       #t "C_save(t1);"
-		       #t "C_reclaim(toplevel_trampoline,NULL);}"
+		       #t "C_reclaim((void*)toplevel_trampoline,NULL);}"
 		       #t "toplevel_initialized=1;")
-		  (gen #t "if(!C_demand_2(" ldemand ")) C_bad_memory_2();")
+		  (gen #t "if(!C_demand_2(" ldemand ")){"
+		       #t "C_save(t1);"
+		       #t "C_rereclaim(" (- ldemand) "*sizeof(C_word));"
+		       #t "t1=C_restore;}")
 		  (gen #t "a=C_alloc(" demand ");")
 		  (when (not (zero? llen))
+		    (gen #t "C_initialize_lf(lf," llen ");")
 		    (literal-frame)
 		    (gen #t "C_register_lf(lf," llen ");") ) ) ]
 	       [rest
@@ -743,7 +789,7 @@
 		(when (and (not unsafe) (not no-argc-checks) (> n 2))
 		  (gen #t "if(c<" n ") C_bad_min_argc(c," n ");") )
 		(when insert-timer-checks (gen #t "C_check_for_interrupt;"))
-		(gen #t "if(!C_stack_probe(&a-" (+ n demand) ")){") ] ; (rough approximation)
+		(gen #t "if(!C_demand(c*C_SIZEOF_PAIR+" demand ")){") ]
 	       [else
 		(cond [(and (not direct) (> demand 0))
 		       (if looping
@@ -767,16 +813,22 @@
 	 (when (and (not (eq? 'toplevel id))
 		    (not direct)
 		    (or rest external (> demand 0)) )
+	   #|
 	   (cond [(> nec 1)
 		  (gen #t "C_adjust_stack(" nec ");")
 		  (do ([i (if empty-closure 1 0) (+ i 1)])
 		      ((>= i n))
 		    (gen #t "C_rescue(t" i #\, (- n i 1) ");") ) ]
 		 [(= nec 1) (gen #t "C_save(" (if empty-closure "t1" "t0") ");")] )
+	   |#
 	   (cond [rest
-		  (gen #t "C_reclaim(tr" n #\r)
+		  (gen #t (if (> nec 0) "C_save_and_reclaim" "C_reclaim") "((void*)tr" n #\r)
 		  (when (eq? rest-mode 'vector) (gen #\v))
-		  (gen #\, id "r);}"
+		  (gen ",(void*)" id "r")
+		  (when (> nec 0)
+		    (gen #\, nec #\,)
+		    (apply gen arglist) )
+		  (gen ");}"
 		       #t "else{"
 		       #t "a=C_alloc((c-" n ")*3);")
 		  (case rest-mode
@@ -796,10 +848,13 @@
 		    (gen #t "C_word t" i #\;) )
 		  (when (> demand 0) (gen #t "C_word *a=C_alloc(" demand ");")) ]
 		 [else 
-		  (gen #t "C_reclaim(tr")
+		  (gen #t (if (> nec 0) "C_save_and_reclaim" "C_reclaim") "((void*)tr")
 		  (if customizable 
 		      (gen id ",NULL")
-		      (gen n #\, id) )
+		      (gen n ",(void*)" id) )
+		  (when (> nec 0)
+		    (gen #\, nec #\,)
+		    (apply gen arglist) )
 		  (gen ");}") ] ) )
 	 (expression
 	  (lambda-literal-body ll)
@@ -816,14 +871,33 @@
   (declarations)
   (generate-external-variables external-variables)
   (generate-foreign-callback-stub-prototypes foreign-callback-stubs)
-  (generate-foreign-stubs foreign-lambda-stubs)
+  (generate-foreign-stubs foreign-lambda-stubs db)
   (prototypes)
-  (generate-foreign-callback-stubs foreign-callback-stubs)
+  (generate-foreign-callback-stubs foreign-callback-stubs db)
   (trampolines)
   (procedures)
   (trailer) )
 
 
+;;; Create name that is safe for C comments:
+
+(define (cleanup s)
+  (let ([s2 #f] 
+	[len (string-length s)] )
+    (let loop ([i 0])
+      (if (>= i len)
+	  (or s2 s)
+	  (let ([c (string-ref s i)])
+	    (if (or (char<? c #\space)
+		    (char>? c #\~)
+		    (and (char=? c #\*) (< i (sub1 len)) (char=? #\/ (string-ref s (add1 i)))) )
+		(begin
+		  (unless s2 (set! s2 (string-copy s)))
+		  (string-set! s2 i #\~) )
+		(when s2 (string-set! s2 i c)) ) 
+	    (loop (add1 i)) ) ) ) ) )
+
+
 ;;; Create list of variables/parameters, interspersed with a special token:
 
 (define (make-variable-list n prefix)
@@ -842,10 +916,9 @@
 (define (generate-external-variables vars)
   (gen #t)
   (for-each
-   (lambda (xvar)
-     (let ([name (car xvar)]
-	   [type (cdr xvar)] )
-       (gen #t (foreign-type-declaration type name) #\;) ) )
+   (match-lambda 
+     [#(name type exported)
+      (gen #t (if exported "" "static ") (foreign-type-declaration type name) #\;) ] )
    vars) )
 
 
@@ -859,10 +932,11 @@
      (gen #\;) )
    stubs) )
 
-(define (generate-foreign-stubs stubs)
+(define (generate-foreign-stubs stubs db)
   (for-each
    (lambda (stub)
      (let* ([id (foreign-stub-id stub)]
+	    [rname (real-name2 id db)]
 	    [types (foreign-stub-argument-types stub)]
 	    [n (length types)]
 	    [varlist (intersperse (cons "C_word C_buf" (make-variable-list n "C_a")) #\,)]
@@ -873,13 +947,15 @@
 	    [rconv (foreign-result-conversion rtype "C_a")] 
 	    [callback (foreign-stub-callback stub)] )
        (gen #t)
+       (when rname
+	 (gen #t "/* from " (cleanup rname) " */") )
        (when body (gen #t "#define return(x) C_cblock C_r = (" rconv "(x))); goto C_return; C_cblockend"))
        (if callback
-	   (gen #t "static void " id "(int c,C_word self,C_word k,")
+	   (gen #t "static void " id "(C_word C_c,C_word C_self,C_word C_k,")
 	   (gen #t "static C_word C_fcall " id #\() )
        (apply gen varlist)
        (if callback
-	   (gen ") C_noret;" #t "static void " id "(int c,C_word self,C_word k,")
+	   (gen ") C_noret;" #t "static void " id "(C_word C_c,C_word C_self,C_word C_k,")
 	   (gen ") C_regparm;" #t "static C_word C_fcall " id #\() )
        (apply gen varlist)
        (gen "){")
@@ -890,17 +966,18 @@
 	       (foreign-type-declaration 
 		type
 		(if name (symbol->string name) (sprintf "t~a" index)) )
-	       #\= (foreign-argument-conversion type) "C_a" index ");") )
+	       "=(" (foreign-type-declaration type "") #\)
+	       (foreign-argument-conversion type) "C_a" index ");") )
 	types (iota n) names)
-       (when callback (gen #t "int C_dummy=C_save_callback_continuation(&C_a,k);"))
+       (when callback (gen #t "int C_dummy=C_save_callback_continuation(&C_a,C_k);"))
        (cond [body
 	      (gen #t body
 		   #t "C_return:")
+	      (gen #t "#undef return" #t)
 	      (if callback
-		  (gen #t "k=C_restore_callback_continuation();"
-		       #t "C_kontinue(k,C_r);")
-		  (gen #t "return C_r;") )
-	      (gen #t "#undef return" #t) ]
+		  (gen #t "C_k=C_restore_callback_continuation();"
+		       #t "C_kontinue(C_k,C_r);")
+		  (gen #t "return C_r;") ) ]
 	     [else
 	      (if (not (eq? rtype 'void))
 		  (gen #t "C_r=" rconv)
@@ -910,16 +987,17 @@
 	      (unless (eq? rtype 'void) (gen #\)))
 	      (gen ");")
 	      (if callback
-		  (gen #t "k=C_restore_callback_continuation();"
-		       #t "C_kontinue(k,C_r);")
+		  (gen #t "C_k=C_restore_callback_continuation();"
+		       #t "C_kontinue(C_k,C_r);")
 		  (gen #t "return C_r;") ) ] )
        (gen #\}) ) )
    stubs) )
 
-(define (generate-foreign-callback-stubs stubs)
+(define (generate-foreign-callback-stubs stubs db)
   (for-each
    (lambda (stub)
      (let* ([id (foreign-callback-stub-id stub)]
+	    [rname (real-name2 id db)]
 	    [rtype (foreign-callback-stub-return-type stub)]
 	    [argtypes (foreign-callback-stub-argument-types stub)]
 	    [n (length argtypes)]
@@ -928,22 +1006,27 @@
        (define (compute-size type var ns)
 	 (case type
 	   [(char int short bool void unsigned-short scheme-object unsigned-char unsigned-int) ns]
-	   [(float double c-pointer unsigned-integer long integer unsigned-long)
+	   [(float double c-pointer unsigned-integer long integer unsigned-long nonnull-c-pointer)
 	    (string-append ns "+3") ]
-	   [(c-string)
+	   [(c-string c-string*)
 	    (string-append ns "+2+(" var "==NULL?1:C_bytestowords(C_strlen(" var ")))") ]
+	   [(nonnull-c-string nonnull-c-string*)
+	    (string-append ns "+2+C_bytestowords(C_strlen(" var "))") ]
 	   [else
 	    (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type)) 
 		   => (lambda (t)
 			(compute-size (if (vector? t) (vector-ref t 0) t) var ns) ) ]
 		  [(pair? type)
 		   (case (car type)
-		     [(pointer function) (string-append ns "+3")]
+		     [(ref pointer c-pointer nonnull-pointer function) (string-append ns "+3")]
+		     [(const) (compute-size (cadr type) var ns)]
 		     [else ns] ) ]
 		  [else ns] ) ] ) )
 
        (let ([sizestr (fold compute-size "0" argtypes vlist)])
 	 (gen #t)
+	 (when rname
+	   (gen #t "/* from " (cleanup rname) " */") )
 	 (generate-foreign-callback-header "" stub)
 	 (gen #\{ #t "C_word x, *a=C_alloc(" sizestr ");")
 	 (for-each
@@ -990,8 +1073,16 @@
       [(float) (str "float")]
       [(double) (str "double")]
       [(unsigned-integer) (str "unsigned int")]
-      [(pointer c-pointer) (str "void *")]
-      [(c-string) (str "char *")]
+      [(pointer nonnull-pointer c-pointer nonnull-c-pointer) (str "void *")]
+      [(byte-vector nonnull-byte-vector u8vector nonnull-u8vector) (str "unsigned char *")]
+      [(u16vector nonnull-u16vector) (str "unsigned short *")]
+      [(s8vector nonnull-s8vector) (str "char *")]
+      [(u32vector nonnull-u32vector) (str "unsigned int *")]
+      [(s16vector nonnull-s16vector) (str "short *")]
+      [(s32vector nonnull-s32vector) (str "int *")]
+      [(f32vector nonnull-f32vector) (str "float *")]
+      [(f64vector nonnull-f64vector) (str "double *")]
+      [(nonnull-c-string c-string nonnull-c-string* c-string*) (str "char *")]
       [(void) (str "void")]
       [else
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
@@ -1000,9 +1091,21 @@
 	     [(string? type) (str type)]
 	     [(pair? type)
 	      (match type
-		[`(pointer ,ptype) (string-append (foreign-type-declaration ptype (string-append "*" target)))]
+		[((or 'pointer 'c-pointer 'nonnull-pointer 'nonnull-c-pointer) ptype)
+		 (foreign-type-declaration ptype (string-append "*" target)) ]
+		[('ref rtype)
+		 (foreign-type-declaration rtype (string-append "&" target)) ]
+		[`(template ,t0 ,ts ...)
+		 (str
+		  (string-append 
+		   (foreign-type-declaration t0 "")
+		   "<"
+		   (string-intersperse (map (cut foreign-type-declaration <> "") ts) ",")
+		   "> ") ) ]
+		[`(const ,t) (string-append "const " (foreign-type-declaration t target))]
 		[`(struct ,sname) (string-append "struct " sname " " target)]
-		[`(union ,uname) (string-append "union " sname " " target)]
+		[`(union ,uname) (string-append "union " uname " " target)]
+		[`(enum ,ename) (string-append "enum " ename " " target)]
 		[`(function ,rtype ,argtypes . ,callconv)
 		 (string-append
 		  (foreign-type-declaration rtype "")
@@ -1012,7 +1115,7 @@
 			cc)
 		      "")
 		  " (*" target ")("
-		  (string-concatenate
+		  (string-intersperse
 		   (map (lambda (at)
 			  (if (eq? '... at) 
 			      "..."
@@ -1031,15 +1134,36 @@
       ((int unsigned-int) "C_unfix(")
       ((short) "C_unfix(")
       ((unsigned-short) "(unsigned short)C_unfix(")
-      ((unsigned-long) "(unsigned long)C_num_to_long(")
+      ((unsigned-long) "C_num_to_unsigned_long(")
       ((float) "C_c_double(")
       ((double) "C_c_double(")
       ((integer) "C_num_to_int(")
       ((long) "C_num_to_long(")
       ((unsigned-integer) "C_num_to_unsigned_int(")
       ((pointer) "C_data_pointer_or_null(")
+      ((nonnull-pointer) "C_data_pointer(")
       ((c-pointer) "C_c_pointer_or_null(")
-      ((c-string) "C_string_or_null(")
+      ((nonnull-c-pointer) "C_c_pointer_nn(")
+      ((byte-vector) "C_c_bytevector_or_null(")
+      ((nonnull-byte-vector) "C_c_bytevector(")
+      ((u8vector) "C_c_u8vector_or_null(")
+      ((nonnull-u8vector) "C_c_u8vector(")
+      ((u16vector) "C_c_u16vector_or_null(")
+      ((nonnull-u16vector) "C_c_u16vector(")
+      ((u32vector) "C_c_u32vector_or_null(")
+      ((nonnull-u32vector) "C_c_u32vector(")
+      ((s8vector) "C_c_s8vector_or_null(")
+      ((nonnull-s8vector) "C_c_s8vector(")
+      ((s16vector) "C_c_s16vector_or_null(")
+      ((nonnull-s16vector) "C_c_s16vector(")
+      ((s32vector) "C_c_s32vector_or_null(")
+      ((nonnull-s32vector) "C_c_s32vector(")
+      ((f32vector) "C_c_f32vector_or_null(")
+      ((nonnull-f32vector) "C_c_f32vector(")
+      ((f64vector) "C_c_f64vector_or_null(")
+      ((nonnull-f64vector) "C_c_f64vector(")
+      ((c-string c-string*) "C_string_or_null(")
+      ((nonnull-c-string nonnull-c-string*) "C_c_string(")
       ((bool) "C_truep(")
       (else
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
@@ -1047,8 +1171,13 @@
 		   (foreign-argument-conversion (if (vector? t) (vector-ref t 0) t)) ) ]
 	     [(pair? type)
 	      (match type
-		[`(pointer ,ptype) "C_c_pointer_or_null("]
+		[((or 'pointer 'c-pointer) ptype) "C_c_pointer_or_null("]
+		[((or 'nonnull-pointer 'nonnull-c-pointer) ptype) "C_c_pointer_nn("]
 		[`(function ,rtype ,@argtypes) "C_c_pointer_or_null("]
+		[`(const ,ctype) (foreign-argument-conversion ctype)]
+		[`(enum ,etype) "C_num_to_int("]
+		[`(ref ,rtype)
+		 (string-append "*(" (foreign-type-declaration rtype "*") ")C_c_pointer_nn(") ]
 		[else (err)] ) ]
 	     [else (err)] ) ) ) ) )
 	    
@@ -1061,9 +1190,13 @@
       ((short) "C_fix((short)")
       ((unsigned-short) "C_fix(0xffff&(C_word)")
       ((float double) (sprintf "C_flonum(&~a," dest))
-      ((c-string c-pointer) (sprintf "C_mpointer(&~a,(void*)" dest))
-      ((integer unsigned-integer) (sprintf "C_int_to_num(&~a,(C_word)" dest))
-      ((long unsigned-long) (sprintf "C_long_to_num(&~a,(C_word)" dest))
+      ((nonnull-c-string c-string nonnull-c-pointer c-string* nonnull-c-string*) 
+       (sprintf "C_mpointer(&~a,(void*)" dest) )
+      ((c-pointer) (sprintf "C_mpointer_or_false(&~a,(void*)" dest))
+      ((integer) (sprintf "C_int_to_num(&~a," dest))
+      ((unsigned-integer) (sprintf "C_unsigned_int_to_num(&~a," dest))
+      ((long) (sprintf "C_long_to_num(&~a," dest))
+      ((unsigned-long) (sprintf "C_unsigned_long_to_num(&~a," dest))
       ((bool) "C_mk_bool(")
       ((void scheme-object) "((C_word)")
       (else
@@ -1072,7 +1205,15 @@
 		   (foreign-result-conversion (if (vector? x) (vector-ref x 0) x) dest)) ]
 	     [(pair? type)
 	      (match type
-		[`(pointer ,ptype) (sprintf "C_mpointer(&~a,(void*)" dest)]
+		[((or 'nonnull-pointer 'nonnull-c-pointer) ptype) 
+		 (sprintf "C_mpointer(&~A,(void*)" dest) ]
+		[('ref rtype) 
+		 (sprintf "C_mpointer(&~A,(void*)&" dest) ]
+		[('const ctype) (foreign-result-conversion ctype dest)]
+		[((or 'pointer 'c-pointer) ptype) 
+		 (sprintf "C_mpointer_or_false(&~a,(void*)" dest) ]
 		[`(function ,rtype ,@argtypes) (sprintf "C_mpointer(&~a,(void*)" dest)]
+		[`(enum ,etype) (sprintf "C_int_to_num(&~a," dest)]
 		[else (err)] ) ]
 	     [else (err)] ) ) ) ) )
+
