@@ -1,6 +1,6 @@
 ;;;; c-platform.scm - Platform specific parameters and definitions
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -38,18 +38,18 @@
 
 #{compiler
   build-information compiler-arguments process-command-line
-  default-analysis-database-size default-standard-bindings default-extended-bindings side-effecting-standard-bindings
+  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  installation-home
-  foreign-type-table-size file-io-only nonwinding-call/cc
+  installation-home debugging
+  foreign-type-table-size
   unit-name insert-timer-checks used-units inlining
   foreign-declarations block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home 
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  dependency-list broken-constant-nodes inline-substitutions-enabled
+  broken-constant-nodes inline-substitutions-enabled
   direct-call-ids foreign-type-table first-analysis
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
@@ -65,10 +65,10 @@
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
   product copyright compiler-features default-declarations default-debugging-declarations units-used-by-default words-per-flonum
-  parameter-limit default-output-filename eq-inline-operator optimizable-rest-argument-operators
+  parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   target-include-file default-profiling-declarations
-  default-optimization-passes
+  default-optimization-passes internal-bindings
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
   foreign-argument-conversion foreign-result-conversion}
 
@@ -79,7 +79,7 @@
 ;;; Parameters:
 
 (define product "the CHICKEN compiler")
-(define copyright "(c)2000-2002 Felix L. Winkelmann")
+(define copyright "(c)2000-2003 Felix L. Winkelmann")
 
 (define default-optimization-passes 3)
 
@@ -91,11 +91,12 @@
      ##sys#standard-input ##sys#standard-output ##sys#standard-error)
     (bound-to-procedure
      ##sys#for-each ##sys#map ##sys#print
-     ##sys#setslot ##sys#dynamic-wind ##sys#call-with-values
+     ##sys#setslot ##sys#dynamic-wind ##sys#call-with-values ##sys#match-error
      ##sys#start-timer ##sys#stop-timer ##sys#gcd ##sys#lcm ##sys#make-promise ##sys#structure? ##sys#slot 
-     ##sys#allocate-vector ##sys#list->vector
+     ##sys#allocate-vector ##sys#list->vector ##sys#record?
      ##sys#list ##sys#cons ##sys#append ##sys#vector ##sys#foreign-char-argument ##sys#foreign-fixnum-argument
-     ##sys#foreign-flonum-argument ##sys#error
+     ##sys#foreign-flonum-argument ##sys#error ##sys#peek-c-string ##sys#peek-nonnull-c-string 
+     ##sys#peek-and-free-c-string ##sys#peek-and-free-nonnull-c-string
      ##sys#foreign-block-argument ##sys#foreign-string-argument ##sys#foreign-pointer-argument ##sys#foreign-integer-argument
      ##sys#call-with-current-continuation) ) )
 
@@ -116,7 +117,6 @@
 (define words-per-flonum 4)
 (define parameter-limit 1024)
 (define small-parameter-limit 128)
-(define default-output-filename "out.c")
 
 (define eq-inline-operator "C_eqp")
 (define optimizable-rest-argument-operators '(car cadr caddr cadddr length pair? null? list-ref))
@@ -127,14 +127,16 @@
 (define target-include-file "chicken.h")
 
 (define valid-compiler-options
-  '(help version verbose explicit-use quiet no-trace no-warnings unsafe optimize block inline assume-self-calls
-    check-syntax expand-only hygienic write-dependencies to-stdout usual-integrations case-sensitive emit-debug-info
-    fixnum-arithmetic disable-interrupts optimize-leaf-routines no-fancy-ports no-winding-callcc debug-calls
-    debug-loops srfi-7 strict strict-srfi-0) )
+  '(help version verbose explicit-use quiet no-trace no-warnings unsafe block assume-self-calls
+    check-syntax hygienic to-stdout usual-integrations case-insensitive emit-debug-info profile
+    fixnum-arithmetic disable-interrupts optimize-leaf-routines debug-calls
+    debug-loops srfi-7 strict strict-srfi-0 strict-reader lambda-lift compile-time-macros-only
+    hygienic-at-run-time analyze-only strict-letrec dynamic ffi) )
 
 (define valid-compiler-options-with-argument
-  '(debug output-file include-path database-size inline-limit inline-passes heap-size stack-size profile
-    dependency-output prelude postlude prologue epilogue nursery extend feature no-feature) )
+  '(debug output-file include-path heap-size stack-size unit uses keyword-style visit require-for-syntax
+    prelude postlude prologue epilogue nursery extend feature no-feature compress-literals
+    heap-growth heap-shrinkage heap-initial-size ffi-define ffi-include-path) )
 
 
 ;;; Standard and extended bindings:
@@ -156,24 +158,41 @@
     string-append string->list list->string vector? vector->list list->vector string read
     read-char substring string-fill! vector-fill! make-string make-vector open-input-file
     open-output-file call-with-input-file call-with-output-file close-input-port close-output-port
-    port? values call-with-values vector procedure? memq memv member assq assv assoc list-tail
+    values call-with-values vector procedure? memq memv member assq assv assoc list-tail
     list-ref abs char-ready? peek-char) )
 
 (define default-extended-bindings
-  '(##sys#slot gc bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 ##sys#setslot fx+ fx- fx* fx/ fxmod
-    fx= fx> fx< fx>= fx<= ##sys#call-with-current-continuation ##sys#size fixnum? fxneg fxmax fxmin ##sys#byte ##sys#setbyte
-    arithmetic-shift signum void ##sys#pointer? ##sys#generic-structure? ##sys#structure? ##sys#check-structure flush-output
+  '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod
+    fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity
+    arithmetic-shift signum void flush-output
+    not-pair? null-list? print print* error cpu-time proper-list?
+    u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector u32vector->byte-vector 
+    s32vector->byte-vector byte-vector-length block-ref block-set! number-of-slots
+    f32vector->byte-vector f64vector->byte-vector byte-vector-ref byte-vector-set!
+    first second third fourth make-record-instance
+    locative-ref locative-set! locative->object locative?
+    null-pointer? pointer->object) )
+
+(define internal-bindings
+  '(##sys#slot ##sys#setslot
+    ##sys#call-with-current-continuation ##sys#size ##sys#byte ##sys#setbyte
+    ##sys#pointer? ##sys#generic-structure? ##sys#structure? ##sys#check-structure
     ##sys#check-exact ##sys#check-number ##sys#check-list ##sys#check-pair ##sys#check-string ##sys#check-symbol 
-    ##sys#check-char ##sys#check-vector not-pair? null-list? print print* ##sys#check-byte-vector ##sys#list
-    ##sys#call-with-values ##sys#tty-port? ##sys#fits-in-int? ##sys#fits-in-unsigned-int? ##sys#flonum-in-fixnum-range? 
-    ##sys#fudge u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector u32vector->byte-vector 
-    s32vector->byte-vector byte-vector-length ##sys#make-structure block-ref block-set! number-of-slots ##sys#apply
-    f32vector->byte-vector f64vector->byte-vector ##sys#bytevector? byte-vector-ref byte-vector-set! ##sys#make-vector
-    ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? first second third fourth ##sys#values
-    ##sys#intern-symbol ##sys#make-symbol ##sys#null-pointer? null-pointer?) )
+    ##sys#check-char ##sys#check-vector ##sys#check-byte-vector ##sys#list
+    ##sys#call-with-values ##sys#fits-in-int? ##sys#fits-in-unsigned-int? ##sys#flonum-in-fixnum-range? 
+    ##sys#fudge ##sys#immediate? ##sys#direct-return
+    ##sys#make-structure ##sys#apply
+    ##sys#bytevector? ##sys#make-vector
+    ##sys#foreign-char-argument ##sys#foreign-fixnum-argument ##sys#foreign-flonum-argument
+    ##sys#foreign-block-argument ##sys#foreign-number-vector-argument
+    ##sys#foreign-string-argument ##sys#foreign-pointer-argument 
+    ##sys#foreign-integer-argument ##sys#foreign-unsigned-integer-argument
+    ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? ##sys#values
+    ##sys#intern-symbol ##sys#make-symbol ##sys#null-pointer? ##sys#peek-byte) )
 
 (define side-effecting-standard-bindings
   '(apply call-with-current-continuation set-car! set-cdr! write-char newline write display
+    peek-char char-ready?
     read read-char for-each map string-set! vector-set! string-fill! vector-fill! open-input-file
     open-output-file close-input-port close-output-port call-with-input-port call-with-output-port
     call-with-values eval) )
@@ -187,11 +206,11 @@
    side-effecting-standard-bindings non-foldable-standard-bindings) )
 
 (define non-foldable-extended-bindings
-  '(##sys#slot ##sys#setslot gc ##sys#call-with-current-continuation void ##sys#fudge flush-output print
+  '(##sys#slot ##sys#setslot ##sys#call-with-current-continuation ##sys#fudge flush-output print void
     u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector u32vector->byte-vector 
-    s32vector->byte-vector ##sys#make-structure print* ##sys#make-vector ##sys#apply
+    s32vector->byte-vector ##sys#make-structure print* ##sys#make-vector ##sys#apply ##sys#setislot
     f32vector->byte-vector f64vector->byte-vector ##sys#byte ##sys#setbyte byte-vector-ref byte-vector-set!
-    ##sys#intern-symbol ##sys#make-symbol) )
+    ##sys#intern-symbol ##sys#make-symbol make-record-instance error cpu-time) )
 
 (define foldable-extended-bindings
   (lset-difference
@@ -213,37 +232,9 @@
 
 ;;; Rewriting-definitions for this platform:
 
-(rewrite 
- '+ 8 
- (lambda (db classargs cont callargs)
-   ;; (+) -> 0
-   ;; (+ <x>) -> <x>
-   ;; (+ <x1> ...) -> (##core#inline "C_fixnum_plus" <x1> (##core#inline "C_fixnum_plus" ...)) [fixnum-mode]
-   ;; (+ <x1> ...) -> (##core#inline "C_u_fixnum_plus" <x1> (##core#inline "C_u_fixnum_plus" ...)) [fixnum-mode + unsafe]
-   ;; - Remove "0" from arguments.
-   (let ([callargs 
-	  (remove
-	   (lambda (x)
-	     (and (eq? 'quote (node-class x))
-		  (zero? (first (node-parameters x))) ) ) 
-	   callargs) ] )
-     (cond [(null? callargs) (make-node '##core#call '(#t) (list cont (qnode 0)))]
-	   [(null? (cdr callargs))
-	    (make-node '##core#call '(#t) (list cont (first callargs))) ]
-	   [(eq? number-type 'fixnum)
-	    (make-node
-	     '##core#call '(#t)
-	     (list
-	      cont
-	      (fold-inner
-	       (lambda (x y)
-		 (make-node '##core#inline 
-			    (if unsafe '("C_u_fixnum_plus") '("C_fixnum_plus"))
-			    (list x y) ) )
-	       callargs) ) ) ]
-	   [else #f] ) ) ) )
+(rewrite '+ 19 0 "C_fixnum_plus" "C_u_fixnum_plus" #f)
 
-(rewrite 
+(rewrite
  '* 8 
  (lambda (db classargs cont callargs)
    ;; (*) -> 1
@@ -476,6 +467,8 @@
 (rewrite 'call-with-values 13 "C_call_with_values" #t)
 (rewrite '##sys#call-with-values 13 "C_u_call_with_values" #f)
 (rewrite '##sys#call-with-values 13 "C_call_with_values" #t)
+(rewrite 'cpu-time 13 "C_cpu_time" #t)
+(rewrite 'locative-ref 13 "C_locative_ref" #t)
 
 (rewrite 'caar 2 1 "C_u_i_caar" #f #f)
 (rewrite 'cdar 2 1 "C_u_i_cdar" #f #f)
@@ -520,6 +513,7 @@
 (rewrite 'not 2 1 "C_i_not" #t #f)
 (rewrite 'char? 2 1 "C_charp" #t #f)
 (rewrite 'string? 2 1 "C_i_stringp" #t #f)
+(rewrite 'locative? 2 1 "C_i_locativep" #t #f)
 (rewrite 'symbol? 2 1 "C_i_symbolp" #t #f)
 (rewrite 'vector? 2 1 "C_i_vectorp" #t #f)
 (rewrite 'pair? 2 1 "C_i_pairp" #t "C_notvemptyp")
@@ -541,6 +535,7 @@
 (rewrite 'inexact? 2 1 "C_i_inexactp" #t #f)
 (rewrite 'inexact? 2 1 "C_u_i_inexactp" #f #f)
 (rewrite 'list? 2 1 "C_i_listp" #t #f)
+(rewrite 'proper-list? 2 1 "C_i_listp" #t #f)
 (rewrite 'eof-object? 2 1 "C_eofp" #t #f)
 (rewrite 'string-ref 2 2 "C_subchar" #f #f)
 (rewrite 'string-ref 2 2 "C_i_string_ref" #t #f)
@@ -555,7 +550,6 @@
 (rewrite 'char<=? 2 2 "C_fixnum_less_or_equal_p" #t #f)
 (rewrite '##sys#slot 2 2 "C_slot" #t #f)		; consider as safe, the primitive is unsafe anyway.
 (rewrite '##sys#size 2 1 "C_block_size" #t #f)
-(rewrite 'bitwise-xor 2 2 "C_fixnum_xor" #t #f)
 (rewrite 'bitwise-not 2 1 "C_fixnum_not" #t #f)
 (rewrite 'fx* 2 2 "C_fixnum_times" #t #f)
 (rewrite 'fx/ 2 2 "C_fixnum_divide" #f #f)
@@ -576,6 +570,7 @@
 (rewrite 'char-downcase 2 1 "C_u_i_char_downcase" #t #f)
 (rewrite 'list-tail 2 2 "C_i_list_tail" #t #f)
 (rewrite '##sys#structure? 2 2 "C_i_structurep" #t #f)
+(rewrite '##sys#record? 2 2 "C_recordp" #t #f)
 (rewrite '##sys#bytevector? 2 2 "C_bytevectorp" #t #f)
 (rewrite 'block-ref 2 2 "C_slot" #f #f)
 (rewrite 'number-of-slots 2 1 "C_block_size" #f #f)
@@ -589,9 +584,6 @@
 (rewrite 'assoc 2 2 "C_i_assoc" #t #f)
 (rewrite 'member 2 2 "C_i_member" #t #f)
 
-(rewrite 'current-input-port 3 '##sys#standard-input)
-(rewrite 'current-output-port 3 '##sys#standard-output)
-
 (rewrite 'set-car! 4 '##sys#setslot 0)
 (rewrite 'set-cdr! 4 '##sys#setslot 1)
 (rewrite 'set-car! 17 2 "C_i_set_car" "C_u_i_set_car")
@@ -620,6 +612,7 @@
 (rewrite 'vector-length 2 1 "C_i_vector_length" #t #f)
 (rewrite 'string-length 2 1 "C_i_string_length" #t #f)
 (rewrite 'inexact->exact 2 1 "C_i_inexact_to_exact" #t #f)
+
 (rewrite '##sys#check-exact 2 1 "C_i_check_exact" #t #f)
 (rewrite '##sys#check-number 2 1 "C_i_check_number" #t #f)
 (rewrite '##sys#check-list 2 1 "C_i_check_list" #t #f)
@@ -630,6 +623,16 @@
 (rewrite '##sys#check-vector 2 1 "C_i_check_vector" #t #f)
 (rewrite '##sys#check-structure 2 2 "C_i_check_structure" #t #f)
 (rewrite '##sys#check-char 2 1 "C_i_check_char" #t #f)
+(rewrite '##sys#check-exact 2 2 "C_i_check_exact_2" #t #f)
+(rewrite '##sys#check-number 2 2 "C_i_check_number_2" #t #f)
+(rewrite '##sys#check-list 2 2 "C_i_check_list_2" #t #f)
+(rewrite '##sys#check-pair 2 2 "C_i_check_pair_2" #t #f)
+(rewrite '##sys#check-symbol 2 2 "C_i_check_symbol_2" #t #f)
+(rewrite '##sys#check-string 2 2 "C_i_check_string_2" #t #f)
+(rewrite '##sys#check-byte-vector 2 2 "C_i_check_bytevector_2" #t #f)
+(rewrite '##sys#check-vector 2 2 "C_i_check_vector_2" #t #f)
+(rewrite '##sys#check-structure 2 3 "C_i_check_structure_2" #t #f)
+(rewrite '##sys#check-char 2 2 "C_i_check_char_2" #t #f)
 
 (rewrite '= 9 "C_eqp" "C_i_equalp" #t #t)
 (rewrite '> 9 "C_fixnum_greaterp" "C_flonum_greaterp" #t #f)
@@ -637,41 +640,22 @@
 (rewrite '>= 9 "C_fixnum_greater_or_equal_p" "C_flonum_greater_or_equal_p" #t #f)
 (rewrite '<= 9 "C_fixnum_less_or_equal_p" "C_flonum_less_or_equal_p" #t #f)
 
-(rewrite 'display 19 string? "C_display_string")
-(rewrite 'display 19 char? "C_display_char")
-(rewrite 'write-char 19 char? "C_display_char")
-
-(rewrite 'read-char 20 "C_read_char" #f)
-(rewrite 'peek-char 20 "C_peek_char" #f)
-(rewrite 'char-ready? 20 "C_char_ready?" #f)
-(rewrite 'flush-output 20 "C_flush_output" #t)
-
-(rewrite 
- 'newline 8
- (lambda (db classargs cont callargs)
-   (and-let* ([file-io-only]
-	      [port (case (length callargs)
-		      [(0) (varnode '##sys#standard-output)]
-		      [(1) (first callargs)]
-		      [else #f] ) ] )
-     (make-node '##core#call '(#t)
-		(list cont (make-node '##core#inline '("C_display_char") (list port (qnode #\newline)))) ) ) ) )
-
 (rewrite 'for-each 11 2 '##sys#for-each #t)
 (rewrite 'map 11 2 '##sys#map #t)
 (rewrite 'block-set! 11 3 '##sys#setslot #t)
+(rewrite 'make-record-instance 11 #f '##sys#make-structure #f)
 
 (rewrite 'vector-set! 11 3 '##sys#setslot #f)
 (rewrite 'vector-set! 2 3 "C_i_vector_set" #t #f)
 
 (rewrite 'gcd 12 '##sys#gcd #t 2)
 (rewrite 'lcm 12 '##sys#lcm #t 2)
+(rewrite 'identity 12 #f #t 1)
 
 (rewrite 'gcd 18 0)
 (rewrite 'lcm 18 1)
 (rewrite 'list 18 '())
 
-(rewrite 'gc 13 "C_gc" #t)
 (rewrite 'argv 13 "C_get_argv" #t)
 
 (rewrite '* 16 2 "C_a_i_times" #t 4)	; words-per-flonum
@@ -755,81 +739,70 @@
 (rewrite 'fx+ 17 2 "C_fixnum_plus" "C_u_fixnum_plus")
 (rewrite 'fx- 17 2 "C_fixnum_difference" "C_u_fixnum_difference")
 (rewrite 'fxneg 17 1 "C_fixnum_negate" "C_u_fixnum_negate")
-(rewrite 'bitwise-and 17 2 "C_fixnum_and" "C_u_fixnum_and")
-(rewrite 'bitwise-ior 17 2 "C_fixnum_or" "C_u_fixnum_or")
+(rewrite 'bitwise-xor 19 0 "C_fixnum_xor" "C_fixnum_xor" #t)
+(rewrite 'bitwise-and 19 -1 "C_fixnum_and" "C_u_fixnum_and" #t)
+(rewrite 'bitwise-ior 19 0 "C_fixnum_or" "C_u_fixnum_or" #t)
+
+(rewrite
+ 'arithmetic-shift 8
+ (lambda (db classargs cont callargs)
+   ;; (arithmetic-shift <x> <-int>) -> (##core#inline "C_fixnum_shift_right" <x> -<int>)
+   ;; (arithmetic-shift <x> <+int>) -> (##core#inline "C_fixnum_shift_left" <x> <int>)
+   ;; _ -> (##core#inline "C_i_arithmetic_shift" <x> <y>)
+   (and (= 2 (length callargs))
+	(let ([val (second callargs)])
+	  (make-node
+	   '##core#call '(#t)
+	   (list cont
+		 (or (and-let* ([(eq? 'quote (node-class val))]
+				[n (first (node-parameters val))]
+				[(fixnum? n)] )
+		       (if (negative? n)
+			   (make-node 
+			    '##core#inline '("C_fixnum_shift_right")
+			    (list (first callargs) (qnode (- n))) )
+			   (make-node
+			    '##core#inline '("C_fixnum_shift_left")
+			    (list (first callargs) val) ) ) )
+		     (make-node '##core#inline '("C_i_arithmetic_shift") callargs) ) ) ) ) ) ) )
+
 (rewrite '##sys#byte 17 2 "C_subbyte")
 (rewrite '##sys#setbyte 17 3 "C_setbyte")
 (rewrite '##sys#peek-fixnum 17 2 "C_peek_fixnum")
+(rewrite '##sys#peek-byte 17 2 "C_peek_byte")
+(rewrite 'pointer->object 17 2 "C_pointer_to_object")
 (rewrite '##sys#setislot 17 3 "C_i_set_i_slot")
 (rewrite '##sys#poke-integer 17 3 "C_poke_integer")
 (rewrite 'string=? 17 2 "C_i_string_equal_p" "C_u_i_string_equal_p")
-(rewrite 'string-ci=? 17 2 "C_i_string_equal_p")
+(rewrite 'string-ci=? 17 2 "C_i_string_ci_equal_p")
 (rewrite '##sys#fudge 17 1 "C_fudge")
-(rewrite '##sys#tty-port? 17 1 "C_tty_portp")
 (rewrite '##sys#fits-in-int? 17 1 "C_fits_in_int_p")
 (rewrite '##sys#fits-in-unsigned-int? 17 1 "C_fits_in_unsigned_int_p")
 (rewrite '##sys#flonum-in-fixnum-range? 17 1 "C_flonum_in_fixnum_range_p")
 (rewrite '##sys#permanent? 17 1 "C_permanentp")
 (rewrite '##sys#null-pointer? 17 1 "C_null_pointerp" "C_null_pointerp")
 (rewrite 'null-pointer? 17 1 "C_i_null_pointerp" "C_null_pointerp")
+(rewrite '##sys#immediate? 17 1 "C_immp")
+(rewrite 'locative->object 17 1 "C_i_locative_to_object")
+(rewrite 'locative-set! 17 2 "C_i_locative_set")
+(rewrite '##sys#foreign-fixnum-argument 17 1 "C_i_foreign_fixnum_argumentp")
+(rewrite '##sys#foreign-char-argument 17 1 "C_i_foreign_char_argumentp")
+(rewrite '##sys#foreign-flonum-argument 17 1 "C_i_foreign_flonum_argumentp")
+(rewrite '##sys#foreign-block-argument 17 1 "C_i_foreign_block_argumentp")
+(rewrite '##sys#foreign-number-vector-argument 17 2 "C_i_foreign_number_vector_argumentp")
+(rewrite '##sys#foreign-string-argument 17 1 "C_i_foreign_string_argumentp")
+(rewrite '##sys#foreign-pointer-argument 17 1 "C_i_foreign_pointer_argumentp")
+(rewrite '##sys#foreign-integer-argument 17 1 "C_i_foreign_integer_argumentp")
+(rewrite '##sys#foreign-unsigned-integer-argument 17 1 "C_i_foreign_unsigned_integer_argumentp")
+(rewrite '##sys#direct-return 17 2 "C_direct_return")
 
 (rewrite 'byte-vector-ref 2 2 "C_subbyte" #f #f)
 (rewrite 'byte-vector-set! 2 3 "C_setbyte" #f #f)
 (rewrite 'byte-vector-length 2 1 "C_block_size" #f #f)
 
-(rewrite 
- 'call-with-current-continuation 8
- (lambda (db classargs cont callargs)
-   ;; (call-with-current-continuation <x>) -> ((##core#proc "C_call_cc") <x>) [no-winding-callcc mode]
-   (and (= (length callargs) 1)
-	nonwinding-call/cc
-	(let ([arg (first callargs)])
-	  (make-node 
-	   '##core#call '(#t)
-	   (cons* (make-node '##core#proc (list "C_call_cc") '())
-		  cont
-		  callargs) ) ) ) ) )
-
 (rewrite 'not-pair? 17 1 "C_i_not_pair_p")
 (rewrite 'null-list? 17 1 "C_i_null_list_p" "C_i_nullp")
 
-(let ()
-  (define (rewrite-print nl)
-    (lambda (db classargs cont callargs)
-      ;; (print[*] <x1> ...) -> <a chain of primitive display ops> ['display' is std binding + no-fancy-ports]
-      ;; - All <xI> have to be character or string literals.
-      (and file-io-only
-	   (get db 'display 'standard-binding)
-	   (every (lambda (x)
-		    (and-let* ([(eq? 'quote (node-class x))]
-			       [c (first (node-parameters x))] )
-		      (or (char? c) (string? c)) ) )
-		  callargs)
-	   (let ([portvar (gensym)])
-	     (make-node
-	      'let (list portvar)
-	      (list
-	       (varnode '##sys#standard-output)
-	       (fold-right
-		(lambda (x body)
-		  (let ([c (first (node-parameters x))])
-		    (make-node
-		     'let (list (gensym))
-		     (list (make-node
-			    '##core#inline (list (if (char? c) "C_display_char" "C_display_string"))
-			    (list (varnode portvar) x) )
-			   body) ) ) )
-		(let ([cont2 (make-node '##core#call '(#t) (list cont (make-node '##core#undefined '() '())))])
-		  (if nl
-		      (make-node
-		       'let (list (gensym))
-		       (list (make-node '##core#inline '("C_display_char") (list (varnode portvar) (qnode #\newline)))
-			     cont2) )
-		      cont2) )
-		callargs) ) ) ) ) ) )
-  (rewrite 'print 8 (rewrite-print #t))
-  (rewrite 'print* 8 (rewrite-print #f)) )
-
 (rewrite 'u8vector->byte-vector 7 1 "C_slot" 1 #f)
 (rewrite 's8vector->byte-vector 7 1 "C_slot" 1 #f)
 (rewrite 'u16vector->byte-vector 7 1 "C_slot" 1 #f)
