@@ -1,6 +1,6 @@
 ;;;; chicken.scm - The CHICKEN Scheme compiler (loader/main-module)
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -33,27 +33,27 @@
 ; Germany
 
 
-(declare (uses extras srfi-1 match modules syntax-case srfi-4 support compiler driver platform backend))
+(declare (uses extras srfi-1 match-support syntax-case srfi-4 script-utils support compiler optimizer driver platform backend easyffi))
 
 
 #{compiler
   build-information compiler-arguments
-  default-analysis-database-size default-standard-bindings default-extended-bindings side-effecting-standard-bindings
+  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home optimization-iterations process-command-line
-  foreign-type-table-size file-io-only nonwinding-call/cc
-  unit-name insert-timer-checks used-units inlining zap-strings-flag
+  foreign-type-table-size file-io-only nonwinding-call/cc debugging
+  unit-name insert-timer-checks used-units zap-strings-flag
   foreign-declarations debugging-executable block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home 
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  dependency-list broken-constant-nodes inline-substitutions-enabled
+  broken-constant-nodes inline-substitutions-enabled
   direct-call-ids foreign-type-table first-analysis
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
+  reorganize-recursive-bindings substitution-table simplify-named-call
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas!
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
@@ -72,7 +72,12 @@
 
 
 (include "tweaks")
-(include "moremacros")
+
+(eval-when (load) 
+  (include "moremacros")
+  (include "match") )
+
+(##sys#provide 'extras 'srfi-1 'syntax-case 'srfi-4)
 
 
 ;;; Retrieve build parameters:
@@ -80,7 +85,18 @@
 (include "build.scm")
 
 (define build-information
-  (sprintf "Version ~s, Build ~s ~c" build-version build-number (##sys#fudge 3)) )
+  (let ()
+    (define (get-config)
+      (let ([bp (build-platform)]
+	    [st (software-type)]
+	    [sv (software-version)]
+	    [mt (machine-type)] )
+	(define (str x)
+	  (if (eq? x 'unknown)
+	      ""
+	      (string-append (symbol->string x) "-") ) )
+	(string-append (str sv) (str st) (str bp) (symbol->string mt)) ) )
+    (sprintf "Version ~s, Build ~s - ~a" build-version build-number (get-config)) ) )
 
 
 ;;; Load other modules if interpreted:
@@ -91,6 +107,7 @@
   (unless (memq '#:match (features)) (include "match"))
   (include "support")
   (include "compiler")
+  (include "optimizer")
   (include "batch-driver")
   (include "c-platform")
   (include "c-backend")
@@ -139,19 +156,14 @@
 		 (case level
 		   [(0) #f]
 		   [(1)
-		    (set! options (cons* 'no-trace 'optimize 'optimize-leaf-routines options)) ]
+		    (set! options (cons* 'optimize-leaf-routines 'lambda-lift options)) ]
 		   [(2)
 		    (set! options
-		      (cons* 'no-trace 'optimize 'optimize-leaf-routines 'usual-integrations options) ) ] 
+		      (cons* 'optimize-leaf-routines 'lambda-lift 'usual-integrations options) ) ] 
 		   [(3)
 		    (set! options
-		      (cons* 'no-trace 'optimize 'optimize-leaf-routines 'usual-integrations 'unsafe options) ) ]
-		   [else
-		    (if (>= level 4)
-			(set! options 
-			  (cons* 'no-trace 'optimize 'optimize-leaf-routines 'usual-integrations 'unsafe 'inline 
-				 options) )
-			(warning "invalid optimization level ~S - ignored" (car rest)) ) ] )
+		      (cons* 'optimize-leaf-routines 'lambda-lift 'usual-integrations 'unsafe options) ) ]
+		   [else (warning "invalid optimization level ~S - ignored" (car rest))] )
 		 (loop (cdr rest)) ) ]
 	      [(eq? 'debug-level o)
 	       (let ([level (string->number (car rest))])
@@ -160,15 +172,16 @@
 		   [(1) #f]
 		   [(2) (set! options (cons* 'no-trace 'emit-debug-info options))]
 		   [(3) (set! options (cons* 'no-trace 'emit-debug-info 'debug-calls options))]
-		   [else
-		    (if (>= level 4)
-			(set! options (cons* 'no-trace 'emit-debug-info 'debug-calls 'debug-loops options))
-			(warning "invalid debug level ~S - ignored" (car rest)) ) ] )
+		   [(4) (set! options (cons* 'no-trace 'emit-debug-info 'debug-calls 'debug-loops options))]
+		   [else (warning "invalid debug level ~S - ignored" (car rest))] )
 		 (loop (cdr rest)) ) ]
+	      [(eq? 'r5rs o)
+	       (set! options (cons* 'hygienic 'strict options))
+	       (loop rest) ]
 	      [(eq? 'benchmark-mode o)
 	       (set! options 
-		 (cons* 'fixnum-arithmetic 'disable-interrupts 'no-trace 'optimize 'usual-integrations 'unsafe
-			'optimize-leaf-routines 'block
+		 (cons* 'fixnum-arithmetic 'disable-interrupts 'no-trace 'usual-integrations 'unsafe
+			'optimize-leaf-routines 'block 'lambda-lift
 			options) )
 	       (loop rest) ]
 	      [(memq o valid-compiler-options) (loop rest)]
