@@ -1,9 +1,9 @@
 ;;;; mmcp.scm - A simple command line tool for copying files with or without mmapped I/O
 ;
-; Taken from Richard Stevens' "Advanced Programming in the UNIX environment"
+; (taken directly from Richard Stevens' "Advanced Programming in the UNIX environment")
 
 
-(declare (uses unistd lolevel))
+(declare (uses posix lolevel))
 
 
 (define-foreign-variable bufsiz int "BUFSIZ")
@@ -27,7 +27,7 @@
 	(let ([buffer (make-string bufsiz)])
 	  (let loop ([n size1])
 	    (when (positive? n)
-	      (let-values ([(buf m) (file-read in bufsiz buffer)])
+	      (let ([m (cadr (file-read in bufsiz buffer))])
 		(file-write out buffer m)
 		(loop (- n m)) ) ) ) ) ) ) )
 
