@@ -3,21 +3,41 @@
 
 (declare
  (unit format)
- (uses extras)
- (standard-bindings)
- (extended-bindings) 
- (no-bound-checks)
- (bound-to-procedure
-  ##format#out-str ##format#out-char ##format#out-substr ##format#out-fill ##format#error ##format#error-save
-  ##format#format ##format#intern-error ##format#out ##format#work ##format#obj->str ##format#iobj-str ##format#par
-  ##format#out-obj-padded ##format#out-num-padded ##format#tabulate ##format#num->old-roman ##format#num->roman
-  ##format#num->cardinal999 ##format#num->cardinal ##format#num->ordinal ##format#out-fixed ##format#out-expon
-  ##format#out-general ##format#out-dollar ##format#parse-float ##format#en-int ##format#en-set
-  ##format#fn-zfill ##format#fn-shiftleft ##format#fn-round ##format#fn-out ##format#en-out
-  ##format#fn-strip ##format#fn-zlead ##format#string-capitalize-first ##format#list-head ##format#abort
-  ##format#char->str) )
+ (disable-interrupts)
+ (hide ##format#fn-max ##format#ordinal-ones-list ##format#ordinal-tens-list
+       ##format#cardinal-ones-list ##format#cardinal-tens-list ##format#fn-zlead ##format#fn-strip
+       ##format#error-save ##format#en-out ##format#ascii-non-printable-charnames ##format#out-expon
+       ##format#fn-out ##format#args ##format#string-capitalize-first ##format#tabulate ##format#en-set
+       ##format#out-fill ##format#en-str ##format#error ##format#space-ch ##format#fn-len
+       ##format#fn-round ##format#out-char ##format#out-general ##format#fn-str ##format#en-max
+       ##format#fn-shiftleft ##format#out-str ##format#par ##format#fn-dot
+       ##format#cardinal-thousands-block-list ##format#out-fixed ##format#out ##format#output-col
+       ##format#flush-output ##format#pos ##format#intern-error ##format#num->cardinal999 ##format#zero-ch
+       ##format#parameter-characters ##format#roman-alist ##format#arg-pos ##format#list-head
+       ##format#fn-zfill ##format#error-continuation ##format#format ##format#en-int
+       ##format#roman-boundary-values ##format#out-dollar ##format#abort ##format#out-num-padded
+       ##format#out-obj-padded ##format#parse-float ##format#en-pos? ##format#fn-pos?
+       ##format#num->ordinal ##format#num->cardinal ##format#num->old-roman ##format#char->str
+       ##format#format-work ##format#en-len ##format#out-substr ##format#num->roman
+       ##format#call-with-output-string)
+ (usual-integrations) )
 
-(register-feature! 'format)
+(cond-expand
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks)
+    (bound-to-procedure
+     ##format#out-str ##format#out-char ##format#out-substr ##format#out-fill ##format#error ##format#error-save
+     ##format#format ##format#intern-error ##format#out ##format#work ##format#obj->str ##format#iobj-str ##format#par
+     ##format#out-obj-padded ##format#out-num-padded ##format#tabulate ##format#num->old-roman ##format#num->roman
+     ##format#num->cardinal999 ##format#num->cardinal ##format#num->ordinal ##format#out-fixed ##format#out-expon
+     ##format#out-general ##format#out-dollar ##format#parse-float ##format#en-int ##format#en-set
+     ##format#fn-zfill ##format#fn-shiftleft ##format#fn-round ##format#fn-out ##format#en-out
+     ##format#fn-strip ##format#fn-zlead ##format#string-capitalize-first ##format#list-head ##format#abort
+     ##format#char->str) ) ] )
+
+(register-feature! 'format 'srfi-28)
 
 
 ;;; "format.scm" Common LISP text output formatter for SLIB
@@ -63,6 +83,68 @@
      "dle" "dc1" "dc2" "dc3" "dc4" "nak" "syn" "etb"
      "can" "em"  "sub" "esc" "fs"  "gs"  "rs"  "us" "space"))
 
+
+;;; Case conversion:
+
+(define ##format#string-upcase! 
+  (lambda (str)
+    (do ((i (fx- (##sys#size str) 1)
+	    (fx- i 1)))
+	((fx< i 0) str)
+      (##core#inline "C_setsubchar"
+		 str i
+		 (##core#inline "C_u_i_char_upcase" (##core#inline "C_subchar" str i))) ) ) )
+
+(define ##format#string-upcase
+  (let ([string-copy string-copy])
+    (lambda (str)
+      (##format#string-upcase! (string-copy str)) ) ) )
+
+(define ##format#string-downcase! 
+  (lambda (str)
+    (do ((i (fx- (##sys#size str) 1)
+	    (fx- i 1)))
+	((fx< i 0) str)
+      (##core#inline "C_setsubchar" 
+		 str i
+		 (##core#inline "C_u_i_char_downcase" (##core#inline "C_subchar" str i))) ) ) )
+
+(define ##format#string-downcase
+  (let ([string-copy string-copy])
+    (lambda (str)
+      (##format#string-downcase! (string-copy str)) ) ) )
+
+(define (##format#string-capitalize! str)	; stolen from slib
+  (let ((non-first-alpha #f)	
+	(str-len (##sys#size str)))
+    (do ((i 0 (fx+ i 1)))	
+	((eq? i str-len) str)
+      (let ((c (##core#inline "C_subchar" str i)))
+	(if (##core#inline "C_u_i_char_alphabeticp" c)
+	    (if non-first-alpha
+		(##core#inline "C_setsubchar" str i (##core#inline "C_u_i_char_downcase" c))
+		(begin
+		  (set! non-first-alpha #t)
+		  (##core#inline "C_setsubchar" str i (##core#inline "C_u_i_char_upcase" c))))
+	    (set! non-first-alpha #f))))))
+
+(define ##format#string-capitalize 
+  (let ([string-copy string-copy])
+    (lambda (str)
+      (##format#string-capitalize! (string-copy str))) ) )
+
+
+;;; We use this so we don't depend on extras:
+
+(define ##format#call-with-output-string
+  (let ((open-output-string open-output-string)
+	(get-output-string get-output-string) )
+    (lambda (proc)
+      (let ((out (open-output-string)))
+	(proc out)
+	(get-output-string out) ) ) ) )
+
+
 ;;; End of configuration ----------------------------------------------------
 
 (define ##format#version "3.0")
@@ -182,7 +264,7 @@
 		    (car arglist) (cdr arglist)))
        ((and (boolean? destination)	; string output
 	     (not destination))
-	(call-with-output-string
+	(##format#call-with-output-string
 	 (lambda (port) (##format#out port (car arglist) (cdr arglist)))) )
        (else
 	(##format#error "illegal destination `~a'" destination))))))
@@ -515,9 +597,9 @@
 	       (set! ##format#case-conversion
 		     (case modifier
 		       ((at) ##format#string-capitalize-first)
-		       ((colon) string-capitalize)
-		       ((colon-at) string-upcase)
-		       (else string-downcase)))
+		       ((colon) ##format#string-capitalize)
+		       ((colon-at) ##format#string-upcase)
+		       (else ##format#string-downcase)))
 	       (anychar-dispatch))
 	      ((#\))			; Case conversion end
 	       (if (not ##format#case-conversion)
@@ -854,7 +936,7 @@
     (##format#iobj->str obj))))		; representation should fall in here
 
 ;; ##format#iobj->str reveals the implementation dependent representation of
-;; #<...> objects with the use of display and call-with-output-string.
+;; #<...> objects with the use of display and ##format#call-with-output-string.
 ;; If ##format#read-proof is set to #t the resulting string is additionally
 ;; set into string quotes.
 
@@ -867,10 +949,10 @@
        (if ##format#read-proof "\"" "")
        (if ##format#iobj-case-conv
 	   (##format#iobj-case-conv
-	    (call-with-output-string (lambda (p) (display iobj p))))
-	   (call-with-output-string (lambda (p) (display iobj p))))
+	    (##format#call-with-output-string (lambda (p) (display iobj p))))
+	   (##format#call-with-output-string (lambda (p) (display iobj p))))
        (if ##format#read-proof "\"" ""))
-      (call-with-output-string (lambda (p) (display iobj p)))))
+      (##format#call-with-output-string (lambda (p) (display iobj p)))))
 
 
 ;; ##format#char->str converts a character into a slashified string as
@@ -1672,9 +1754,7 @@
 ;; Aborts the program when a formatting error occures. This is a null
 ;; argument closure to jump to the interpreters toplevel continuation.
 
-(define ##format#abort (lambda () (error "error in format")))
-
-(define format ##format#format)
+(define ##format#abort (lambda () (##sys#error "error in format")))
 
 ;; If this is not possible then a continuation is used to recover
 ;; properly from a format error. In this case format returns #f.
@@ -1689,4 +1769,14 @@
 ;       (set! ##format#error-continuation cont)
 ;       (apply ##format#format args)))))
 
-;eof
+
+;;; Always use intermediate string, if printing to a port (this module is not thread-safe):
+
+(define (format dest . more)
+  (if (or (not dest) (string? dest)) 
+      (apply ##format#format dest more)
+      (display
+       (##format#call-with-output-string (lambda (p) (apply ##format#format p more)))
+       (if (eq? #t dest)
+	   ##sys#standard-output
+	   dest) ) ) )
