@@ -1,6 +1,6 @@
 ;;;; profiler.scm - Support code for profiling applications
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -36,10 +36,16 @@
 (declare
   (unit profiler)
   (usual-integrations)
+  (hide ##sys#profile-name ##sys#profile-vector-list)
   (disable-interrupts)
-  (no-bound-checks)
   (fixnum-arithmetic) )
 
+(cond-expand
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks) ) ] )
+
 
 (include "parameters.scm")
 
