@@ -4,7 +4,7 @@
 
    - Runs 'csi' with the given arguments like this:
 
-     csi -quiet -batch -hygienic -srfi-7 SCRIPTFILENAME -no-init -strict -eval "(##sys#script-main 12)" -- ARGUMENT1 ...
+     csi -quiet -batch -hygienic -srfi-7 SCRIPTFILENAME -no-init -strict -eval "(##sys#script-main 5 11 main)" -- ARGUMENT1 ...
 */
 
 
@@ -12,11 +12,12 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <errno.h>
+#include <string.h>
 
 
 #define PREFIX_ARGS        11
 #define SCRIPT_ARG         5
-#define CMD_PARAMETER      "(##sys#script-main 5 12)"
+#define CMD_PARAMETER      "(##sys#script-main 5 11 main)"
 
 
 int main(int argc, char *argv[])
@@ -27,7 +28,7 @@ int main(int argc, char *argv[])
 		     "-no-init", "-strict",
 		     "-eval", CMD_PARAMETER,
 		     "--" },
-       **args = (char **)malloc(argc + 12);
+       **args = (char **)malloc((argc + 12) * sizeof(char *));
   int i;
 
   if(args == NULL) {
