@@ -5,7 +5,10 @@
   (unit srfi-13)
   (uses srfi-14)
   (fixnum)
-  (no-bound-checks)
+  (hide %string-prefix? %string-hash %finish-string-concatenate-reverse %string-suffix-length %string-prefix-length
+	%string-map %string-copy! %string-compare %substring/shared %string-suffix? %multispan-repcopy!
+	%string-prefix-length-ci %string-suffix-length-ci %string-prefix-ci? %string-suffix-ci?
+	%string-titlecase! %string-map! %kmp-search %string-compare-ci ##srfi13#string-fill!)
   (standard-bindings not boolean? apply call-with-current-continuation eq? eqv? equal? pair? cons car cdr caar cadr
 		     cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar caaadr caadar caaddr cadaar
 		     cadadr caddar cadddr cdaaar cdaadr cdadar cdaddr cddaar cddadr cdddar cddddr set-car! set-cdr!
@@ -24,26 +27,34 @@
 		     open-output-file call-with-input-file call-with-output-file close-input-port close-output-port
 		     port? values call-with-values vector procedure? memq memv assq assv member assoc) 
   (extended-bindings)
-  (interrupts-disabled) )
+  (disable-interrupts) )
+
+(cond-expand
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks) ) ] )
 
 (cond-expand
  [unsafe
   (eval-when (compile)
-    (define-macro (##sys#check-structure x y) '(##core#undefined))
-    (define-macro (##sys#check-range x y z) '(##core#undefined))
-    (define-macro (##sys#check-pair x) '(##core#undefined))
-    (define-macro (##sys#check-list x) '(##core#undefined))
-    (define-macro (##sys#check-symbol x) '(##core#undefined))
-    (define-macro (##sys#check-string x) '(##core#undefined))
-    (define-macro (##sys#check-char x) '(##core#undefined))
-    (define-macro (##sys#check-exact x) '(##core#undefined))
-    (define-macro (##sys#check-port x) '(##core#undefined))
-    (define-macro (##sys#check-number x) '(##core#undefined))
-    (define-macro (##sys#check-bytevector x) '(##core#undefined)) ) ]
+    (define-macro (##sys#check-structure . _) '(##core#undefined))
+    (define-macro (##sys#check-range . _) '(##core#undefined))
+    (define-macro (##sys#check-pair . _) '(##core#undefined))
+    (define-macro (##sys#check-list . _) '(##core#undefined))
+    (define-macro (##sys#check-symbol . _) '(##core#undefined))
+    (define-macro (##sys#check-string . _) '(##core#undefined))
+    (define-macro (##sys#check-char . _) '(##core#undefined))
+    (define-macro (##sys#check-exact . _) '(##core#undefined))
+    (define-macro (##sys#check-port . _) '(##core#undefined))
+    (define-macro (##sys#check-number . _) '(##core#undefined))
+    (define-macro (##sys#check-bytevector . _) '(##core#undefined)) ) ]
  [else] )
 
 (register-feature! 'srfi-13)
 
+(include "srfi-13-syntax")
+
 
 (cond-expand [hygienic-macros
 (define char-cased? char-alphabetic?)
@@ -205,7 +216,7 @@
 ;;; Returns three values: rest start end
 
 (define (string-parse-start+end proc s args)
-  (if (not (string? s)) (error "Non-string value" proc s))
+  (##sys#check-string s 'string-parse-start+end)
   (let ((slen (string-length s)))
     (if (pair? args)
 
@@ -220,18 +231,18 @@
 ;			(if (and (integer? end) (exact? end) (<= end slen))
 			(if (and (fixnum? end) (<= end slen))
 			    (values end args)
-			    (error "Illegal substring END spec" proc end s)))
+			    (##sys#error 'string-parse-start+end "Illegal substring END spec" proc end s)))
 		      (values slen args))
 		(if (<= start end) (values args start end)
-		    (error "Illegal substring START/END spec"
+		    (##sys#error 'string-parse-start+end "Illegal substring START/END spec"
 			   proc start end s)))
-	      (error "Illegal substring START spec" proc start s)))
+	      (##sys#error 'string-parse-start+end "Illegal substring START spec" proc start s)))
 
 	(values '() 0 slen))))
 
 (define (string-parse-final-start+end proc s args)
   (receive (rest start end) (string-parse-start+end proc s args)
-    (if (pair? rest) (error "Extra arguments to procedure" proc rest)
+    (if (pair? rest) (##sys#error 'string-parse-final-start+end "Extra arguments to procedure" proc rest)
 	(values start end))))
 
 (define (substring-spec-ok? s start end)
@@ -248,7 +259,7 @@
 
 (define (check-substring-spec proc s start end)
   (if (not (substring-spec-ok? s start end))
-      (error "Illegal substring spec." proc s start end)))
+      (##sys#error 'check-substring-spec "Illegal substring spec." proc s start end)))
 
 
 ;;; Defined by R5RS, so commented out here.
@@ -278,7 +289,7 @@
 ;    (check-arg (lambda (start) (and (integer? start) (exact? start) (<= 0 start)))
 ;	       start substring/shared)
     (let ([n (:optional maybe-end slen)])
-      (##sys#check-exact n)
+      (##sys#check-exact n 'substring/shared)
       (%substring/shared s start n) ) ) )
 #|
     (%substring/shared s start
@@ -560,7 +571,7 @@
 		   (if (= i1 end) (criteria c)	; Tail call.
 		       (and (criteria c) (lp i1)))))))
 
-	  (else (error "Second param is neither char-set, char, or predicate procedure."
+	  (else (##sys#error 'string-every "Second param is neither char-set, char, or predicate procedure."
 		       string-every criteria)))))
 
 
@@ -586,7 +597,7 @@
 		    (if (= i1 end) (criteria c)	; Tail call
 			(or (criteria c) (lp i1)))))))
 
-	  (else (error "Second param is neither char-set, char, or predicate procedure."
+	  (else (##sys#error 'string-any "Second param is neither char-set, char, or predicate procedure."
 		       string-any criteria)))))
 
 
@@ -594,7 +605,7 @@
 ;  (check-arg procedure? proc string-tabulate)
 ;  (check-arg (lambda (val) (and (integer? val) (exact? val) (<= 0 val)))
 ;	     len string-tabulate)
-  (##sys#check-exact len)
+  (##sys#check-exact len 'string-tabulate)
   (let ((s (make-string len)))
     (do ((i (- len 1) (- i 1)))
 	((< i 0))
@@ -963,7 +974,7 @@
 							;     (<= 0 bound)))
 					 rest)
     (if (zero? bound) (set! bound 4194304))
-    (##sys#check-exact bound)		
+    (##sys#check-exact bound 'string-hash)		
     (let-string-start+end (start end) string-hash s rest
       (%string-hash s char->integer bound start end))))
 
@@ -973,7 +984,7 @@
 							 ;    (<= 0 bound)))
 					 rest)
     (if (zero? bound) (set! bound 4194304))
-    (##sys#check-exact bound)
+    (##sys#check-exact bound 'string-hash-ci)
     (let-string-start+end (start end) string-hash-ci s rest
       (%string-hash s (lambda (c) (char->integer (char-downcase c)))
 		    bound start end))))
@@ -1052,7 +1063,7 @@
 ;  (check-arg (lambda (val) (and (integer? n) (exact? n)
 ;				(<= 0 n (string-length s))))
 ;	     n string-take)
-  (##sys#check-exact n)
+  (##sys#check-exact n 'string-take)
   (%substring/shared s 0 n))
 
 (define (string-take-right s n)
@@ -1060,7 +1071,7 @@
   (let ((len (string-length s)))
 ;    (check-arg (lambda (val) (and (integer? n) (exact? n) (<= 0 n len)))
 ;	       n string-take-right)
-    (##sys#check-exact n)
+    (##sys#check-exact n 'string-take-right)
     (%substring/shared s (- len n) len)))
 
 (define (string-drop s n)
@@ -1068,7 +1079,7 @@
   (let ((len (string-length s)))
 ;    (check-arg (lambda (val) (and (integer? n) (exact? n) (<= 0 n len)))
 ;	       n string-drop)
-    (##sys#check-exact n)
+    (##sys#check-exact n 'string-drop)
   (%substring/shared s n len)))
 
 (define (string-drop-right s n)
@@ -1076,7 +1087,7 @@
   (let ((len (string-length s)))
 ;    (check-arg (lambda (val) (and (integer? n) (exact? n) (<= 0 n len)))
 ;	       n string-drop-right)
-    (##sys#check-exact n)
+    (##sys#check-exact n 'string-drop-right)
     (%substring/shared s 0 (- len n))))
 
 
@@ -1104,7 +1115,7 @@
 
 
 (define (string-pad-right s n . char+start+end)
-  (##sys#check-exact n)
+  (##sys#check-exact n 'string-pad-right)
   (let-optionals* char+start+end ((char #\space) rest) ; (char? char)) rest)
     (let-string-start+end (start end) string-pad-right s rest
       (let ((len (- end start)))
@@ -1115,7 +1126,7 @@
 	      ans))))))
 
 (define (string-pad s n . char+start+end)
-  (##sys#check-exact n)
+  (##sys#check-exact n 'string-pad)
   (let-optionals* char+start+end ((char #\space) rest) ; (char? char)) rest)
     (let-string-start+end (start end) string-pad s rest
       (let ((len (- end start)))
@@ -1155,7 +1166,7 @@
 
 	(let* ((cset (cond ((char-set? criteria) criteria)
 			   ((char? criteria) (char-set criteria))
-			   (else (error "string-delete criteria not predicate, char or char-set" criteria))))
+			   (else (##sys#error 'string-delete "string-delete criteria not predicate, char or char-set" criteria))))
 	       (len (string-fold (lambda (c i) (if (char-set-contains? cset c)
 						   i
 						   (+ i 1)))
@@ -1183,7 +1194,7 @@
 
 	(let* ((cset (cond ((char-set? criteria) criteria)
 			   ((char? criteria) (char-set criteria))
-			   (else (error "string-delete criteria not predicate, char or char-set" criteria))))
+			   (else (##sys#error 'string-filter "string-delete criteria not predicate, char or char-set" criteria))))
 
 	       (len (string-fold (lambda (c i) (if (char-set-contains? cset c)
 						   (+ i 1)
@@ -1226,7 +1237,7 @@
 	     (and (< i end)
 		  (if (criteria (string-ref str i)) i
 		      (lp (+ i 1))))))
-	  (else (error "Second param is neither char-set, char, or predicate procedure."
+	  (else (##sys#error 'string-index "Second param is neither char-set, char, or predicate procedure."
 		       string-index criteria)))))
 
 (define (string-index-right str criteria . maybe-start+end)
@@ -1246,7 +1257,7 @@
 	     (and (>= i 0)
 		  (if (criteria (string-ref str i)) i
 		      (lp (- i 1))))))
-	  (else (error "Second param is neither char-set, char, or predicate procedure."
+	  (else (##sys#error 'string-index-right "Second param is neither char-set, char, or predicate procedure."
 		       string-index-right criteria)))))
 
 (define (string-skip str criteria . maybe-start+end)
@@ -1268,7 +1279,7 @@
 	     (and (< i end)
 		  (if (criteria (string-ref str i)) (lp (+ i 1))
 		      i))))
-	  (else (error "Second param is neither char-set, char, or predicate procedure."
+	  (else (##sys#error 'string-skip "Second param is neither char-set, char, or predicate procedure."
 		       string-skip criteria)))))
 
 (define (string-skip-right str criteria . maybe-start+end)
@@ -1290,11 +1301,13 @@
 	     (and (>= i 0)
 		  (if (criteria (string-ref str i)) (lp (- i 1))
 		      i))))
-	  (else (error "CRITERIA param is neither char-set or char."
+	  (else (##sys#error 'string-skip-right "CRITERIA param is neither char-set or char."
 		       string-skip-right criteria)))))
 
 
-(define (string-count criteria s . maybe-start+end)
+; [felix] Boooh! original code had "s" and "criteria" in the wrong order:
+
+(define (string-count s criteria . maybe-start+end)
   (let-string-start+end (start end) string-count s maybe-start+end
     (cond ((char? criteria)
 	   (do ((i start (+ i 1))
@@ -1315,7 +1328,7 @@
 		(count 0 (if (criteria (string-ref s i)) (+ count 1) count)))
 	       ((>= i end) count)))
 
-	  (else (error "CRITERIA param is neither char-set or char."
+	  (else (##sys#error 'string-count "CRITERIA param is neither char-set or char."
 		       string-count criteria)))))
 
 
@@ -1336,7 +1349,7 @@
 (define (string-copy! to tstart from . maybe-fstart+fend)
   (let-string-start+end (fstart fend) string-copy! from maybe-fstart+fend
 ;    (check-arg integer? tstart string-copy!)
-    (##sys#check-exact tstart)			
+    (##sys#check-exact tstart 'string-copy!)			
     (check-substring-spec string-copy! to tstart (+ tstart (- fend fstart)))
     (%string-copy! to tstart from fstart fend)))
 
@@ -1456,7 +1469,7 @@
 (define (make-kmp-restart-vector pattern . maybe-c=+start+end)
   (let-optionals* maybe-c=+start+end
                   ((c= char=?) rest) ; (procedure? c=))
-     (receive (start end) (string-parse-start+end make-kmp-restart-vector pattern args)
+     (receive (start end) (string-parse-start+end make-kmp-restart-vector pattern rest)
        (let* ((rvlen (- end start))
 	   (rv (make-vector rvlen -1)))
       (if (> rvlen 0)
@@ -1526,7 +1539,7 @@
   (let-optionals* c=+p-start+s-start+s-end
       ((c=      char=?) ; (procedure? c=))
        (p-start 0) rest) ; (and (integer? p-start) (exact? p-start) (<= 0 p-start)))
-    (receive (s-start s-end) (string-parse-start+end string-kmp-partial-search s args)
+    (receive (s-start s-end) (string-parse-start+end string-kmp-partial-search s rest)
     ;; Enough prelude. Here's the actual code.
     (let ((patlen (vector-length rv)))
       (let lp ((si s-start)		; An index into S.
@@ -1552,7 +1565,7 @@
 ;;; (reverse-list->string clist)
 ;;; (string->list s [start end])
 
-(define (string-null? s) (zero? (string-length s)))
+(define (string-null? s) (##core#inline "C_i_string_null_p" s))
 
 (define (string-reverse s . maybe-start+end)
   (let-string-start+end (start end) string-reverse s maybe-start+end
@@ -1675,7 +1688,7 @@
 ;					(and (integer? end)
 ;					     (exact? end)
 ;					     (<= 0 end (string-length final)))))
-    (##sys#check-exact end)
+    (##sys#check-exact end 'string-concatenate-reverse)
     (let ((len (let lp ((sum 0) (lis string-list))
 		 (if (pair? lis)
 		     (lp (+ sum (string-length (car lis))) (cdr lis))
@@ -1689,7 +1702,7 @@
 ;					(and (integer? end)
 ;					     (exact? end)
 ;					     (<= 0 end (string-length final)))))
-    (##sys#check-exact end)
+    (##sys#check-exact end 'string-concatenate-reverse/shared)
     ;; Add up the lengths of all the strings in STRING-LIST; also get a
     ;; pointer NZLIST into STRING-LIST showing where the first non-zero-length
     ;; string starts.
@@ -1795,7 +1808,7 @@
 (define (xsubstring s from . maybe-to+start+end)
 ;  (check-arg (lambda (val) (and (integer? val) (exact? val)))
 ;	     from xsubstring)
-  (##sys#check-exact from)
+  (##sys#check-exact from 'xsubstring)
   (receive (to start end)
            (if (pair? maybe-to+start+end)
 	       (let-string-start+end (start end) xsubstring s (cdr maybe-to+start+end)
@@ -1804,7 +1817,7 @@
 ;						 (exact? val)
 ;						 (<= from val)))
 ;			      to xsubstring)
-		   (##sys#check-exact to)
+		   (##sys#check-exact to 'xsubstring)
 		   (values to start end)))
 ;	       (let ((slen (string-length (check-arg string? s xsubstring))))
 	       (let ((slen (string-length s)))
@@ -1812,7 +1825,7 @@
     (let ((slen   (- end start))
 	  (anslen (- to  from)))
       (cond ((zero? anslen) "")
-	    ((zero? slen) (error "Cannot replicate empty (sub)string"
+	    ((zero? slen) (##sys#error 'xsubstring "Cannot replicate empty (sub)string"
 				  xsubstring s from to start end))
 
 	    ((= 1 slen)		; Fast path for 1-char replication.
@@ -1836,17 +1849,19 @@
 ;;; This operation is not defined if (EQ? TARGET S) -- you cannot copy
 ;;; a string on top of itself.
 
+(define ##srfi13#string-fill! string-fill!) ; or we use std-binding.
+
 (define (string-xcopy! target tstart s sfrom . maybe-sto+start+end)
 ;  (check-arg (lambda (val) (and (integer? val) (exact? val)))
 ;	     sfrom string-xcopy!)
-  (##sys#check-exact sfrom)
+  (##sys#check-exact sfrom 'string-xcopy!)
   (receive (sto start end)
            (if (pair? maybe-sto+start+end)
 	       (let-string-start+end (start end) string-xcopy! s (cdr maybe-sto+start+end)
 		 (let ((sto (car maybe-sto+start+end)))
 ;		   (check-arg (lambda (val) (and (integer? val) (exact? val)))
 ;			      sto string-xcopy!)
-		   (##sys#check-exact sto)
+		   (##sys#check-exact sto 'string-xcopy!)
 		   (values sto start end)))
 	       (let ((slen (string-length s)))
 		 (values (+ sfrom slen) 0 slen)))
@@ -1856,12 +1871,12 @@
 	   (slen (- end start)))
       (check-substring-spec string-xcopy! target tstart tend)
       (cond ((zero? tocopy))
-	    ((zero? slen) (error "Cannot replicate empty (sub)string"
+	    ((zero? slen) (##sys#error 'string-xcopy! "Cannot replicate empty (sub)string"
 				 string-xcopy!
 				 target tstart s sfrom sto start end))
 
 	    ((= 1 slen)			; Fast path for 1-char replication.
-	     (string-fill! target (string-ref s start) tstart tend))
+	     (##srfi13#string-fill! target (string-ref s start) tstart tend))
 
 	    ;; Selected text falls entirely within one span.
 	    ((= (floor (/ sfrom slen)) (floor (/ sto slen)))
@@ -1932,16 +1947,16 @@
 		((suffix)
 		 (cons (car strings) (buildit (cdr strings) (list delim))))
 
-		(else (error "Illegal join grammar"
+		(else (##sys#error 'string-join "Illegal join grammar"
 			     grammar string-join)))))
 
 	     ((not (null? strings))
-	      (error "STRINGS parameter not list." strings string-join))
+	      (##sys#error 'string-join "STRINGS parameter not list." strings string-join))
 
 	     ;; STRINGS is ()
 
 	     ((eq? grammar 'strict-infix)
-	      (error "Empty list cannot be joined with STRICT-INFIX grammar."
+	      (##sys#error 'string-join "Empty list cannot be joined with STRICT-INFIX grammar."
 		     string-join))
 
 	     (else "")))))		; Special-cased for infix grammar.
