@@ -4,25 +4,33 @@
 (declare
   (unit srfi-14)
   (fixnum)
-  (interrupts-disabled)
-  (no-bound-checks)
+  (disable-interrupts)
   (standard-bindings)
-  (extended-bindings) )
+  (extended-bindings) 
+  (hide %char-set:s/check %string-iter %char-set-diff+intersection! %char->latin %latin1->char
+	%ucs-range->char-set! %string->char-set! %list->char-set! %set-char-set! %char-set-unfold!
+	%char-set-algebra %char-set-cursor-next %char-set-filter! %set-char-set) )
+
+(cond-expand 
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks) ) ] )
 
 (cond-expand
  [unsafe
   (eval-when (compile)
-    (define-macro (##sys#check-structure x y) '(##core#undefined))
-    (define-macro (##sys#check-range x y z) '(##core#undefined))
-    (define-macro (##sys#check-pair x) '(##core#undefined))
-    (define-macro (##sys#check-list x) '(##core#undefined))
-    (define-macro (##sys#check-symbol x) '(##core#undefined))
-    (define-macro (##sys#check-string x) '(##core#undefined))
-    (define-macro (##sys#check-char x) '(##core#undefined))
-    (define-macro (##sys#check-exact x) '(##core#undefined))
-    (define-macro (##sys#check-port x) '(##core#undefined))
-    (define-macro (##sys#check-number x) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector x) '(##core#undefined)) ) ]
+    (define-macro (##sys#check-structure . _) '(##core#undefined))
+    (define-macro (##sys#check-range . _) '(##core#undefined))
+    (define-macro (##sys#check-pair . _) '(##core#undefined))
+    (define-macro (##sys#check-list . _) '(##core#undefined))
+    (define-macro (##sys#check-symbol . _) '(##core#undefined))
+    (define-macro (##sys#check-string . _) '(##core#undefined))
+    (define-macro (##sys#check-char . _) '(##core#undefined))
+    (define-macro (##sys#check-exact . _) '(##core#undefined))
+    (define-macro (##sys#check-port . _) '(##core#undefined))
+    (define-macro (##sys#check-number . _) '(##core#undefined))
+    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
  [else] )
 
 (register-feature! 'srfi-14)
@@ -130,8 +138,8 @@
 	    (tail (cdr maybe-base)))
 	(if (null? tail)
 	    (if (char-set? bcs) (%string-copy (char-set:s bcs))
-		(error "BASE-CS parameter not a char-set" proc bcs))
-	    (error "Expected final base char set -- too many parameters"
+		(##sys#error "BASE-CS parameter not a char-set" proc bcs))
+	    (##sys#error "Expected final base char set -- too many parameters"
 		   proc maybe-base)))
       (make-string 256 (%latin1->char 0))))
 
@@ -142,7 +150,7 @@
 (define (%char-set:s/check cs proc)
   (let lp ((cs cs))
     (if (char-set? cs) (char-set:s cs)
-	(lp (error "Not a char-set" cs proc)))))
+	(lp (##sys#error "Not a char-set" cs proc)))))
 
 
 
@@ -229,7 +237,7 @@
 (define (char-set-hash cs . maybe-bound)
   (let ((bound (:optional maybe-bound 4194304)))
     (if (zero? bound) (set! bound 4194304))
-    (##sys#check-exact bound)
+    (##sys#check-exact bound 'char-set-hash)
     (let* ((s (%char-set:s/check cs char-set-hash))
 	   ;; Compute a 111...1 mask that will cover BOUND-1:
 	   (mask (let lp ((i #x10000))	; Let's skip first 16 iterations, eh?
@@ -305,8 +313,8 @@
 (define (char-set-ref cset cursor) (%latin1->char cursor))
 
 (define (char-set-cursor-next cset cursor)
-  (check-arg (lambda (i) (and (integer? i) (exact? i) (<= 0 i 255))) cursor
-	     char-set-cursor-next)
+;  (check-arg (lambda (i) (and (integer? i) (exact? i) (<= 0 i 255))) cursor
+;	     char-set-cursor-next)
   (%char-set-cursor-next cset cursor char-set-cursor-next))
 
 (define (%char-set-cursor-next cset cursor proc)	; Internal
@@ -456,7 +464,7 @@
   (##sys#check-exact upper)
 
   (if (and (< lower upper) (< 256 upper) error?)
-      (error "Requested UCS range contains unavailable characters -- this implementation only supports Latin-1"
+      (##sys#error "Requested UCS range contains unavailable characters -- this implementation only supports Latin-1"
 	     proc lower upper))
 
   (let lp ((i (- (min upper 256) 1)))
@@ -478,7 +486,7 @@
 ;;; -- predicate -> char-set
 
 (define (%char-set-filter! pred ds bs proc)
-  (check-arg procedure? pred proc)
+;  (check-arg procedure? pred proc)
   (let lp ((i 255))
     (cond ((>= i 0)
 	   (if (and (si=1? ds i) (pred (%latin1->char i)))
@@ -507,7 +515,7 @@
   (cond ((char-set? x) x)
 	((string? x) (string->char-set x))
 	((char? x) (char-set x))
-	(else (error "->char-set: Not a charset, string or char." x))))
+	(else (##sys#error '->char-set "Not a charset, string or char." x))))
 
 
 
