@@ -1,6 +1,6 @@
 ;;;; srfi-18.scm - Simple thread unit - felix
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -35,36 +35,45 @@
 
 (declare
  (unit srfi-18)
- (interrupts-disabled)
- (standard-bindings)
- (extended-bindings)
- (no-bound-checks)
- (bound-to-procedure
-  ##sys#add-to-ready-queue ##sys#remove-from-ready-queue
-  ##sys#schedule ##sys#make-thread
-  ##sys#update-thread-state-buffer ##sys#restore-thread-state
-  ##sys#check-number ##sys#error ##sys#signal-hook ##sys#signal
-  ##sys#current-exception-handler ##sys#abandon-mutexes ##sys#check-structure ##sys#structure? ##sys#make-mutex
-  ##sys#delq ##sys#compute-time-limit ##sys#fudge) )
+ (uses scheduler)
+ (disable-interrupts)
+ (usual-integrations)
+ (hide ##sys#compute-time-limit ##sys#delq) )
+
+(cond-expand
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks)
+    (bound-to-procedure
+     ##sys#add-to-ready-queue
+     ##sys#schedule ##sys#make-thread
+     ##sys#check-number ##sys#error ##sys#signal-hook ##sys#signal
+     ##sys#current-exception-handler ##sys#abandon-mutexes ##sys#check-structure ##sys#structure? ##sys#make-mutex
+     ##sys#delq ##sys#compute-time-limit ##sys#fudge) ) ] )
 
 (cond-expand
  [unsafe
   (eval-when (compile)
-    (define-macro (##sys#check-structure x y) '(##core#undefined))
-    (define-macro (##sys#check-range x y z) '(##core#undefined))
-    (define-macro (##sys#check-pair x) '(##core#undefined))
-    (define-macro (##sys#check-list x) '(##core#undefined))
-    (define-macro (##sys#check-symbol x) '(##core#undefined))
-    (define-macro (##sys#check-string x) '(##core#undefined))
-    (define-macro (##sys#check-char x) '(##core#undefined))
-    (define-macro (##sys#check-exact x) '(##core#undefined))
-    (define-macro (##sys#check-port x) '(##core#undefined))
-    (define-macro (##sys#check-number x) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector x) '(##core#undefined)) ) ]
+    (define-macro (##sys#check-structure . _) '(##core#undefined))
+    (define-macro (##sys#check-range . _) '(##core#undefined))
+    (define-macro (##sys#check-pair . _) '(##core#undefined))
+    (define-macro (##sys#check-list . _) '(##core#undefined))
+    (define-macro (##sys#check-symbol . _) '(##core#undefined))
+    (define-macro (##sys#check-string . _) '(##core#undefined))
+    (define-macro (##sys#check-char . _) '(##core#undefined))
+    (define-macro (##sys#check-exact . _) '(##core#undefined))
+    (define-macro (##sys#check-port . _) '(##core#undefined))
+    (define-macro (##sys#check-number . _) '(##core#undefined))
+    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
  [else] )
 
 (register-feature! 'srfi-18)
 
+(define-macro (dbg . args) #f)
+#;(define-macro (dbg . args)
+  `(print "DBG: " ,@args) )
+
 
 ;;; Helper routines:
 
@@ -79,23 +88,26 @@
     (lambda (tm)
       (and tm
 	   (cond [(##sys#structure? tm 'time) (##sys#slot tm 1)]
-		 [(number? tm) (fx+ (##sys#fudge 6) (inexact->exact (truncate (* tm 1000))))]
+		 [(number? tm) (fx+ (##sys#fudge 16) (inexact->exact (truncate (* tm 1000))))]
 		 [else (##sys#signal-hook #:type-error "invalid timeout argument" tm)] ) ) ) ) )
 
 
 ;;; Time objects:
 
 (define (current-time) 
-  (##sys#make-structure 'time (##sys#fudge 6)) )
+  (let* ([s (##sys#fudge 20)]
+	 [ms (##sys#fudge 21)] )
+    (##sys#make-structure 'time (##sys#fudge 22) s ms) ) )
 
-(define time->seconds
-  (lambda (tm)
-    (##sys#check-structure tm 'time)
-    (/ (##sys#slot tm 1) 1000) ) )
+(define (time->seconds tm)
+  (##sys#check-structure tm 'time 'time->seconds)
+  (+ (##sys#slot tm 2) (/ (##sys#slot tm 3) 1000)) )
 
 (define (seconds->time n)
-  (##sys#check-number n)
-  (##sys#make-structure 'time (inexact->exact (truncate (* n 1000)))) )
+  (##sys#check-number n 'seconds->time)
+  (let* ([n2 (- n (##sys#fudge 23))]	; seconds since startup
+	 [n3 (inexact->exact (truncate (* n2 1000)))] )	; milliseconds since startup
+    (##sys#make-structure 'time n3 (inexact->exact (truncate n)) 0) ) )
 
 (define (time? x) (##sys#structure? x 'time))
 
@@ -121,8 +133,7 @@
        (memq 'uncaught-exception (##sys#slot x 1)) ) )
 
 (define uncaught-exception-reason
-  (let ([get-reason (condition-property-accessor 'uncaught-exception 'reason)])
-    (lambda (ux) (get-reason ux)) ) )
+  (condition-property-accessor 'uncaught-exception 'reason) )
 
 
 ;;; Threads:
@@ -141,43 +152,46 @@
 	 (lambda () 
 	   (let ((result (thunk)))
 	     (##sys#setslot thread 2 result)
-	     (##sys#setslot thread 3 'dead) 
-	     (##sys#abandon-mutexes thread)
-	     result) ) )
+	     (##sys#thread-kill! thread 'dead)
+	     (##sys#schedule) ) ) )
 	thread) ) ) )
 
 (define (thread? x) (##sys#structure? x 'thread))
 (define (current-thread) ##sys#current-thread)
 
+(define (thread-state thread)
+  (##sys#check-structure thread 'thread 'thread-state)
+  (##sys#slot thread 3) )
+
 (define (thread-specific thread)
-  (##sys#check-structure thread 'thread)
+  (##sys#check-structure thread 'thread 'thread-specific)
   (##sys#slot thread 10) )
 
 (define (thread-specific-set! thread x)
-  (##sys#check-structure thread 'thread)
+  (##sys#check-structure thread 'thread 'thread-specific-set!)
   (##sys#setslot thread 10 x) )
 
 (define (thread-quantum thread)
-  (##sys#check-structure thread 'thread)
+  (##sys#check-structure thread 'thread 'thread-quantum)
   (##sys#slot thread 9) )
 
 (define (thread-quantum-set! thread q)
-  (##sys#check-structure thread 'thread)
-  (##sys#check-exact q)
+  (##sys#check-structure thread 'thread 'thread-quantum-set!)
+  (##sys#check-exact q 'thread-quantum-set!)
   (##sys#setislot thread 9 (fxmax q 10)) )
 
 (define (thread-name x)
-  (##sys#check-structure x 'thread)
+  (##sys#check-structure x 'thread 'thread-name)
   (##sys#slot x 6) )
 
 (define thread-start!
-    (lambda (thread)
-      (unless (eq? 'created (##sys#slot thread 3))
-	(##sys#error "thread can not be started a second time" thread) )
-      (##sys#check-structure thread 'thread)
-      (##sys#setslot thread 3 'ready)
-      (##sys#add-to-ready-queue thread) 
-      thread) )
+  (lambda (thread)
+    (##sys#check-structure thread 'thread 'thread-start!)
+    (unless (eq? 'created (##sys#slot thread 3))
+      (##sys#error 'thread-start! "thread can not be started a second time" thread) )
+    (##sys#setslot thread 3 'ready)
+    (##sys#add-to-ready-queue thread) 
+    thread) )
 
 (define (thread-yield!)
   (##sys#call-with-current-continuation
@@ -187,66 +201,44 @@
        (##sys#schedule) ) ) ) )
 
 (define thread-join!
-    (lambda (thread . timeout)
-      (##sys#check-structure thread 'thread)
-      (let* ((limit (and (pair? timeout) (##sys#compute-time-limit (##sys#slot timeout 0))))
-	     (rest (and (pair? timeout) (##sys#slot timeout 1)))
-	     (tosupplied (and rest (pair? rest)))
-	     (toval (and tosupplied (##sys#slot rest 0))) )
-	(##sys#call-with-current-continuation
-	 (lambda (return)
-	   (let ([ct ##sys#current-thread])
-	     (##sys#setslot ct 1 (lambda () (return (##sys#slot thread 2))))
-	     (##sys#setslot ct 3 'blocked)
-	     (##sys#setslot 
-	      ct 4
-	      (if limit
-		  (lambda () 
-		    (case (##sys#slot thread 3)
-		      [(dead) #t]
-		      [(terminated)
-		       (##sys#setslot
-			ct 1
-			(lambda ()
-			  (return 
-			   (##sys#signal
-			    (##sys#make-structure 'condition '(uncaught-exception) (list 'reason (##sys#slot thread 7)) ) ) ) ) )
-		       #t]
-		      [else
-		       (and (fx> (##sys#fudge 6) limit)
-			    (begin
-			      (##sys#setslot 
-			       ct 1 
-			       (lambda ()
-				 (return
-				  (if tosupplied
-				      toval
-				      (##sys#signal (##sys#make-structure 'condition '(join-timeout-exception) '())) ) ) ) )
-			      #t) ) ] ) )
-		  (lambda () 
-		    (case (##sys#slot thread 3)
-		      [(dead) #t]
-		      [(terminated)
-		       (##sys#setslot
-			ct 1
-			(lambda ()
-			  (return 
-			   (##sys#signal
-			    (##sys#make-structure 'condition '(uncaught-exception) (list 'reason (##sys#slot thread 7)) ) ) ) ) )
-		       #t]
-		      [else #f] ) ) ) )
-	     (##sys#schedule) ) ) ) ) ) )
-
+  (lambda (thread . timeout)
+    (##sys#check-structure thread 'thread 'thread-join!)
+    (let* ((limit (and (pair? timeout) (##sys#compute-time-limit (##sys#slot timeout 0))))
+	   (rest (and (pair? timeout) (##sys#slot timeout 1)))
+	   (tosupplied (and rest (pair? rest)))
+	   (toval (and tosupplied (##sys#slot rest 0))) )
+      (##sys#call-with-current-continuation
+       (lambda (return)
+	 (let ([ct ##sys#current-thread])
+	   (when limit (##sys#thread-block-for-timeout! ct limit))
+	   (##sys#setslot
+	    ct 1
+	    (lambda ()
+	      (case (##sys#slot thread 3)
+		[(dead) (return (##sys#slot thread 2))]
+		[(terminated)
+		 (return 
+		  (##sys#signal
+		   (##sys#make-structure 'condition '(uncaught-exception) (list 'reason (##sys#slot thread 7)) ) ) ) ]
+		[else
+		 (return
+		  (if tosupplied
+		      toval
+		      (##sys#signal (##sys#make-structure 'condition '(join-timeout-exception) '())) ) ) ] ) ) )
+	   (##sys#thread-block-for-termination! ct thread) 
+	   (##sys#schedule) ) ) ) ) ) )
+	   
 (define (thread-terminate! thread)
-  (##sys#check-structure thread 'thread)
-  (##sys#setslot thread 3 'terminated)
+  (##sys#check-structure thread 'thread 'thread-terminate!)
+  (when (eq? thread ##sys#primordial-thread)
+    ((##sys#exit-handler)) )
   (##sys#setislot thread 2 (##core#undefined))
   (##sys#setslot thread 7 (##sys#make-structure 'condition '(terminated-thread-exception) '()))
-  (##sys#abandon-mutexes thread)
+  (##sys#thread-kill! thread 'terminated)
   (when (eq? thread ##sys#current-thread) (##sys#schedule)) )
 
 (define (thread-suspend! thread)
-  (##sys#check-structure thread 'thread)
+  (##sys#check-structure thread 'thread 'thread-suspend!)
   (##sys#setslot thread 3 'suspended)
   (when (eq? thread ##sys#current-thread)
     (##sys#call-with-current-continuation
@@ -255,116 +247,110 @@
        (##sys#schedule) ) ) ) )
 
 (define (thread-resume! thread)
-  (##sys#check-structure thread 'thread)
+  (##sys#check-structure thread 'thread 'thread-resume!)
   (when (eq? (##sys#slot thread 3) 'suspended)
     (##sys#setslot thread 3 'ready)
     (##sys#add-to-ready-queue thread) ) )
 
 (define thread-sleep!
   (lambda (tm)
-    (unless tm (##sys#signal-hook #:tpe-error "invalid timeout argument" tm))
+    (unless tm (##sys#signal-hook #:type-error 'thread-sleep! "invalid timeout argument" tm))
     (##sys#call-with-current-continuation
      (lambda (return)
        (let ([limit (##sys#compute-time-limit tm)]
 	     [ct ##sys#current-thread] )
 	 (##sys#setslot ct 1 (lambda () (return (##core#undefined))))
-	 (##sys#setslot ct 3 'blocked)
-	 (##sys#setslot ct 4 (lambda () (fx>= (##sys#fudge 6) limit)))
+	 (##sys#thread-block-for-timeout! ct limit)
 	 (##sys#schedule) ) ) ) ) )
 
 
 ;;; Mutexes:
 
-(define (mutex? x) (##sys#structure? mutex 'mutex))
+(define (mutex? x) (##sys#structure? x 'mutex))
 
 (define make-mutex
   (let ((gensym gensym))
-    (lambda owner-and-id
-      (let* ((owner (if (pair? owner-and-id) owner-and-id #f))
-	     (ownert (if owner (##sys#slot owner 0) ##sys#current-thread))
-	     (rest (and owner (##sys#slot owner 1)))
-	     (id (if (and owner (pair? rest)) (##sys#slot rest 0) (gensym 'mutex)))
-	     (m (##sys#make-mutex id ownert)) )
-	(if ownert
-	    (begin
-	      (##sys#check-structure ownert 'thread)
-	      (##sys#setslot ownert 8 (cons m (##sys#slot ownert 8))) ) ) 
+    (lambda id
+      (let* ((id (if (pair? id) (car id) (gensym 'mutex)))
+	     (m (##sys#make-mutex id ##sys#current-thread)) )
 	m) ) ) )
 
 (define (mutex-name x)
-  (##sys#check-structure x 'mutex) 
+  (##sys#check-structure x 'mutex 'mutex-name) 
   (##sys#slot x 1) )
 
-(define (mutex-owner mutex)
-  (##sys#check-structure mutex 'mutex)
-  (##sys#slot mutex 2) )
-
 (define (mutex-specific mutex)
-  (##sys#check-structure mutex 'mutex)
+  (##sys#check-structure mutex 'mutex 'mutex-specific)
   (##sys#slot mutex 6) )
 
 (define (mutex-specific-set! mutex x)
-  (##sys#check-structure mutex 'mutex)
+  (##sys#check-structure mutex 'mutex 'mutex-specific-set!)
   (##sys#setslot mutex 6 x) )
 
 (define (mutex-state mutex)
-  (##sys#check-structure mutex 'mutex)
+  (##sys#check-structure mutex 'mutex 'mutex-state)
   (cond [(##sys#slot mutex 5) (or (##sys#slot mutex 2) 'not-owned)]
 	[(##sys#slot mutex 4) 'abandoned]
 	[else 'not-abandoned] ) )
 
 (define mutex-lock! 
   (lambda (mutex . ms-and-t)
-    (##sys#check-structure mutex 'mutex)
-    (let ([limit (and (pair? ms-and-t) (##sys#compute-time-limit (car ms-and-t)))]
-	  [thread (and (fx> (length ms-and-t) 1) (cadr ms-and-t))] )
-      (when thread (##sys#check-structure thread 'thread))
+    (##sys#check-structure mutex 'mutex 'mutex-lock!)
+    (let* ([limitsup (pair? ms-and-t)]
+	   [limit (and limitsup (##sys#compute-time-limit (car ms-and-t)))]
+	   [threadsup (fx> (length ms-and-t) 1)]
+	   [thread (and threadsup (cadr ms-and-t))] 
+	   [abd (##sys#slot mutex 4)] )
+      (when thread (##sys#check-structure thread 'thread 'mutex-lock!))
       (##sys#call-with-current-continuation
        (lambda (return)
-	 (let ([result
-		(let ([ct ##sys#current-thread]
-		      [success #t] )
-
-		  (define (switch)
-		    (##sys#setslot ct 1 (lambda () (return success)))
-		    (##sys#setslot mutex 3 (##sys#append (##sys#slot mutex 3) (list ct)))
-		    (##sys#schedule) )
-
-		  (cond [(not (##sys#slot mutex 5)) #t]
-			[limit
-			 (##sys#setslot ct 3 'blocked)
-			 (##sys#setslot 
-			  ct 4 
-			  (lambda ()
-			    (and (fx>= (##sys#fudge 6) limit)
-				 (begin 
-				   (##sys#setslot mutex 3 (##sys#delq ct (##sys#slot mutex 3)))
-				   (##sys#setslot mutex 2 thread)
-				   (set! success #f)
-				   #t) ) ) )
-			 (switch) ]
-			[else
-			 (##sys#setslot ct 3 'sleeping)
-			 (switch) ] ) ) ] )
-
-	   (cond [(##sys#slot mutex 4)
-		  (##sys#signal (##sys#make-structure 'condition '(abandoned-mutex-exception) '())) ]
-		 [result
-		  (let ([t (or thread ##sys#current-thread)])
-		    (cond [(eq? 'terminated (##sys#slot t 3)) (##sys#setislot mutex 4 #t)]
-			  [else
-			   (##sys#setislot mutex 5 #t)
-			   (##sys#setslot mutex 2 thread) ] ) )
-		  result]
-		 [else #f] ) ) ) ) ) ) )
+	 (let ([ct ##sys#current-thread])
+	   (define (switch)
+	     (##sys#setslot mutex 3 (##sys#append (##sys#slot mutex 3) (list ct)))
+	     (##sys#schedule) )
+	   (define (check)
+	     (when abd
+	       (return (##sys#signal (##sys#make-structure 'condition '(abandoned-mutex-exception) '()))) ) )
+	   (dbg ct ": locking " mutex)
+	   (cond [(not (##sys#slot mutex 5))
+		  (if (and threadsup (not thread))
+		      (begin
+			(##sys#setislot mutex 2 #f)
+			(##sys#setislot mutex 5 #t) )
+		      (let* ([t (or thread ct)]
+			     [ts (##sys#slot t 3)] )
+			(if (or (eq? 'terminated ts) (eq? 'dead ts))
+			    (##sys#setislot mutex 4 #t)
+			    (begin
+			      (##sys#setislot mutex 5 #t)
+			      (##sys#setslot t 8 (cons mutex (##sys#slot t 8)))
+			      (##sys#setslot mutex 2 t) ) ) ) )
+		  (check)
+		  (return #t) ]
+		 [limit
+		  (check)
+		  (##sys#setslot
+		   ct 1 
+		   (lambda ()
+		     (##sys#setslot mutex 3 (##sys#delq ct (##sys#slot mutex 3)))
+		     (##sys#setslot ##sys#current-thread 8 (cons mutex (##sys#slot ##sys#current-thread 8)))
+		     (##sys#setslot mutex 2 thread)
+		     #f) )
+		  (##sys#thread-block-for-timeout! ct limit)
+		  (switch) ]
+		 [else
+		  (##sys#setslot ct 3 'sleeping)
+		  (##sys#setslot ct 1 (lambda () (return #t)))
+		  (switch) ] ) ) ) ) ) ) )
 
 (define mutex-unlock!
   (lambda (mutex . cvar-and-to)
-    (##sys#check-structure mutex 'mutex)
+    (##sys#check-structure mutex 'mutex 'mutex-unlock!)
     (let ([ct ##sys#current-thread]
 	  [cvar (and (pair? cvar-and-to) (car cvar-and-to))]
 	  [timeout (and (fx> (length cvar-and-to) 1) (cadr cvar-and-to))] )
-      (when cvar (##sys#check-structure cvar 'condition-variable))
+      (dbg ct ": unlocking " mutex)
+      (when cvar (##sys#check-structure cvar 'condition-variable 'mutex-unlock!))
       (##sys#call-with-current-continuation
        (lambda (return)
 	 (let ([waiting (##sys#slot mutex 3)]
@@ -377,27 +363,25 @@
 	   (when cvar
 	     (##sys#setslot cvar 2 (##sys#append (##sys#slot cvar 2) (##sys#list ct)))
 	     (cond [limit
-		    (##sys#setslot ct 3 'blocked)
 		    (##sys#setslot 
-		     ct 4
+		     ct 1
 		     (lambda () 
-		       (and (fx> (##sys#fudge 6) limit)
-			    (begin 
-			      (##sys#setslot cvar 2 (##sys#delq ct (##sys#slot cvar 2)))
-			      (set! result #f)
-			      #t) ) ) ) ]
-		   [else (##sys#setslot ct 3 'sleeping)] ) )
+		       (##sys#setslot cvar 2 (##sys#delq ct (##sys#slot cvar 2)))
+		       (return #f) ) )
+		    (##sys#thread-block-for-timeout! ct limit) ]
+		   [else 
+		    (##sys#setslot ct 3 'sleeping)] ) )
 	   (unless (null? waiting)
 	     (let* ([wt (##sys#slot waiting 0)]
 		    [wts (##sys#slot wt 3)] )
 	       (##sys#setslot mutex 3 (##sys#slot waiting 1))
-	       (when (memq wts '(blocked sleeping))
+	       (when (or (eq? wts 'blocked) (eq? wts 'sleeping))
 		 (##sys#setslot mutex 2 wt)
-		 (##sys#setslot wt 3 'ready)
 		 (##sys#setslot wt 8 (cons mutex (##sys#slot wt 8)))
 		 (when (eq? wts 'sleeping) (##sys#add-to-ready-queue wt) ) ) ) )
 	   (##sys#schedule) ) ) ) ) ) )
 
+
 ;;; Condition variables:
 
 (define make-condition-variable
@@ -415,29 +399,45 @@
   (##sys#structure? x 'condition-variable) )
 
 (define (condition-variable-specific cv)
-  (##sys#check-structure cv 'condition-variable)
+  (##sys#check-structure cv 'condition-variable 'condition-variable-specific)
   (##sys#slot cv 3) )
 
 (define (condition-variable-specific-set! cv x)
-  (##sys#check-structure cv 'condition-variable)
+  (##sys#check-structure cv 'condition-variable 'condition-variable-specific-set!)
   (##sys#setslot cv 3 x) )
 
 (define (condition-variable-signal! cvar)
-  (##sys#check-structure cvar 'condition-variable)
+  (##sys#check-structure cvar 'condition-variable 'condition-variable-signal!)
+  (dbg "signalling " cvar)
   (let ([ts (##sys#slot cvar 2)])
     (unless (null? ts)
-      (let ([t0 (##sys#slot ts 0)])
+      (let* ([t0 (##sys#slot ts 0)]
+	     [t0s (##sys#slot t0 3)] )
 	(##sys#setslot cvar 2 (##sys#slot ts 1))
-	(when (memq (##sys#slot t0 3) '(blocked sleeping))
-	  (##sys#setslot t0 3 'ready)
-	  (##sys#add-to-ready-queue t0) ) ) ) ) )
+	(when (or (eq? t0s 'blocked) (eq? t0s 'sleeping))
+	  (##sys#thread-unblock! t0) ) ) ) ) )
 
 (define (condition-variable-broadcast! cvar)
-  (##sys#check-structure cvar 'condition-variable)
+  (##sys#check-structure cvar 'condition-variable 'condition-variable-broadcast!)
+  (dbg "broadcasting " cvar)
   (##sys#for-each
    (lambda (ti)
-     (when (memq (##sys#slot ti 3) '(blocked sleeping))
-       (##sys#setslot ti 3 'ready)
-       (##sys#add-to-ready-queue ti) ) )
+     (let ([tis (##sys#slot ti 3)])
+       (when (or (eq? tis 'blocked) (eq? tis 'sleeping))
+	 (##sys#thread-unblock! ti) ) ) )
    (##sys#slot cvar 2) )
   (##sys#setislot cvar 2 '()) )
+
+
+;;; Change continuation of thread to signal an exception:
+
+(define (thread-deliver-signal! thread exn)
+  (##sys#check-structure thread 'thread 'thread-deliver-exception!)
+  (if (eq? thread ##sys#current-thread)
+      (##sys#signal exn)
+      (let ([old (##sys#slot thread 1)])
+	(##sys#setslot
+	 thread 1
+	 (lambda ()
+	   (##sys#signal exn)
+	   (old) ) ) ) ) )
