@@ -1,6 +1,6 @@
 ;;;; support.scm - Miscellaneous support code for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -37,41 +37,42 @@
 
 
 #{compiler
-  build-information compiler-arguments process-command-line
-  default-analysis-database-size default-standard-bindings default-extended-bindings side-effecting-standard-bindings
+  build-information compiler-arguments process-command-line dump-nodes
+  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home optimization-iterations compiler-cleanup-hook decompose-lambda-list
-  foreign-type-table-size file-io-only nonwinding-call/cc
-  unit-name insert-timer-checks used-units inlining debug-info-index debug-info-vector-name
+  foreign-type-table-size file-io-only
+  unit-name insert-timer-checks used-units source-filename pending-canonicalizations
   foreign-declarations block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home 
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
   dependency-list broken-constant-nodes inline-substitutions-enabled
-  always-bound-to-procedure block-variable-literal
-  direct-call-ids foreign-type-table first-analysis expand-debug-call
+  always-bound-to-procedure block-variable-literal copy-node!
+  direct-call-ids foreign-type-table first-analysis scan-sharp-greater-string enable-sharp-greater-read-syntax
   expand-profile-lambda profile-lambda-list profile-lambda-index profile-info-vector-name
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
+  reorganize-recursive-bindings substitution-table simplify-named-call
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! expand-debug-lambda expand-debug-assignment debug-lambda-list debug-variable-list
+  transform-direct-lambdas! finish-foreign-result compressable-literal
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
-  string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
+  string->c-identifier c-ify-string words words->bytes check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database write-dependency-list
+  topological-sort print-version print-usage initialize-analysis-database estimate-foreign-result-location-size
+  real-name real-name-table set-real-name! real-name2 display-real-name-table display-line-number-database
   product copyright compiler-features default-declarations units-used-by-default words-per-flonum
-  foreign-string-result-reserve parameter-limit default-output-filename eq-inline-operator optimizable-rest-argument-operators
+  foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  default-optimization-iterations chop-separator chop-extension current-source-file source-file-changed show-source-file
+  default-optimization-iterations chop-separator chop-extension follow-without-loop
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion final-foreign-type
+  foreign-argument-conversion foreign-result-conversion final-foreign-type debugging
   make-random-name foreign-type-convert-result foreign-type-convert-argument}
 
 
@@ -85,8 +86,6 @@
 
 (define debugging-chicken '())
 (define warnings-enabled #t)
-(define current-source-file #f)
-(define source-file-changed #t)
 
 (define (bomb . msg-and-args)
   (if (pair? msg-and-args)
@@ -105,39 +104,30 @@
 	 (flush-output)
 	 #t) ) )
 
-(define (show-source-file)
-  (when (and current-source-file source-file-changed)
-    (fprintf (current-error-port) "~A:~%" current-source-file)
-    (set! source-file-changed #f) ) )
-
 (define (warning msg . args)
   (when warnings-enabled
     (let ((out (current-error-port)))
-      (show-source-file)
       (apply fprintf out (string-append "Warning: " msg) args)
       (newline out) ) ) )
 
-(define quit-handler
-  (make-parameter (lambda (code) (exit code))) )
-
 (define (quit msg . args)
   (let ([out (current-error-port)])
-    (show-source-file)
     (apply fprintf out (string-append "Error: " msg) args)
     (newline out)
-    ((quit-handler) 1) ) )
+    (exit 1) ) )
 
 (set! ##sys#syntax-error-hook
   (lambda (msg . args)
     (let ([out (current-error-port)])
-      (show-source-file)
       (fprintf out "Syntax error: ~a~%" msg) 
       (for-each (lambda (x) (write x out) (newline out)) args)
       (exit 70) ) ) )
 
-(define (check-signature args llist)
+(define (check-signature var args llist)
   (define (err)
-    (quit "Arguments to inlined call do not match parameter-list ~S" llist) )
+    (quit "Arguments to inlined call of `~A' do not match parameter-list ~A" 
+	  (real-name var)
+	  (map real-name (cdr llist)) ) )
   (let loop ([as args] [ll llist])
     (cond [(null? ll) (unless (null? as) (err))]
 	  [(symbol? ll)]
@@ -171,16 +161,7 @@
     (cond ((or (zero? n) (null? vars)) (or rest '()))
           (else (cons (car vars) (loop (cdr vars) (sub1 n)))) ) ) )
 
-(define (string->c-identifier str)
-  (let ((lst (string->list str)))
-    (list->string
-     (map (lambda (c) 
-	    (if (and (not (char-numeric? c)) (not (char-alphabetic? c)))
-		#\_
-		c) )
-	  (if (not (char-alphabetic? (car lst)))
-	      (cons #\_ lst)
-	      lst) ) ) ) )
+(define string->c-identifier ##sys#string->c-identifier)
 
 (define (c-ify-string str)
   (list->string
@@ -191,7 +172,7 @@
 	  '(#\")
 	  (let* ((c (car chars))
 		 (code (char->integer c)) )
-	    (if (or (< code 32) (memq c '(#\" #\' #\\)))
+	    (if (or (< code 32) (memq c '(#\" #\' #\\ #\?)))
 		(append '(#\\)
 			(cond ((< code 8) '(#\0 #\0))
 			      ((< code 64) '(#\0))
@@ -201,13 +182,15 @@
 		(cons c (loop (cdr chars))) ) ) ) ) ) ) )
 
 (eval-when (load)
-  (define words (foreign-lambda int "C_bytestowords" int)) )
+  (define words (foreign-lambda int "C_bytestowords" int)) 
+  (define words->bytes (foreign-lambda int "C_wordstobytes" int)) )
 
 (eval-when (eval)
-  (define words
+  (define (words n)
     (let ([wordsize (##sys#fudge 7)])
-      (lambda (n)
-	(+ (quotient n wordsize) (if (zero? (modulo n wordsize)) 0 1)) ) ) ) )
+      (+ (quotient n wordsize) (if (zero? (modulo n wordsize)) 0 1)) ) )
+  (define (words->bytes n)
+    (* n (##sys#fudge 7)) ) )
 
 (define (check-and-open-input-file fname . line)
   (cond [(string=? fname "-") (current-input-port)]
@@ -228,6 +211,12 @@
 	     (list (cadr xs) (car xs))
 	     (list (fold (cdr xs)) (car xs)) ) ) ) ) )
 
+(define (follow-without-loop seed proc abort)
+  (let loop ([x seed] [done '()])
+    (if (member x done)
+	(abort)
+	(proc x (lambda (x2) (loop x2 (cons x done)))) ) ) )
+
 
 ;;; Predicates on expressions and literals:
 
@@ -251,6 +240,18 @@
       (char? x)
       (boolean? x) ) )
 
+(define (compressable-literal lit t)
+  (let* ([count 0]
+	 [f (let rec ([x lit])
+	      (set! count (add1 count))
+	      (cond [(or (number? x) (char? x) (string? x) (boolean? x) (null? x) (symbol? x))] ; 1
+		    [(pair? x)		; car + cdr
+		     (set! count (sub1 count))
+		     (and (rec (car x)) (rec (cdr x))) ]
+		    [(vector? x) (every rec (vector->list x))] ; 1 + elements
+		    [else #f] ) ) ] )
+    (and f (> count t) count) ) )
+
 
 ;;; Expression manipulation:
 
@@ -258,7 +259,11 @@
   (let loop ((xs body))
     (cond ((null? xs) '(##core#undefined))
 	  ((null? (cdr xs)) (car xs))
-	  ((or (equal? (car xs) '(##core#undefined)) (constant? (car xs))) (loop (cdr xs)))
+	  ((let ([h (car xs)])
+	     (or (equal? h '(##core#undefined))
+		 (constant? h) 
+		 (equal? h '(##sys#void)) ) )
+	   (loop (cdr xs)) )
 	  (else `(let ((,(gensym 't) ,(car xs)))
 		   ,(loop (cdr xs))) ) ) ) )
 
@@ -273,12 +278,13 @@
 	       (set! mlist (alist-cons var c mlist))
 	       var)
 	     x) ]
-	[`(let ((vars vals) ...) . body)
+	[`(let ((,vars ,vals) ...) . ,body)
 	 `(let ,(map (lambda (var val) (list var (walk val))) vars vals) ,@(map walk body)) ]
 	[(op . args)
 	 (case op
 	   [(##core#include ##core#declare ##core#immutable ##core#undefined ##core#primitive ##core#inline_ref) x]
-	   [(set! lambda ##core#inline ##core#inline_allocate ##core#inline_update)
+	   [(##core#set! set! lambda ##core#inline ##core#inline_allocate ##core#inline_update ##core#inline_loc_ref
+			 ##core#inline_loc_update)
 	    (cons* op (first args) (map walk (cdr args))) ]
 	   [(if ##core#compiletimeonly ##core#compiletimetoo)
 	    (cons op (map walk args)) ]
@@ -287,55 +293,24 @@
     (let ([exp2 (walk exp)])
       (values exp2 mlist) ) ) )
 
-(define (string->expr str)
-  (parameterize ([error-handler
-		  (lambda (msg . args)
-		    (quit "can not parse expression: ~s [~a]~%" str msg) ) ] )
-    (let ([xs (with-input-from-string str (lambda () (unfold eof-object? values (lambda (x) (read)) (read))))])
-      (cond [(null? xs) '(##core#undefined)]
-	    [(null? (cdr xs)) (car xs)]
-	    [else `(begin ,@xs)] ) ) ) )
+(define string->expr
+  (let ([exn? (condition-predicate 'exn)]
+	[exn-msg (condition-property-accessor 'exn 'message)] )
+    (lambda (str)
+      (handle-exceptions ex
+	  (quit "can not parse expression: ~s [~a]~%" 
+		str
+		(if (exn? ex) 
+		    (exn-msg ex)
+		    (->string ex) ) ) 
+	(let ([xs (with-input-from-string str (lambda () (unfold eof-object? values (lambda (x) (read)) (read))))])
+	  (cond [(null? xs) '(##core#undefined)]
+		[(null? (cdr xs)) (car xs)]
+		[else `(begin ,@xs)] ) ) ) ) ) )
 
 (define decompose-lambda-list ##sys#decompose-lambda-list)
 
 
-;;; Debugging expression instrumentation:
-
-(define (expand-debug-lambda name llist ollist argc rest body obody)
-  (let* ([exit (gensym 'exit)]
-	 [args (gensym)]
-	 [source 
-	  (##sys#string->pbytevector
-	   (with-output-to-string
-	     (lambda () (pretty-print `(lambda ,ollist ,@obody))) ) ) ] 
-	 ;; #(name lambda-list source argc rest-flag break? exec?)
-	 [info (vector (or name anonymous-object-identifier) ollist source argc (if rest #t #f) #f #f)] 
-	 [index debug-info-index] )
-    (set! debug-lambda-list (alist-cons index info debug-lambda-list))
-    (set! debug-info-index (add1 index))
-    `(lambda ,args
-       (call-with-current-continuation
-	(lambda (,exit)
-	  (##sys#dynamic-wind
-	   (lambda () (##sys#push-debug-frame (##core#inline "C_slot" ,debug-info-vector-name ',index) ,exit) )
-	   (lambda () (apply (lambda ,llist ,body) (##sys#check-debug-entry ,args)))
-	   (lambda () (##sys#pop-debug-frame)) ) ) ) ) ) )
-
-(define (expand-debug-assignment var val oexp)
-  (let* ([source 
-	  (##sys#string->pbytevector
-	   (with-output-to-string (lambda () (pretty-print oexp))) ) ] 
-	 ;; #(name source watched?)
-	 [info (vector var source #f)]
-	 [index debug-info-index] )
-    (set! debug-info-index (add1 index))
-    (set! debug-variable-list (alist-cons index info debug-variable-list))
-    `(##sys#check-debug-assignment (##core#inline "C_slot" ,debug-info-vector-name ',index) ,val) ) )
-
-(define (expand-debug-call name exp)
-  `(##sys#debug-call ',name ,@exp) )
-
-
 ;;; Profiling instrumentation:
 
 (define (expand-profile-lambda name llist body)
@@ -408,6 +383,9 @@
 		[else (##sys#setslot plist 1 (alist-cons prop n (##sys#slot plist 1)))] ) )
 	(##sys#hash-table-set! db key (list (cons prop val)))) ) )
 
+
+;;; Line-number database management:
+
 (define (get-line exp)
   (get ##sys#line-number-database (car exp) exp) )
 
@@ -424,11 +402,17 @@
 	(let ([c (get db id 'contained-in)])
 	  (and c (loop c)) ) ) ) )
 
+(define (display-line-number-database)
+  (##sys#hash-table-for-each
+   (lambda (key val)
+     (when val (printf "~S ~S~%" key (map cdr val))) )
+   ##sys#line-number-database) )
+
 
 ;;; Display analysis database:
 
 (define display-analysis-database
-  (let ((names '((captured . cpt) (assigned . set) (boxed . box) (global . glo)
+  (let ((names '((captured . cpt) (assigned . set) (boxed . box) (global . glo) (assigned-locally . stl)
 		 (contractable . con) (standard-binding . stb) (foldable . fld) (simple . sim)
 		 (side-effecting . sef) (collapsable . col) (removable . rem) (constant . con)
 		 (undefined . und) (replacing . rpg) (unused . uud) (extended-binding . xtb)
@@ -436,23 +420,26 @@
     (lambda (db)
       (##sys#hash-table-for-each
        (lambda (sym plist)
-	 (let ((val #f)
-	       (csites '())
-	       (refs '()) )
+	 (let ([val #f]
+	       [pval #f]
+	       [csites '()]
+	       [refs '()] )
 	   (write sym)
 	   (let loop ((es plist))
 	     (if (pair? es)
 		 (begin
 		   (case (caar es)
-		     ((captured assigned boxed global contractable standard-binding foldable
+		     ((captured assigned boxed global contractable standard-binding foldable assigned-locally
 		       side-effecting collapsable removable undefined replacing unused simple
 		       has-unused-parameters extended-binding customizable constant boxed-rest)
 		      (printf "\t~a" (cdr (assq (caar es) names))) )
 		     ((unknown)
 		      (set! val 'unknown) )
 		     ((value)
-		      (if (not (eq? val 'unknown)) (set! val (cdar es))) )
-		     ((replacable home outermost contains contained-in use-expr closure-size rest-parameter
+		      (unless (eq? val 'unknown) (set! val (cdar es))) )
+		     ((potential-value)
+		      (set! pval (cdar es)) )
+		     ((replacable home contains contained-in use-expr closure-size rest-parameter
 		       o-r/access-count captured-variables explicit-rest)
 		      (printf "\t~a=~s" (caar es) (cdar es)) )
 		     ((references)
@@ -461,10 +448,11 @@
 		      (set! csites (cdar es)) )
 		     (else (bomb "Illegal property" (car es))) )
 		   (loop (cdr es)) ) ) )
-	   (if (and val (not (eq? val 'unknown)))
-	       (printf "\tval=~s" (cons (node-class val) (node-parameters val))) )
-	   (if (pair? refs) (printf "\trefs=~s" (length refs)))
-	   (if (pair? csites) (printf "\tcss=~s" (length csites)))
+	   (cond [(and val (not (eq? val 'unknown)))
+		  (printf "\tval=~s" (cons (node-class val) (node-parameters val))) ]
+		 [pval (printf "\tpval=~s" (cons (node-class val) (node-parameters val)))] )
+	   (when (pair? refs) (printf "\trefs=~s" (length refs)))
+	   (when (pair? csites) (printf "\tcss=~s" (length csites)))
 	   (newline) ) )
        db) ) ) )       
 
@@ -486,6 +474,7 @@
 	    ((not-pair? x) (bomb "bad expression"))
 	    ((symbol? (car x))
 	     (case (car x)
+	       ((##core#global-ref) (make-node '##core#global-ref (list (cadr x)) '()))
 	       ((if ##core#undefined) (make-node (car x) '() (map walk (cdr x))))
 	       ((quote)
 		(let ((c (cadr x)))
@@ -507,14 +496,18 @@
 	       ((lambda) (make-node 'lambda (list (cadr x)) (list (walk (caddr x)))))
 	       ((set! ##core#inline ##core#callunit ##core#primitive ##core#proc) 
 		(make-node (car x) (list (cadr x)) (map walk (cddr x))) )
+	       ((##core#set!) (make-node 'set! (list (cadr x)) (map walk (cddr x))))
 	       ((##core#foreign-callback-wrapper)
 		(let ([name (cadr (second x))])
 		  (make-node
 		   '##core#foreign-callback-wrapper
 		   (list name (cadr (third x)) (cadr (fourth x)) (cadr (fifth x)))
 		   (list (walk (sixth x))) ) ) )
-	       ((##core#inline_allocate ##core#inline_ref ##core#inline_update)
+	       ((##core#inline_allocate ##core#inline_ref ##core#inline_update
+					##core#inline_loc_ref ##core#inline_loc_update)
 		(make-node (first x) (second x) (map walk (cddr x))) )
+	       ((##core#app)
+		(make-node '##core#call '(#t) (map walk (cdr x))) )
 	       (else
 		(receive
 		    (name ln) (get-line-2 x)
@@ -525,7 +518,14 @@
 				#t]
 			       [else #f] )
 			 (if ln
-			     (string-append (##sys#symbol->qualified-string name) "@" (number->string ln))
+			     (let ([rn (real-name name)]
+				   [lns (number->string ln)] )
+			       (string-append 
+				source-filename ": " 
+				lns
+				(make-string (max 0 (- 4 (string-length lns))) #\space)
+				" "
+				(or rn (##sys#symbol->qualified-string name)) ) )
 			     (##sys#symbol->qualified-string name) ) )
 		   (map walk x) ) ) ) ) )
 	    (else (make-node '##core#call '(#f) (map walk x))) ) )
@@ -540,9 +540,11 @@
 	  (class (node-class n)) )
       (case class
 	((if ##core#closure ##core#box ##core#cond) (cons class (map walk subs)))
-	((##core#variable) (car params))
+	((##core#variable ##core#global-ref) (car params))
 	((quote) `(quote ,(car params)))
-	((let) `(let ((,(first params) ,(walk (first subs)))) ,(walk (second subs))))
+	((let)
+	 `(let ,(map list params (map walk (butlast subs)))
+	    ,(walk (last subs)) ) )
 	((##core#lambda) 
 	 (list (if (second params)
 		   'lambda
@@ -633,12 +635,15 @@
 ;;; Test nodes for certain properties:
 
 (define (expression-has-side-effects? node db)
-  (let walk ((n node))
-    (let ((subs (node-subexpressions n)))
+  (let walk ([n node])
+    (let ([subs (node-subexpressions n)])
       (case (node-class n)
-	((##core#variable quote ##core#undefined ##core#lambda ##core#proc) #f)
-	((if let) (any walk subs))
-	(else #t) ) ) ) )
+	[(##core#variable quote ##core#undefined ##core#proc ##core#global-ref) #f]
+	[(##core#lambda) 
+	 (let ([id (first (node-parameters n))])
+	   (find (lambda (fs) (eq? id (foreign-callback-stub-id fs))) foreign-callback-stubs) ) ]
+	[(if let) (any walk subs)]
+	[else #t] ) ) ) )
 
 (define (simple-lambda-node? node)
   (let* ([params (node-parameters node)]
@@ -695,23 +700,18 @@
 	    nprocs
 	    nglobs
 	    nsites
-	    entries
-	    (let ((len (vector-length db))
-		  (load 0) )
-	      (do ((i 0 (+ i 1)))
-		  ((>= i len) (/ load len))
-		(set! load (+ load (length (vector-ref db i)))) ) ) ) ) )
+	    entries) ) )
 
 (define (print-program-statistics db)
   (receive
-   (size osize kvars kprocs globs sites entries load) (compute-database-statistics db)
+   (size osize kvars kprocs globs sites entries) (compute-database-statistics db)
    (when (debugging 's "program statistics:")
      (printf ";   program size: \t~s \toriginal program size: \t~s\n" size osize)
      (printf ";   variables with known values: \t~s\n" kvars)
      (printf ";   known procedures: \t~s\n" kprocs)
      (printf ";   global variables: \t~s\n" globs)
      (printf ";   known call sites: \t~s\n" sites) 
-     (printf ";   database entries: \t~s \tload: \t~s\n" entries load) ) ) )
+     (printf ";   database entries: \t~s\n" entries) ) ) )
 
 
 ;;; Pretty-print expressions:
@@ -730,44 +730,88 @@
 
 ;;; Create foreign type checking expression:
 
-(define (foreign-type-check param type)
-  (case type
-    [(char unsigned-char) (if unsafe param `(##sys#foreign-char-argument ,param))]
-    [(int unsigned-int short unsigned-short) (if unsafe param `(##sys#foreign-fixnum-argument ,param))]
-    [(float double) (if unsafe param `(##sys#foreign-flonum-argument ,param))]
-    [(pointer)
-     `(if ,param
-	  ,(if unsafe
-	       param
-	       `(##sys#foreign-block-argument ,param) )
-	  '#f) ]
-    [(integer long) (if unsafe param `(##sys#foreign-integer-argument ,param))]
-    [(unsigned-integer unsigned-long)
-     (if unsafe
-	 param
-	 `(##sys#foreign-unsigned-integer-argument ,param) ) ]
-    [(c-pointer)
-     `(if ,param
-	  (##sys#foreign-pointer-argument ,param)
-	  '#f) ]
-    [(c-string)
-     `(if ,param
-	  ,(if unsafe 
-	       `(##sys#make-c-string ,param)
-	       `(##sys#make-c-string (##sys#foreign-string-argument ,param)) )
-	  '#f) ]
-    [else
-     (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
-	    => (lambda (t)
-		 (foreign-type-check param (if (vector? t) (vector-ref t 0) t)) ) ]
-	   [(pair? type)
-	    (case (car type)
-	      [(pointer function)
-	       `(if ,param
-		    (##sys#foreign-pointer-argument ,param)
-		    '#f) ]
-	      [else param] ) ]
-	   [else param] ) ] ) )
+(define foreign-type-check
+  (let ([tmap '((nonnull-u8vector . u8vector) (nonnull-u16vector . u16vector)
+		(nonnull-s8vector . s8vector) (nonnull-s16vector . s16vector)
+		(nonnull-u32vector . u32vector) (nonnull-s32vector . s32vector)
+		(nonnull-f32vector . f32vector) (nonnull-f64vector . f64vector) ) ] )
+    (lambda (param type)
+      (follow-without-loop
+       type
+       (lambda (t next)
+	 (case t
+	   [(char unsigned-char) (if unsafe param `(##sys#foreign-char-argument ,param))]
+	   [(int unsigned-int short unsigned-short) (if unsafe param `(##sys#foreign-fixnum-argument ,param))]
+	   [(float double) (if unsafe param `(##sys#foreign-flonum-argument ,param))]
+	   [(pointer byte-vector)
+	    (let ([tmp (gensym)])
+	      `(let ([,tmp ,param])
+		 (if ,tmp
+		     ,(if unsafe
+			  tmp
+			  `(##sys#foreign-block-argument ,tmp) )
+		     '#f) ) ) ]
+	   [(nonnull-pointer nonnull-byte-vector)
+	    (if unsafe
+		param
+		`(##sys#foreign-block-argument ,param) ) ]
+	   [(u8vector u16vector s8vector s16vector u32vector s32vector f32vector f64vector)
+	    (let ([tmp (gensym)])
+	      `(let ([,tmp ,param])
+		 (if ,tmp
+		     ,(if unsafe
+			  tmp
+			  `(##sys#foreign-number-vector-argument ',t ,tmp) )
+		     '#f) ) ) ]
+	   [(nonnull-u8vector nonnull-u16vector nonnull-s8vector nonnull-s16vector nonnull-u32vector nonnull-s32vector 
+			      nonnull-f32vector nonnull-f64vector)
+	    (if unsafe
+		param
+		`(##sys#foreign-number-vector-argument 
+		  ',(##sys#slot (assq t tmap) 1)
+		  ,param) ) ]
+	   [(integer long) (if unsafe param `(##sys#foreign-integer-argument ,param))]
+	   [(unsigned-integer unsigned-long)
+	    (if unsafe
+		param
+		`(##sys#foreign-unsigned-integer-argument ,param) ) ]
+	   [(c-pointer)
+	    (let ([tmp (gensym)])
+	      `(let ([,tmp ,param])
+		 (if ,tmp
+		     (##sys#foreign-pointer-argument ,tmp)
+		     '#f) ) ) ]
+	   [(nonnull-c-pointer)
+	    `(##sys#foreign-pointer-argument ,param) ]
+	   [(c-string c-string*)
+	    (let ([tmp (gensym)])
+	      `(let ([,tmp ,param])
+		 (if ,tmp
+		     ,(if unsafe 
+			  `(##sys#make-c-string ,tmp)
+			  `(##sys#make-c-string (##sys#foreign-string-argument ,tmp)) )
+		     '#f) ) ) ]
+	   [(nonnull-c-string nonnull-c-string*)
+	    (if unsafe 
+		`(##sys#make-c-string ,param)
+		`(##sys#make-c-string (##sys#foreign-string-argument ,param)) ) ]
+	   [else
+	    (cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
+		   => (lambda (t)
+			(next (if (vector? t) (vector-ref t 0) t)) ) ]
+		  [(pair? t)
+		   (case (car t)
+		     [(ref pointer function c-pointer)
+		      (let ([tmp (gensym)])
+			`(let ([,tmp ,param])
+			   (if ,tmp
+			       (##sys#foreign-pointer-argument ,tmp)
+			       '#f) ) )  ]
+		     [(nonnull-pointer nonnull-c-pointer)
+		      `(##sys#foreign-pointer-argument ,param) ]
+		     [else param] ) ]
+		  [else param] ) ] ) )
+       (lambda () (quit "foreign type `~S' refers to itself" type)) ) ) ) )
 
 
 ;;; Compute foreign-type conversions:
@@ -786,28 +830,75 @@
 	(list (vector-ref ft 1) a) )
       a) )
 
-(define (final-foreign-type t)
-  (cond [(##sys#hash-table-ref foreign-type-table t) 
-	 => (lambda (t)
-	      (final-foreign-type (if (vector? t) (vector-ref t 0) t))) ]
-	[else t] ) )
+(define (final-foreign-type t0)
+  (follow-without-loop
+   t0
+   (lambda (t next)
+     (cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
+	    => (lambda (t2)
+		 (next (if (vector? t2) (vector-ref t2 0) t2)) ) ]
+	   [else t] ) )
+   (lambda () (quit "foreign type `~S' refers to itself" t0)) ) )
 
 
 ;;; Compute foreign result size:
 
 (define (estimate-foreign-result-size type)
+  (follow-without-loop
+   type
+   (lambda (t next)
+     (case t
+       ((char int short bool void unsigned-short scheme-object unsigned-char unsigned-int) 0)
+       ((c-string nonnull-c-string c-pointer nonnull-c-pointer unsigned-integer long integer unsigned-long
+		  c-string* nonnull-c-string*)
+	(words->bytes 3) )
+       ((float double) 
+	(words->bytes 4) )		; possibly 8-byte aligned 32-bit double
+       (else
+	(cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
+	       => (lambda (t2)
+		    (next (if (vector? t2) (vector-ref t2 0) t2)) ) ]
+	      [(pair? t)
+	       (case (car t)
+		 [(ref nonnull-pointer pointer c-pointer nonnull-c-pointer function) (words->bytes 3)]
+		 [else 0] ) ]
+	      [else 0] ) ) ) )
+   (lambda () (quit "foreign type `~S' refers to itself" type)) ) )
+
+(define (estimate-foreign-result-location-size type)
+  (define (err t) 
+    (quit "can not compute size of location for foreign type `~S'" t) )
+  (follow-without-loop
+   type
+   (lambda (t next)
+     (case t
+       ((char int short bool unsigned-short unsigned-char unsigned-int long unsigned-long
+	      c-pointer pointer nonnull-c-pointer unsigned-integer integer float c-string
+	      nonnull-c-string c-string* nonnull-c-string*) 
+	(words->bytes 1) )
+       ((double)
+	(words->bytes 2) )
+       (else
+	(cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
+	       => (lambda (t2)
+		    (next (if (vector? t2) (vector-ref t2 0) t2)) ) ]
+	      [(pair? t)
+	       (case (car t)
+		 [(ref nonnull-pointer pointer c-pointer nonnull-c-pointer function) (words->bytes 1)]
+		 [else (err t)] ) ]
+	      [else (err t)] ) ) ) )
+   (lambda () (quit "foreign type `~S' refers to itself" type)) ) )
+
+
+;;; Convert result value, if a string:
+
+(define (finish-foreign-result type body)
   (case type
-    ((char int short bool void unsigned-short scheme-object unsigned-char unsigned-int) 0)
-    ((c-string float double c-pointer unsigned-integer long integer unsigned-long) 12)
-    (else
-     (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
-	    => (lambda (t)
-		 (estimate-foreign-result-size (if (vector? t) (vector-ref t 0) t)) ) ]
-	   [(pair? type)
-	    (case (car type)
-	      [(pointer function) 12]
-	      [else 0] ) ]
-	   [else 0] ) ) ) )
+    [(c-string) `(##sys#peek-c-string ,body '0)]
+    [(nonnull-c-string) `(##sys#peek-nonnull-c-string ,body '0)]
+    [(c-string*) `(##sys#peek-and-free-c-string ,body '0)]
+    [(nonnull-c-string*) `(##sys#peek-and-free-nonnull-c-string ,body '0)]
+    [else body] ) )
 
 
 ;;; Scan expression-node for variable usage:
@@ -909,24 +1000,6 @@
 	sorted) ) )
 
 
-;;; Write list of dependecies for makefile:
-
-(define (write-dependency-list strlst limit out)
-  (let loop ([len 0]
-	     [lst (reverse strlst)] )
-    (if (null? lst)
-	(newline out)
-	(let* ([d (car lst)]
-	       [rest (cdr lst)] 
-	       [len2 (+ len (string-length d) 1)] )
-	  (cond [(>= len2 limit)
-		 (fprintf out "\\~%\t ~a " d)
-		 (loop 0 rest) ]
-		[else 
-		 (fprintf out "~a " d)
-		 (loop len2 rest) ] ) ) ) ) )
-
-
 ;;; Some pathname operations:
 
 (define (chop-separator str)
@@ -946,67 +1019,97 @@
 ;;; Print version/usage information:
 
 (define (print-version)
-  (printf "This is ~a, ~a~%~a" product build-information copyright) )
+  (printf "This is ~a.~%~a~%~a" product build-information copyright) )
 
 (define (print-usage)
   (print-version)
   (newline)
   (display #<<EOF
-Usage: chicken {FILENAME | OPTION}
+Usage: chicken FILENAME OPTION ...
 
   FILENAME should be a complete source file name with extension, or "-" for
   standard input. OPTION may be one of the following:
 
+  General options:
+
     -help                       display this text and exit
     -version                    display compiler version and exit
     -verbose                    display information on compilation progress
     -quiet                      do not display compile information
-    -debug MODES                display debugging output for the given modes
-    -explicit-use               do not use units 'library' and 'eval' by default
+
+  File and pathname options:
+
     -output-file FILENAME       specifies output-filename, default is 'out.c'
     -include-path PATHNAME      specifies alternative path for included files
-    -database-size NUMBER       specifies size of analysis-database
+    -to-stdout                  write compiled file to stdout instead of file
+
+  Language options:
+
+    -feature SYMBOL             register feature identifier
+    -no-feature SYMBOL          unregister feature identifier
+    -srfi-7                     process source file as SRFI-7 configuration
+    -ffi-define SYMBOL          define preprocessor macro for ``easy'' FFI parser
+    -r5rs                       equivalent to `-hygienic -strict'
+
+  Syntax related options:
+
+    -case-insensitive           don't preserve case of read symbols
+    -hygienic                   use syntax-case macro package
+    -keyword-style STYLE        allow alternative keyword syntax (prefix or suffix)
+    -strict                     disable non-standard macros and symbolsyntax
+    -strict-srfi-0              disable non-standard macros except `cond-expand'
+    -strict-reader              disable non-standard read syntax
+    -strict-letrec              enable fully R5RS compliant `letrec'
+    -compile-time-macros-only   low-level macros are not made available at run-time
+    -hygienic-at-run-time       enable hygienic macros at run-time
+    -visit FILENAME             reads and macroexpands source file FILENAME
+
+  Translation options:
+
+    -explicit-use               do not use units 'library' and 'eval' by default
+    -check-syntax               stop compilation after macro-expansion
+    -analyze-only               stop compilation after first analysis pass
+
+  Debugging options:
+
+    -no-warnings                disable warnings
     -debug-level NUMBER         set level of available debugging information
-    -emit-debug-info            emit extended debugging information
-    -debug-calls                emit extra code to track non-loop procedure calls
-    -debug-loops                emit extra code to track all procedure calls
     -no-trace                   disable tracing information
     -profile FILENAME           executable emits profiling information 
-    -no-warnings                disable warnings
-    -optimize                   enable all optimizations
+
+  Optimization options:
+
     -optimize-level NUMBER      enable certain sets of optimization options
     -optimize-leaf-routines     enable leaf routine optimization
+    -lambda-lift                enable lambda-lifting
     -usual-integrations         assume standard procedures are not redefined
     -unsafe                     disable safety checks
     -block                      enable block-compilation
-    -inline                     enable inlining
-    -inline-limit NUMBER        maximum percentage of growth through inlining
-    -inline-passes NUMBER       maximal number of inlining passes
     -disable-interrupts         disable interrupts in compiled code
     -fixnum-arithmetic          assume all numbers are fixnums
     -benchmark-mode             fixnum mode, no interrupts and opt.-level 3
-    -no-fancy-ports             use only file- and std-ports
-    -no-winding-callcc          use non-winding semantics for call/cc
-    -feature SYMBOL             register feature identifier
-    -no-feature SYMBOL          unregister feature identifier
+
+  Configuration options:
+
+    -unit NAME                  compile file as a library unit
+    -uses NAME                  declare library unit as used.
     -heap-size NUMBER           specifies heap-size of compiled executable
+    -heap-initial-size NUMBER   specifies heap-size at startup time
+    -heap-growth PERCENTAGE     specifies growth-rate of expanding heap
+    -heap-shrinkage PERCENTAGE  specifies shrink-rate of contracting heap
     -nursery NUMBER
     -stack-size NUMBER          specifies nursery size of compiled executable
-    -case-sensitive             preserve case of read symbols
-    -check-syntax               abort compilation after macro-expansion
-    -expand-only                print macro-expanded source to file and exit
-    -to-stdout                  write compiled file to stdout instead of file
-    -srfi-7                     process source file as SRFI-7 configuration
-    -hygienic                   use syntax-case macro package
     -extend FILENAME            load file before compilation commences
-    -write-dependencies         output include-file dependencies
-    -write-distribution-dependencies
-                                output library unit dependencies
-    -dependency-output FILENAME destination for dependency output
     -prelude EXPRESSION         add expression to front of source file
     -postlude EXPRESSION        add expression to end of source file
     -prologue FILENAME          include file before main source file
     -epilogue FILENAME          include file after main source file
+    -dynamic                    compile as dynamically loadable code
+
+  Obscure options:
+
+    -debug MODES                display debugging output for the given modes
+    -compress-literals NUMBER   compile literals above threshold as strings
 
 EOF
 ) )
@@ -1026,3 +1129,121 @@ EOF
 	    (:optional prefix (gensym))
 	    (current-seconds)
 	    (random 1000) ) ) )
+
+
+;;; Register/lookup real names:
+;
+; - The real-name-table contains the following mappings:
+;
+;     <variable-alias> -> <variable>
+;     <lambda-id> -> <variable> or <variable-alias>
+
+(define (set-real-name! name rname)
+  (##sys#hash-table-set! real-name-table name rname) )
+
+(define (real-name var . db)
+  (define (resolve n)
+    (let ([n2 (##sys#hash-table-ref real-name-table n)])
+      (if n2
+	  (or (##sys#hash-table-ref real-name-table n2)
+	      n2) 
+	  n) ) )
+  (let ([rn (resolve var)])
+    (cond [(not rn) (##sys#symbol->qualified-string var)]
+	  [(pair? db)
+	   (let ([db (car db)])
+	     (let loop ([prev (##sys#symbol->qualified-string rn)] 
+			[container (get db var 'contained-in)] )
+	       (if container
+		   (let ([rc (resolve container)])
+		     (if (eq? rc container)
+			 prev
+			 (loop (sprintf "~A in ~A" prev rc)
+			       (get db container 'contained-in) ) ) )
+		   prev) ) ) ]
+	  [else (##sys#symbol->qualified-string rn)] ) ) )
+
+(define (real-name2 var db)
+  (and-let* ([rn (##sys#hash-table-ref real-name-table var)])
+    (real-name rn db) ) )
+
+(define (display-real-name-table)
+  (##sys#hash-table-for-each
+   (lambda (key val)
+     (printf "~S ~S~%" key val) )
+   real-name-table) )
+
+
+;;; We need this for constant folding:
+
+(define (string-null? x) (string-null? x))
+
+
+;;; Handle inline- and constant-definitions in visited files:
+
+(set! ##sys#visit-define-hook
+  (lambda (x)
+    (set! pending-canonicalizations (cons x pending-canonicalizations)) ) )
+
+
+;;; Dump node structure:
+
+(define (dump-nodes n)
+  (let loop ([i 0] [n n])
+    (let ([class (node-class n)]
+	  [params (node-parameters n)]
+	  [subs (node-subexpressions n)] 
+	  [ind (make-string i #\space)] 
+	  [i2 (+ i 2)] )
+      (printf "~%~A<~A ~S" ind class params)
+      (for-each (cut loop i2 <>) subs)
+      (write-char #\>) ) )
+  (newline) )
+
+
+;;; "#> ... <#" syntax:
+
+(define (enable-sharp-greater-read-syntax)
+  (set! ##sys#user-read-hook
+    (let ([old-hook ##sys#user-read-hook])
+      (lambda (char port)
+	(if (char=? #\> char)	       
+	    (let ([_ (read-char port)])	; swallow #\>
+	      (case (peek-char port)
+		[(#\!)
+		 (read-char port)
+		 (let ([text (scan-sharp-greater-string port)])
+		   `(declare (foreign-declare ,text)
+			     (foreign-parse ,text) ) ) ]
+		[(#\?)
+		 (read-char port)
+		 (let ([text (scan-sharp-greater-string port)])
+		   `(declare (foreign-parse ,text)) ) ]
+		[else
+		 `(declare 
+		    (foreign-declare 
+		     ,(scan-sharp-greater-string port) ) ) ] ) )
+	    (old-hook char port) ) ) ) ) )
+
+(define (scan-sharp-greater-string port)
+  (let ([out (open-output-string)])
+    (let loop ()
+      (let ([c (read-char port)])
+	(cond [(eof-object? c) (quit "unexpected end of `#> ... <#' sequence")]
+	      [(char=? c #\newline)
+	       (newline out)
+	       (set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) 
+	       (loop) ]
+	      [(char=? c #\<)
+	       (let ([c (read-char port)])
+		 (if (eqv? #\# c)
+		     (get-output-string out)
+		     (begin
+		       (write-char #\< out)
+		       (write-char c out) 
+		       (loop) ) ) ) ]
+	      [else
+	       (write-char c out)
+	       (loop) ] ) ) ) ) )
+
+(enable-sharp-greater-read-syntax)
