@@ -22,11 +22,11 @@
 ; **********************************************************************
 ;
 ;
-; (felix: I have stolen several optimizations from Eli Barzilay's Swindle code)
+; [felix] I have stolen several optimizations from Eli Barzilay's Swindle code - thanks anyway!
        
 
 #{tinyclos
-  tiny-clos-version every getl filter-in ensure-generic add-global-method
+  every1 every2 getl filter-in ensure-generic add-global-method
   compute-std-cpl top-sort std-tie-breaker build-transitive-closure build-constraints
   %allocate-instance %allocate-entity get-field set-field! lookup-slot-info the-slots-of-a-class
   getters-n-setters-for-class generic-invocation-generics applicable? more-specific? 
@@ -37,24 +37,33 @@
 
 (include "parameters")
 
-(define tiny-clos-version "1.2 (Chicken)")
-
 (declare
   (unit tinyclos)
   (uses extras)
   (usual-integrations)
   (fixnum) 
   (disable-interrupts)
-  (no-bound-checks)
-  (block-global method-caching-enabled method-cache-tag entity-tag symbol-vector)
-  (bound-to-procedure
-   every getl filter-in ensure-generic add-global-method
-   compute-std-cpl top-sort std-tie-breaker build-transitive-closure build-constraints
-   %allocate-instance %allocate-entity get-field set-field! lookup-slot-info
-   applicable? more-specific? ##sys#error ##sys#signal-hook
-   make-primitive-class method-cache-lookup make-method-cache
-   method-cache-update! hash-arg-list
-   quick-getl) )
+  (hide every1 every2 every2-eq? method-caching-enabled method-cache-tag symbol-vector ##tinyclos#slot-ref
+	symbol-vector getl make-method-cache set-field! method-cache-lookup %allocate-entity
+	hash-arg-list std-tie-breaker build-constraints %allocate-instance compute-std-cpl
+	get-field ensure-generic ##tinyclos#slot-set! lookup-slot-info
+	getters-n-setters-for-class the-slots-of-a-class make-primitive-class quick-getl
+	build-transitive-closure applicable? more-specific? method-cache-tag top-sort
+	filter-in) )
+
+(cond-expand
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks)
+    (bound-to-procedure
+     every1 every2 getl filter-in ensure-generic add-global-method
+     compute-std-cpl top-sort std-tie-breaker build-transitive-closure build-constraints
+     %allocate-instance %allocate-entity get-field set-field! lookup-slot-info
+     applicable? more-specific? ##sys#error ##sys#signal-hook
+     make-primitive-class method-cache-lookup make-method-cache
+     method-cache-update! hash-arg-list
+     quick-getl) ) ] )
 
 
 ;
@@ -149,22 +158,21 @@
 ;       COMPUTE-APPLY-METHODS
 ;
 
-;
-; OK, now let's get going.  But, as usual, before we can do anything
-; interesting, we have to muck around for a bit first.  First, we need  
-; to load the support library.
-;
-;
+
+(register-feature! 'tinyclos)
 
 
-(define (every test . lists)
-  (let scan ([tails lists])
-    (let loop ([tails2 tails])
-      (if (null? tails2)
-	  (and (apply test (map (lambda (x) (##sys#slot x 0)) tails))
-	       (scan (map (lambda (x) (##sys#slot x 1)) tails)) ) 
-	  (or (null? (##sys#slot tails2 0))
-	      (loop (##sys#slot tails2 1)) ) ) ) ) )
+(define (every1 test lst)
+  (let loop ([lst lst])
+    (or (null? lst)
+	(and (test (##sys#slot lst 0))
+	     (loop (##sys#slot lst 1)) ) ) ) )
+
+(define (every2 test list1 list2)
+  (let loop ([list1 list1] [list2 list2])
+    (or (null? list1)
+	(and (test (##sys#slot list1 0) (##sys#slot list2 0))
+	     (loop (##sys#slot list1 1) (##sys#slot list2 1)) ) ) ) )
 
 (define quick-getl
   (foreign-lambda* scheme-object ((scheme-object initargs) (scheme-object name) (scheme-object def)) "
@@ -183,7 +191,7 @@
 		     (cond ((null? tail)
 			    (if (pair? not-found)
 				(car not-found)
-				(##sys#error "getl: couldn't find item" name initargs)))
+				(##sys#error 'getl "couldn't find item" name initargs)))
 			   ((eq? (car tail) name) (cadr tail))
 			   (else (scan (cddr tail)))))))
       (scan initargs))))
@@ -207,11 +215,11 @@
   (cons method-cache-tag (make-vector (arithmetic-shift method-cache-size 1) #f)) )
 
 (define method-cache-lookup
-  (foreign-lambda* scheme-object ((scheme-object mcache) (int hash) (scheme-object classes)) "
+  (foreign-lambda* scheme-object ((scheme-object mcache) (scheme-object hash) (scheme-object classes)) "
     C_word v = C_block_item(mcache, 1);
     C_word clist, x, y;
     int free_index = -1;
-    int i = ((hash & (C_METHOD_CACHE_SIZE - 1)) << 1) & 0xffff,
+    int i = ((C_unfix(hash) & (C_METHOD_CACHE_SIZE - 1)) << 1) & 0xffff,
         i0, i2;
     for(i0 = i;; i = i2) {
       clist = C_block_item(v, i);
@@ -258,13 +266,13 @@
 	    (let ((can-go-in-now
 		    (filter-in
 		      (lambda (x)
-			(every (lambda (constraint)
+			(every1 (lambda (constraint)
 				 (or (not (eq? (cadr constraint) x))
 				     (memq (car constraint) result)))
 			       constraints))
 		      elements)))
 	      (if (null? can-go-in-now)
-		  (##sys#error "top-sort: invalid constraints")
+		  (##sys#error 'top-sort "invalid constraints")
 		  (let ((choice (if (null? (cdr can-go-in-now))
 				    (car can-go-in-now)
 				    (tie-breaker result
@@ -288,7 +296,7 @@
 					  min-elts)))
 		   (if (null? common)
 		       (if (null? (cdr pcpl))
-			   (##sys#error "std-tie-breaker: nothing valid")
+			   (##sys#error 'std-tie-breaker "nothing valid")
 			   (loop (cdr pcpl)))
 		       (car common)))))))))
 
@@ -345,7 +353,7 @@
 
 (define (%allocate-instance class nfields)
   (let ((instance (make-vector (+ nfields 3) #f)))
-    (##sys#vector->structure! instance)
+    (##core#inline "C_vector_to_structure" instance)
     (##sys#setslot instance 0 'instance)
     (##sys#setslot instance 1 class)
     instance))
@@ -361,17 +369,16 @@
 (define-record-printer (instance x out)
   (print-object x out) )
 
-
 (define entity-tag (gensym))
 
 (let ([default-proc
 	(lambda args
-	  (##sys#error "called entity without first setting proc"))] )
+	  (##sys#error '%allocate-entity "called entity without first setting proc"))] )
   (set! %allocate-entity 
     (lambda (class nfields name)
       (letrec ((vector (make-vector (+ nfields 5) #f))
 	       (closure (lambda args (apply (##sys#slot vector 1) args))))
-	(##sys#vector->structure! vector)
+	(##core#inline "C_vector_to_structure" vector)
 	(##sys#setslot vector 0 'entity)
 	(##sys#setslot vector 1 default-proc)
 	(##sys#setslot vector 2 class)
@@ -384,7 +391,8 @@
 		((>= i len)
 		 (##sys#setslot p2 i vector)
 		 (##sys#setslot p2 (add1 i) entity-tag)
-		 (##sys#vector->closure! p2 (##sys#peek-unsigned-integer closure 0))
+		 (##core#inline "C_vector_to_closure" p2)
+		 (##core#inline "C_copy_pointer" closure p2)
 		 p2)
 	      (##sys#setslot p2 i (##sys#slot closure i)) ) ) ) ) ) ) )
 
@@ -486,61 +494,63 @@
 ; Note that this implementation of class-of assumes the name of a the
 ; primitive classes that are set up later.
 ; 
-(define class-of
-  (lambda (x)
-    (cond ((null? x)       <null>)
-	  ((fixnum? x)     <exact>)
-	  ((boolean? x)    <boolean>)
-	  ((char? x)       <char>)
-	  ((eq? x (##core#undefined)) <void>)
-	  ((eof-object? x) <end-of-file>)
-	  ((%instance? x)  (%instance-class x))
-	  ((%entity? x)    (%entity-class x))
-	  ((symbol? x)     <symbol>)
-	  ((vector? x)     <vector>)
-	  ((pair? x)       <pair>)
-	  ((number? x)     <inexact>)
-	  ((string? x)     <string>)
-	  ((procedure? x)  <procedure>)
-	  ((port? x)  
-	   (if (input-port? x)
-	       <input-port>
-	       <output-port>) )
-	  ((##sys#bytevector? x) <byte-vector>)
-	  ((##sys#generic-structure? x)
-	   (case (##sys#slot x 0)
-	     [(dummy-environment) <environment>]
-	     [(hash-table) <hash-table>]
-	     [(queue) <queue>]
-	     [(condition) <condition>]
-	     [(char-set) <char-set>]
-	     [(time) <time>]
-	     [(lock) <lock>]
-	     [(mmap) <mmap>]
-	     [(u8vector) <u8vector>]
-	     [(s8vector) <s8vector>]
-	     [(u16vector) <u16vector>]
-	     [(s16vector) <s16vector>]
-	     [(u32vector) <u32vector>]
-	     [(s32vector) <s32vector>]
-	     [(f32vector) <f32vector>]
-	     [(f64vector) <f64vector>]
-	     [else <structure>] ) )
-	  (else (##sys#error "can not compute class of primitive object" x)) ) ) )
+(define (class-of x)
+  (cond [(null? x)       <null>]
+	[(fixnum? x)     <exact>]
+	[(boolean? x)    <boolean>]
+	[(char? x)       <char>]
+	[(eq? x (##core#undefined)) <void>]
+	[(eof-object? x) <end-of-file>]
+	[(%instance? x)  (%instance-class x)]
+	[(%entity? x)    (%entity-class x)]
+	[(symbol? x)     <symbol>]
+	[(vector? x)     <vector>]
+	[(pair? x)       <pair>]
+	[(number? x)     <inexact>]
+	[(string? x)     <string>]
+	[(procedure? x)  <procedure>]
+	[(port? x) (if (input-port? x) <input-port> <output-port>)]
+	[(##core#inline "C_pointerp" x) <pointer>]
+	[(##core#inline "C_taggedpointerp" x) <tagged-pointer>]
+	[(##core#inline "C_locativep" x) <locative>]
+	[(##sys#bytevector? x) <byte-vector>]
+	[(##sys#generic-structure? x)
+	 (case (##sys#slot x 0)
+	   [(dummy-environment) <environment>]
+	   [(array) <array>]
+	   [(hash-table) <hash-table>]
+	   [(queue) <queue>]
+	   [(condition) <condition>]
+	   [(char-set) <char-set>]
+	   [(time) <time>]
+	   [(lock) <lock>]
+	   [(mmap) <mmap>]
+	   [(promise) <promise>]
+	   [(u8vector) <u8vector>]
+	   [(s8vector) <s8vector>]
+	   [(u16vector) <u16vector>]
+	   [(s16vector) <s16vector>]
+	   [(u32vector) <u32vector>]
+	   [(s32vector) <s32vector>]
+	   [(f32vector) <f32vector>]
+	   [(f64vector) <f64vector>]
+	   [(tcp-listener) <tcp-listener>]
+	   [else <structure>] ) ]
+	[else (##sys#error 'class-of "can not compute class of primitive object" x)] ) )
 
 (define get-field
   (lambda (object field)
     (cond ((%instance? object) (%instance-ref object field))
 	  ((%entity?   object) (%entity-ref   object field))
 	  (else
-	   (##sys#signal-hook #:type-error "can only get-field of instances and entities" object)))))
+	   (##sys#signal-hook #:type-error 'get-field "can only get-field of instances and entities" object)))))
 
 (define set-field!
   (lambda (object field new-value)
     (cond ((%instance? object) (%instance-set! object field new-value))
 	  ((%entity?   object) (%entity-set!   object field new-value))
 	  (else
-	   (##sys#signal-hook #:type-error "can only set-field! of instances and entities" object)))))
+	   (##sys#signal-hook #:type-error 'set-field! "can only set-field! of instances and entities" object)))))
 
 
 
@@ -574,9 +584,9 @@
 				       (cons (car sups)
 					     so-far)))))
 		  (slots (apply append
-				(cons dslots
-				      (map class-direct-slots
-					   (cdr cpl)))))
+				dslots
+				(map class-direct-slots
+				     (cdr cpl))))
 		  (nfields 0)
 		  (field-initializers '())
 		  (allocator
@@ -696,6 +706,18 @@
   (##tinyclos#slot-ref class 'cpl) )
 
 
+;;; Inline procedures inside this module only:
+
+(eval-when (compile)
+  (define-macro (class-direct-slots class) `(##tinyclos#slot-ref ,class 'direct-slots))
+  (define-macro (class-direct-supers class) `(##tinyclos#slot-ref ,class 'direct-supers))
+  (define-macro (class-slots class) `(##tinyclos#slot-ref ,class 'slots))
+  (define-macro (class-name class) `(##tinyclos#slot-ref ,class 'name))
+  (define-macro (generic-methods generic) `(##tinyclos#slot-ref ,generic 'methods))
+  (define-macro (method-specializers method) `(##tinyclos#slot-ref ,method 'specializers))
+  (define-macro (method-procedure method) `(##tinyclos#slot-ref ,method 'procedure))
+  (define-macro (class-cpl class) `(##tinyclos#slot-ref ,class 'cpl)) )
+
 ;
 ; The next 7 clusters define the 6 initial classes.  It takes 7 to 6
 ; because the first and fourth both contribute to <class>.
@@ -849,7 +871,7 @@
 	    (lambda (m) 
 	      (let ([ms1 (method-specializers m)]
 		    [ms2 (method-specializers method)] )
-		(not (every (lambda (x y) (eq? x y)) ms1 ms2) ) ) )
+		(not (every2 (lambda (x y) (eq? x y)) ms1 ms2) ) ) )
 	    (##tinyclos#slot-ref generic 'methods))))
     (if (memq generic generic-invocation-generics)
 	(set! method-cache-tag (vector))
@@ -878,7 +900,7 @@
 				   ;it doesn't get cnm.
        (lambda args
 	 (apply (method-procedure (car (generic-methods generic)))
-		(cons #f args))))) ;But, the ONE time it is run,
+		#f args)))) ;But, the ONE time it is run,
 				   ;it needs to pass a dummy
 				   ;value for cnm!
 
@@ -906,7 +928,7 @@
 				       (if (null? tail)
 					   (##sys#slot lis 0)
 					   (lp tail)) ) ) ) ] )
-			     (lambda (args) (apply proc (cons #f args))) )
+			     (lambda (args) (apply proc #f args)) )
 			   (let ([x (compute-apply-methods generic)]
 				 [y ((compute-methods generic) args)] )
 			     (lambda (args) (x y args)) ) ) ] )
@@ -922,7 +944,7 @@
       (lambda (args)
 	(let ([applicable
 	       (filter-in (lambda (method)
-			    (every applicable?
+			    (every2 applicable?
 				   (method-specializers method)
 				   args))
 			  (generic-methods generic) ) ] )
@@ -978,7 +1000,7 @@
 			(when (null? tail)
 			  (##sys#error "call-next-method: no methods left" generic) ) ] )
 		      (apply (method-procedure (##sys#slot tail 0))
-			     (cons (one-step (##sys#slot tail 1)) args))))))
+			     (one-step (##sys#slot tail 1)) args)))))
 	  ((one-step methods)))))))
 
 (define applicable?
@@ -1014,7 +1036,7 @@
 	     [namestr 
 	      (cond [(symbol? name) (##sys#symbol->string name)]
 		    [(string? name) name]
-		    [else (##sys#signal-hook #:type-error "invalid class name" name)] ) ] 
+		    [else (##sys#signal-hook #:type-error 'initialize "invalid class name" name)] ) ] 
 	     [namestrlen (##sys#size namestr)] )
 	(##tinyclos#slot-set! class 'direct-slots dslots)
 	(##tinyclos#slot-set! class 'cpl (compute-cpl class))
@@ -1193,12 +1215,14 @@
 (define <end-of-file> (make-primitive-class "end-of-file"))
 (define <environment> (make-primitive-class "environment" <structure>))	; (Benedikt insisted on this)
 (define <hash-table> (make-primitive-class "hash-table" <structure>))
+(define <promise> (make-primitive-class "promise" <structure>))
 (define <queue> (make-primitive-class "queue" <structure>))
 (define <condition> (make-primitive-class "condition" <structure>))
 (define <char-set> (make-primitive-class "char-set" <structure>))
 (define <time> (make-primitive-class "time" <structure>))
 (define <lock> (make-primitive-class "lock" <structure>))
 (define <mmap> (make-primitive-class "mmap" <structure>))
+(define <array> (make-primitive-class "array" <structure>))
 (define <u8vector> (make-primitive-class "u8vector" <byte-vector>))
 (define <s8vector> (make-primitive-class "s8vector" <byte-vector>))
 (define <u16vector> (make-primitive-class "u16vector" <byte-vector>))
@@ -1207,6 +1231,11 @@
 (define <s32vector> (make-primitive-class "s32vector" <byte-vector>))
 (define <f32vector> (make-primitive-class "f32vector" <byte-vector>))
 (define <f64vector> (make-primitive-class "f64vector" <byte-vector>))
+(define <pointer> (make-primitive-class "pointer"))
+(define <tagged-pointer> (make-primitive-class "tagged-pointer" <pointer>))
+(define <locative> (make-primitive-class "locative"))
+(define <tcp-listener> (make-primitive-class "tcp-listener" <structure>))
+(define <c++-object> (make <class> 'name "c++-object" 'direct-supers (list <object>) 'direct-slots '(this)))
 
 (set! method-caching-enabled #t)
 
@@ -1268,16 +1297,20 @@
     (lambda (call-next-method x port)
       (fprintf port "generic ~A~%" (%entity-name x)) ) ) )
 
-(define (ensure-generic sym)
-  (let ([x (##sys#slot sym 0)])
-    (if (%entity? x)
-	x
-	(let ([g (make-generic (##sys#symbol->string sym))])
-	  (##sys#setslot sym 0 g)
-	  g) ) ) )
+(define ensure-generic
+  (let ([make-generic make-generic])
+    (lambda (x sym)
+      (if (%entity? x)
+	  x
+	  (make-generic (##sys#symbol->string sym)) ) ) ) )
 
-(define (add-global-method sym specializers proc)
-  (add-method (ensure-generic sym) (make-method specializers proc)) )
+(define add-global-method
+  (let ([make-method make-method]
+	[add-method add-method] )
+    (lambda (val sym specializers proc)
+      (let ([g (ensure-generic val sym)])
+	(add-method g (make-method specializers proc))
+	g) ) ) )
 
 (define (instance? x)
   (or (%instance? x) (%entity? x)) )
