@@ -1,6 +1,6 @@
 ;;;; tweaks.scm - Some inline-routines
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2003, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -33,12 +33,18 @@
 ; Germany
 
 
-(declare
-  (fixnum)
-  (disable-interrupts)
-  (no-bound-checks)
-  (no-procedure-checks)
-  (no-argc-checks) )
+(cond-expand
+ [compiler-unsafe
+  (declare
+    (fixnum)
+    (disable-interrupts)
+    (no-bound-checks)
+    (no-procedure-checks)
+    (no-argc-checks) ) ]
+ [else
+  (declare
+    (fixnum)
+    (disable-interrupts) ) ] )
 
 
 (define-inline (make-node c p s) (##sys#make-structure 'node c p s))
